@@ -43,10 +43,23 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <uim/uim.h>
-#include <uim/uim-helper.h>
-#include <uim/gettext.h>
+#include <locale.h>
 
+#include "uim/uim.h"
+#include "uim/uim-util.h"
+#include "uim/uim-helper.h"
+#include "uim/uim-im-switcher.h"
+#include "uim/config.h"
+#include "uim/gettext.h"
+#include "uim-cand-win-gtk.h"
+
+/* exported symbols */
+GtkIMContext *im_module_create(const gchar *context_id);
+void im_module_list(const GtkIMContextInfo ***contexts, int *n_contexts);
+void im_module_exit(void);
+void im_module_init(GTypeModule *type_module);
+
+/**/
 #define NR_CANDIDATES 20
 #define DEFAULT_SEPARATOR_STR "|"
 
@@ -59,26 +72,13 @@ struct preedit_segment {
   char *str;
 };
 
-struct candidate_window {
-  GtkWidget *top_win;
-  GtkWidget *clist;
-  GtkWidget *numlabel;
-  int nr_candidates;
-  int candidate_index;
-  int layout_begin;
-  int is_active;
-  int display_limit;
-  int select_count;
-  gboolean left;
-  GdkRectangle cursor;
-};
-
 typedef struct _IMUIMContext {
   /**/
   struct _GtkIMContext parent;
   struct _GtkIMContext *slave;
   uim_context uc;
-  struct candidate_window cwin;
+  UIMCandWinGtk *cwin;
+  gboolean cwin_is_active;
   int nr_psegs;
   struct preedit_segment *pseg;
   /**/
@@ -93,41 +93,45 @@ typedef struct _IMUIMContext {
 static IMUIMContext context_list;
 static IMUIMContext *focused_context = NULL;
 static gboolean disable_focused_context = FALSE;
+static gboolean snooper_installed = FALSE;
 
 static GObjectClass *parent_class;
 
-typedef struct _IMContextUIMClass 
+typedef struct _IMContextUIMClass
 {
   GtkIMContextClass parent_class;
 } IMContextUIMClass;
 
-static void im_uim_class_init (GtkIMContextClass *class);
+
+static void im_uim_class_init(GtkIMContextClass *class);
+static void im_uim_class_finalize(GtkIMContextClass *class);
+static void im_uim_init(IMUIMContext *uic);
 
 static void show_preedit(GtkIMContext *ic, GtkWidget *preedit_label);
 
 static void im_uim_helper_disconnect_cb(void);
 static void helper_read_cb(gpointer p, int fd, GdkInputCondition c);
 static GdkFilterReturn
-toplevel_window_candidate_cb (GdkXEvent *xevent, GdkEvent *ev, gpointer data);
-
-static void layout_candidate(IMUIMContext *uic);
-static void update_candidate(IMUIMContext *uic);
+toplevel_window_candidate_cb(GdkXEvent *xevent, GdkEvent *ev, gpointer data);
 
 static void
-im_uim_parse_helper_str(char *str);
+im_uim_parse_helper_str(const char *str);
 
 static gboolean get_user_defined_color(PangoColor *color, const gchar *uim_symbol);
 
+static gboolean
+uim_key_snoop(GtkWidget *grab_widget, GdkEventKey *key, gpointer data);
+
 static const GTypeInfo class_info = {
   sizeof(IMContextUIMClass),
   (GBaseInitFunc) NULL,
   (GBaseFinalizeFunc) NULL,
   (GClassInitFunc) im_uim_class_init,
-  (GClassFinalizeFunc)NULL,
+  (GClassFinalizeFunc)im_uim_class_finalize,
   NULL,/*for class data*/
   sizeof(IMUIMContext), /* size of instance */
   0,
-  (GInstanceInitFunc)NULL, /* constructor */
+  (GInstanceInitFunc) im_uim_init, /* constructor */
 };
 
 static GType type_im_uim = 0;
@@ -135,16 +139,20 @@ static GType type_im_uim = 0;
 #define IM_UIM_CONTEXT(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj),type_im_uim,IMUIMContext))
 
 static const GtkIMContextInfo im_uim_info = {
-  "uim",
-  "UIM()",
-  "gtk+",
-  "",
-  ""
+  "uim", /* id */
+  "uim", /* human-readable name*/
+  "uim", /* domain for gettext*/
+  LOCALEDIR,
+  "ja:ko:zh:*"
 };
 
-static
-void
-im_uim_commit_string(void *ptr, char *str)
+static const GtkIMContextInfo *im_uim_info_list = {
+  &im_uim_info
+};
+
+
+static void
+im_uim_commit_string(void *ptr, const char *str)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
   g_return_if_fail(str);
@@ -165,12 +173,12 @@ clear_cb(void *ptr)
 }
 
 static void
-pushback_cb(void *ptr, int attr, char *str)
+pushback_cb(void *ptr, int attr, const char *str)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
   g_return_if_fail(str);
-  
-  if(!strcmp(str,"")
+
+  if (!strcmp(str, "")
      && !(attr & (UPreeditAttr_Cursor | UPreeditAttr_Separator)))
     return;
 
@@ -207,6 +215,7 @@ convert_keyval(int key)
   case GDK_Next: return UKey_Next;
   case GDK_Home: return UKey_Home;
   case GDK_End: return UKey_End;
+  case GDK_Kanji:
   case GDK_Zenkaku_Hankaku: return UKey_Zenkaku_Hankaku;
   case GDK_Multi_key: return UKey_Multi_key;
   case GDK_Mode_switch: return UKey_Mode_switch;
@@ -226,10 +235,10 @@ convert_keyval(int key)
   case GDK_Hyper_R: return UKey_Hyper_key;
   }
 
-  if(key >= GDK_F1 && key <= GDK_F35) {
+  if (key >= GDK_F1 && key <= GDK_F35) {
     return key - GDK_F1 + UKey_F1;
   }
-  if(key >= GDK_KP_0 && key <= GDK_KP_9) {
+  if (key >= GDK_KP_0 && key <= GDK_KP_9) {
     return key - GDK_KP_0 + UKey_0;
   }
   if (key < 256) {
@@ -277,7 +286,7 @@ get_user_defined_color(PangoColor *color, const gchar *uim_symbol)
 {
   gboolean parsed = FALSE;
   gchar *literal = uim_symbol_value_str(uim_symbol);
-  if(literal != NULL && literal[0] != '\0') {
+  if (literal != NULL && literal[0] != '\0') {
     parsed = pango_color_parse(color, literal);
   }
   g_free(literal);
@@ -316,40 +325,40 @@ get_preedit_segment(struct preedit_segment *ps,
       }
       if (get_user_defined_color(&color, separator_fg_symbol))
       {
-	attr = pango_attr_foreground_new (color.red, color.green, color.blue);
+	attr = pango_attr_foreground_new(color.red, color.green, color.blue);
 	attr->start_index = begin;
 	attr->end_index = end;
-	pango_attr_list_change (attrs, attr);
+	pango_attr_list_change(attrs, attr);
       }
 
       if (get_user_defined_color(&color, separator_bg_symbol))
       {
-	attr = pango_attr_background_new (color.red, color.green, color.blue);
+	attr = pango_attr_background_new(color.red, color.green, color.blue);
 	attr->start_index = begin;
 	attr->end_index = end;
-	pango_attr_list_change (attrs, attr);
+	pango_attr_list_change(attrs, attr);
       }
     } else if (ps->attr & UPreeditAttr_Reverse) {
       if (get_user_defined_color(&color, "reversed-preedit-foreground")
 	  || pango_color_parse(&color, "#fff"))
       {
-	attr = pango_attr_foreground_new (color.red, color.green, color.blue);
+	attr = pango_attr_foreground_new(color.red, color.green, color.blue);
 	attr->start_index = begin;
 	attr->end_index = end;
-	pango_attr_list_change (attrs, attr);
+	pango_attr_list_change(attrs, attr);
       }
 
       if (get_user_defined_color(&color, "reversed-preedit-background")
 	  || pango_color_parse(&color, "#000"))
       {
-	attr = pango_attr_background_new (color.red, color.green, color.blue);
+	attr = pango_attr_background_new(color.red, color.green, color.blue);
 	attr->start_index = begin;
 	attr->end_index = end;
-	pango_attr_list_change (attrs, attr);
+	pango_attr_list_change(attrs, attr);
       }
     }
   }
-  str = (char *)realloc(str, strlen(str) + strlen(segment_str)+1);
+  str = (char *)realloc(str, strlen(str) + strlen(segment_str) + 1);
   strcat(str, segment_str);
   return str;
 }
@@ -370,7 +379,7 @@ im_uim_get_preedit_string(GtkIMContext *ic, gchar **str,
   tmp = g_strdup("");
   for (i = 0; i < uic->nr_psegs; i++) {
     if (uic->pseg[i].attr & UPreeditAttr_Cursor) {
-      pos = g_utf8_strlen(tmp,-1);
+      pos = g_utf8_strlen(tmp, -1);
     }
     if (attrs) {
       tmp = get_preedit_segment(&uic->pseg[i], *attrs, tmp);
@@ -394,14 +403,11 @@ im_uim_set_cursor_location(GtkIMContext *ic,
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
 
-  if(uic->cwin.left &&  uic->nr_psegs > 1){
+  if (uic->cwin->left &&  uic->nr_psegs > 1) {
     return;
   }
 
-  uic->cwin.cursor.x= area->x;
-  uic->cwin.cursor.y= area->y;
-  uic->cwin.cursor.height= area->height;
-  uic->cwin.cursor.width= area->width;
+  uim_cand_win_gtk_set_cursor_location(uic->cwin, area);
 
   uic->preedit_pos.x = area->x;
   uic->preedit_pos.y = area->y;
@@ -420,8 +426,8 @@ im_uim_commit_cb(GtkIMContext *ic,
 static void
 check_helper_connection()
 {
-  if(im_uim_fd < 0) {
-    im_uim_fd     = uim_helper_init_client_fd(im_uim_helper_disconnect_cb);
+  if (im_uim_fd < 0) {
+    im_uim_fd = uim_helper_init_client_fd(im_uim_helper_disconnect_cb);
     if (im_uim_fd >= 0) {
       read_tag = gdk_input_add(im_uim_fd, (GdkInputCondition)GDK_INPUT_READ,
 			       helper_read_cb, 0);
@@ -438,23 +444,27 @@ focus_in(GtkIMContext *ic)
   focused_context = uic;
   disable_focused_context = FALSE;
 
+  /* XXX:This is not recommended way!! */
+  if(snooper_installed == FALSE) {
+    snooper_id = gtk_key_snooper_install((GtkKeySnoopFunc)uim_key_snoop, NULL );
+    snooper_installed = TRUE;
+  }
+
   check_helper_connection();
 
   uim_prop_list_update(uic->uc);
   uim_prop_label_update(uic->uc);
 
   for (cc = context_list.next; cc != &context_list; cc = cc->next) {
-    if (cc != uic) {
-      if (cc->cwin.top_win) {
-	gtk_widget_hide(cc->cwin.top_win);
-      }
+    if (cc != uic && cc->cwin) {
+      gtk_widget_hide(GTK_WIDGET(cc->cwin));
     }
   }
-  
-  if(uic->cwin.top_win && uic->cwin.is_active){
-    gtk_widget_show(uic->cwin.top_win);
+
+  if (uic->cwin && uic->cwin_is_active) {
+    gtk_widget_show(GTK_WIDGET(uic->cwin));
   }
-  
+
   uim_helper_client_focus_in(uic->uc);
 }
 
@@ -463,33 +473,35 @@ focus_out(GtkIMContext *ic)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
 
+  if(snooper_installed == TRUE) {
+    gtk_key_snooper_remove(snooper_id);
+    snooper_installed = FALSE;
+  }
+
   uim_helper_client_focus_out(uic->uc);
 
-  if(uic->cwin.top_win){
-    gtk_widget_hide(uic->cwin.top_win);
+  if (uic->cwin) {
+    gtk_widget_hide(GTK_WIDGET(uic->cwin));
   }
 }
 
-void im_uim_reset(GtkIMContext *ic)
+static void
+im_uim_reset(GtkIMContext *ic)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
-
   uim_reset_context(uic->uc);
-
-  //  g_warning (_("reset called: implement me"));
-
 }
 
-void
+static void
 set_use_preedit(GtkIMContext *ic, gboolean use_preedit)
 {
   GtkWidget *preedit_window;
   GtkWidget *preedit_label;
-  if( use_preedit == FALSE ) {
+  if (use_preedit == FALSE ) {
     preedit_window = gtk_window_new(GTK_WINDOW_POPUP);
     preedit_label = gtk_label_new("");
     gtk_container_add(GTK_CONTAINER(preedit_window), preedit_label);
-    
+
     g_signal_connect(G_OBJECT(ic), "preedit-changed",
 		     G_CALLBACK(show_preedit), preedit_label );
     gtk_widget_show_all(preedit_window);
@@ -507,11 +519,11 @@ show_preedit(GtkIMContext *ic, GtkWidget *preedit_label)
   PangoAttrList *attrs;
 
   preedit_window = gtk_widget_get_parent(preedit_label);
-  
-  gtk_im_context_get_preedit_string (ic, &str, &attrs, &cursor_pos);
 
-  
-  if ( strlen (str) > 0) {
+  gtk_im_context_get_preedit_string(ic, &str, &attrs, &cursor_pos);
+
+
+  if (strlen(str) > 0) {
       gint x,y,w,h;
       PangoLayout *layout;
 
@@ -519,41 +531,43 @@ show_preedit(GtkIMContext *ic, GtkWidget *preedit_label)
       gtk_label_set_attributes(GTK_LABEL(preedit_label), attrs);
 
       gdk_window_get_origin(uic->win, &x, &y);
-      
-      gtk_window_move( GTK_WINDOW(preedit_window),
-		       x + uic->preedit_pos.x,
-		       y + uic->preedit_pos.y );
+
+      gtk_window_move(GTK_WINDOW(preedit_window),
+		      x + uic->preedit_pos.x,
+		      y + uic->preedit_pos.y );
 
       layout = gtk_label_get_layout(GTK_LABEL(preedit_label));
 
-      pango_layout_get_cursor_pos (layout, 0, NULL, NULL);
+      pango_layout_get_cursor_pos(layout, 0, NULL, NULL);
 
-      pango_layout_get_pixel_size (layout, &w, &h);
-      gtk_window_resize (GTK_WINDOW(preedit_window), w, h);
+      pango_layout_get_pixel_size(layout, &w, &h);
+      gtk_window_resize(GTK_WINDOW(preedit_window), w, h);
 
-      gtk_widget_show( preedit_window );
+      gtk_widget_show(preedit_window);
 
-    }
-  else {
+  } else {
       gtk_label_set_text(GTK_LABEL(preedit_label), "");
-      gtk_widget_hide( preedit_window );
-      gtk_window_resize( GTK_WINDOW(preedit_window), 0, 0 );
+      gtk_widget_hide(preedit_window);
+      gtk_window_resize(GTK_WINDOW(preedit_window), 0, 0);
   }
-  g_free (str);
-  pango_attr_list_unref (attrs);
-  
+  g_free(str);
+  pango_attr_list_unref(attrs);
+
 }
 
 static GdkFilterReturn
-toplevel_window_candidate_cb (GdkXEvent *xevent, GdkEvent *ev, gpointer data)
+toplevel_window_candidate_cb(GdkXEvent *xevent, GdkEvent *ev, gpointer data)
 {
   IMUIMContext *uic = data;
 
-  if(!uic)
+  if (!uic)
     return GDK_FILTER_CONTINUE;
 
-  if(uic->cwin.top_win && uic->cwin.is_active ){
-    layout_candidate(uic);
+  if (uic->cwin && uic->cwin_is_active) {
+    gint topwin_x, topwin_y;
+
+    gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
+    uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
   }
   return GDK_FILTER_CONTINUE;
 }
@@ -564,10 +578,36 @@ set_client_window(GtkIMContext *ic, GdkWindow *w)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(ic);
 
-  if(w) {
+  if (w) {
     g_object_ref(w);
+    uic->win = w;
+  } else {
+    if (uic->win)
+      g_object_unref(uic->win);
+    uic->win = NULL;
   }
-  uic->win = w;
+}
+
+static void
+index_changed_cb(UIMCandWinGtk *cwin, IMUIMContext *uic)
+{
+  g_return_if_fail(UIM_IS_CAND_WIN_GTK(cwin));
+
+  uim_set_candidate_index(uic->uc, uim_cand_win_gtk_get_index(cwin));
+}
+
+static void
+im_uim_init(IMUIMContext *uic)
+{
+  uic->win = NULL;
+  uic->menu = NULL;
+  uic->pseg = 0;
+  uic->nr_psegs = 0;
+
+  uic->cwin = uim_cand_win_gtk_new();
+  uic->cwin_is_active = FALSE;
+  g_signal_connect(G_OBJECT(uic->cwin), "index-changed",
+		   G_CALLBACK(index_changed_cb), uic);
 }
 
 
@@ -578,7 +618,7 @@ static void
 im_uim_finalize(GObject *obj)
 {
   IMUIMContext *uic = IM_UIM_CONTEXT(obj);
-  //  set_client_window(GTK_IM_CONTEXT(uic), NULL);
+  /* set_client_window(GTK_IM_CONTEXT(uic), NULL); */
 
   uic->next->prev = uic->prev;
   uic->prev->next = uic->next;
@@ -586,27 +626,28 @@ im_uim_finalize(GObject *obj)
   if (uic->menu) {
     gtk_widget_destroy(uic->menu);
   }
-  if (uic->cwin.top_win) {
-    gtk_widget_destroy(uic->cwin.top_win);
+  if (uic->cwin) {
+    gtk_widget_destroy(GTK_WIDGET(uic->cwin));
+    uic->cwin = NULL;
   }
   uic->menu = NULL;
 
   uim_release_context(uic->uc);
-  
+
   g_signal_handlers_disconnect_by_func(uic->slave,
-				       im_uim_commit_cb,
+				       (gpointer)im_uim_commit_cb,
 				       uic);
   g_object_unref(uic->slave);
-  parent_class->finalize (obj);
+  parent_class->finalize(obj);
 
-  if(uic == focused_context) {
+  if (uic == focused_context) {
     focused_context = NULL;
     disable_focused_context = TRUE;
   }
 }
 
 static void
-im_uim_class_init (GtkIMContextClass *class)
+im_uim_class_init(GtkIMContextClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS(class);
   parent_class = g_type_class_peek_parent(class);
@@ -622,8 +663,15 @@ im_uim_class_init (GtkIMContextClass *class)
   object_class->finalize = im_uim_finalize;
 }
 
+
 static void
-update_prop_list_cb(void *ptr, char *str)
+im_uim_class_finalize(GtkIMContextClass *class)
+{
+  uim_quit();
+}
+
+static void
+update_prop_list_cb(void *ptr, const char *str)
 {
   GString *tmp = g_string_new("");
   g_string_printf(tmp, "prop_list_update\ncharset=UTF-8\n%s", str);
@@ -633,7 +681,7 @@ update_prop_list_cb(void *ptr, char *str)
 }
 
 static void
-update_prop_label_cb(void *ptr, char *str)
+update_prop_label_cb(void *ptr, const char *str)
 {
   GString *tmp = g_string_new("");
   g_string_printf(tmp, "prop_label_update\ncharset=UTF-8\n%s", str);
@@ -642,256 +690,73 @@ update_prop_label_cb(void *ptr, char *str)
   g_string_free(tmp, TRUE);
 }
 
-
-/* This function is called by Gtk+, when user select candidate
-   by clicking candidate, etc. */
 static void
-select_candidate_cb(GtkWidget *clist,
-		    gint row, gint column,
-		    GdkEventButton *event,
-		    gpointer data)
+im_uim_send_im_list(void)
 {
-  IMUIMContext *uic = IM_UIM_CONTEXT(data);
-
-  if (uic->cwin.select_count == 0) {
-    int idx = row + uic->cwin.layout_begin;
-  
-    if( idx >= uic->cwin.nr_candidates ) {
-      gtk_clist_select_row(GTK_CLIST(uic->cwin.clist),
-			   uic->cwin.candidate_index - uic->cwin.layout_begin,
-			   0);
-      return;
-    }
-
-    uim_set_candidate_index(uic->uc, idx);
-    uic->cwin.candidate_index = idx;
+  int nr, i;
+  GString *msg;
+  const char *current_im_name = uim_get_current_im_name(focused_context->uc);
+  if(focused_context) {
+    nr = uim_get_nr_im(focused_context->uc);
   } else {
-    uic->cwin.select_count = 0;
-  }
-}
-
-static void
-init_candidate_win(IMUIMContext *uic)
-{
-  struct candidate_window *cwin = &uic->cwin;
-  GtkWidget *vbox;
-  GtkWidget *numlabel;
-  gchar *leftp;
-
-  if (cwin->top_win) {
-    return ;
+    return;
   }
 
-  leftp = uim_symbol_value_str("candidate-window-position");
- 
-  if(leftp && !strcmp(leftp,"left")) {
-    cwin->left = TRUE;
-  } else {
-    cwin->left = FALSE;
-  }
-  g_free(leftp);
+  msg = g_string_new("im_list\ncharset=UTF-8\n");
+  for (i = 0; i < nr; i++) {
+    const char *name = uim_get_im_name(focused_context->uc, i);
+    /* return value of uim_get_im_language() is an ISO 639-1
+       compatible language code such as "ja". Since it is unfriendly
+       for human reading, we convert it into friendly one by
+       uim_get_language_name_from_locale() here */
+    const char *langcode = uim_get_im_language(focused_context->uc, i);
+    const char *lang = uim_get_language_name_from_locale(langcode);
+    const char *short_desc = uim_get_im_short_desc(focused_context->uc, i);
 
-  cwin->top_win = gtk_window_new(GTK_WINDOW_POPUP);
-  cwin->clist = gtk_clist_new(2);
-  cwin->is_active = 0;
-  g_signal_connect(G_OBJECT(cwin->clist), "select_row",
-		   G_CALLBACK(select_candidate_cb),
-		   uic);
-  vbox = gtk_vbox_new(FALSE, 0);
-  numlabel = gtk_label_new("");
-  cwin->numlabel = numlabel;
-
-  gtk_box_pack_start(GTK_BOX(vbox), cwin->clist, FALSE, FALSE, 0);
-  gtk_box_pack_start(GTK_BOX(vbox), numlabel, FALSE, FALSE, 0);
-
-  gtk_container_add(GTK_CONTAINER(cwin->top_win), vbox);
-  {
-    gint x,y,w,h,d;
-    gdk_window_get_geometry(uic->win, &x, &y,&w,&h,&d);
-    gdk_window_get_origin(uic->win, &x, &y);
-    gtk_window_move(GTK_WINDOW(cwin->top_win), x , y+h);
-  }
-
-  gtk_clist_set_column_auto_resize( GTK_CLIST(cwin->clist), 0, TRUE );
-  gtk_clist_set_column_auto_resize( GTK_CLIST(cwin->clist), 1, TRUE );
-
-  gtk_clist_set_shadow_type( GTK_CLIST(cwin->clist), GTK_SHADOW_NONE );
-  
-  gtk_widget_show(cwin->clist);
-  gtk_widget_show(numlabel);
-  gtk_widget_show(vbox);
-  
-}
-
-static int
-trim_index(IMUIMContext *uic)
-{
-  struct candidate_window *cwin = &uic->cwin;
-  int changed = 0;
-
-  if(cwin->display_limit != 0){
-    
-    while (cwin->layout_begin + cwin->display_limit <= cwin->candidate_index) {
-      changed = 1;
-      cwin->layout_begin += cwin->display_limit;
-    }  
-    while (cwin->layout_begin > cwin->candidate_index) {
-      changed = 1;
-      cwin->layout_begin -= cwin->display_limit;
-    }
-    if (cwin->layout_begin < 0) {
-      changed = 1;
-      cwin->layout_begin = 0;
+    g_string_append(msg, name);
+    g_string_append(msg, "\t");
+    if(lang) g_string_append(msg, lang);
+    g_string_append(msg, "\t");
+    if(short_desc) g_string_append(msg, short_desc);
+    g_string_append(msg, "\t");
+    if(strcmp(name, current_im_name) == 0) {
+      g_string_append(msg, "selected");
+      g_print("%s has been selected\n", name);
     }
+    g_string_append(msg, "\n");
   }
-  else {
-    while (cwin->layout_begin + NR_CANDIDATES <= cwin->candidate_index) {
-      changed = 1;
-      cwin->layout_begin += NR_CANDIDATES;
-    }  
-    while (cwin->layout_begin > cwin->candidate_index) {
-      changed = 1;
-      cwin->layout_begin -= NR_CANDIDATES;     
-    }
-    if (cwin->layout_begin < 0) {
-      changed = 1;
-      cwin->layout_begin = 0;
-    }
-  }
-  return changed;
-}
-
-
-static void
-calc_candidate_window_size(struct candidate_window cwin, int *width, int *height)
-{
-  GtkRequisition req_clist;
-  GtkRequisition req_numlabel;
-  gtk_widget_size_request(GTK_WIDGET(cwin.clist), &req_clist);
-  gtk_widget_size_request(GTK_WIDGET(cwin.numlabel), &req_numlabel);
-
-  *height = req_clist.height + req_numlabel.height;
-  *width  = req_clist.width;
-
-  if(*width < 80)
-    *width = 80;
-
-  gtk_window_resize(GTK_WINDOW(cwin.top_win), *width, *height);
-}
-
-static void
-layout_candidate(IMUIMContext *uic)
-{
-  int  i, x, y, topwin_x, topwin_y;
-  int  sc_he, cw_he; /*screen height, candidate window height*/
-  int  sc_wi, cw_wi;
-  char idx[20];
-  char *old_cand_str = NULL;
-  char *new_cand_str = NULL;
-
-  trim_index(uic);
-
-  calc_candidate_window_size(uic->cwin, &cw_wi, &cw_he);
-
-  sc_he = gdk_screen_get_height(gdk_screen_get_default ());
-  sc_wi = gdk_screen_get_width (gdk_screen_get_default ());
-
-  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
-
-  if(sc_wi <  topwin_x + uic->cwin.cursor.x + cw_wi ) {
-    x = topwin_x + uic->cwin.cursor.x - cw_wi;
-  } else {
-    x = topwin_x + uic->cwin.cursor.x;
-  }
-
-  if(sc_he <  topwin_y + uic->cwin.cursor.y +  uic->cwin.cursor.height + cw_he ) {
-    y = topwin_y + uic->cwin.cursor.y - cw_he;
-  } else {
-    y = topwin_y + uic->cwin.cursor.y +  uic->cwin.cursor.height;
-  }
-
-  gtk_window_move(GTK_WINDOW(uic->cwin.top_win), x, y );
-  
-  /* XXX: Consider if display_limit == 0 */
-  for (i = 0; i < NR_CANDIDATES && i < uic->cwin.display_limit; i++) {
-    if (i < uic->cwin.nr_candidates - uic->cwin.layout_begin) {
-      uim_candidate new_cand = uim_get_candidate(uic->uc, uic->cwin.layout_begin + i, i);
-      g_return_if_fail(new_cand);
-      
-      new_cand_str = (gchar*)uim_candidate_get_cand_str(new_cand);
-      g_return_if_fail(new_cand_str);
-
-      gtk_clist_get_text(GTK_CLIST(uic->cwin.clist), i, 1, &old_cand_str);
-
-      /* Checking old cand and new cand is the same or not. If not, call update. 
-	 Caching new result is good idea for speed up and should be implement.*/
-      if(!old_cand_str || strcmp(new_cand_str, old_cand_str) != 0) {
-	update_candidate(uic);
-	uim_candidate_free(new_cand);
-	break;
-      }
-      uim_candidate_free(new_cand);
-    }
-  }
-  
-  uic->cwin.select_count = 1;
-
-  if(uic->cwin.candidate_index >= 0) {
-    gtk_clist_select_row(GTK_CLIST(uic->cwin.clist),
-			 uic->cwin.candidate_index - uic->cwin.layout_begin,
-			 0);
-    
-    sprintf(idx, "%d / %d", uic->cwin.candidate_index + 1 , uic->cwin.nr_candidates); 
-  } else {
-    gtk_clist_unselect_all(GTK_CLIST(uic->cwin.clist));
-    sprintf(idx, "- / %d", uic->cwin.nr_candidates); 
-  }
-  gtk_label_set_text( GTK_LABEL(uic->cwin.numlabel), idx);
-  
-}
-
-
-static void
-update_candidate(IMUIMContext *uic)
-{
-  char *buf[2];
-  int i;
-  uim_candidate cand = NULL;
-  gtk_clist_freeze(GTK_CLIST(uic->cwin.clist));
-  gtk_clist_clear(GTK_CLIST(uic->cwin.clist));
-  
-  /* XXX: Consider if display_limit == 0 */
-  for (i = 0; i < NR_CANDIDATES && i < uic->cwin.display_limit; i++) {
-    if (i < uic->cwin.nr_candidates - uic->cwin.layout_begin) {
-      cand = uim_get_candidate(uic->uc, uic->cwin.layout_begin + i, i);
-      buf[0] = (char *) uim_candidate_get_heading_label(cand);
-      buf[1] = (char *) uim_candidate_get_cand_str(cand);
-      gtk_clist_append(GTK_CLIST(uic->cwin.clist), buf);
-      uim_candidate_free(cand);
-    } else {
-      buf[0] = "";
-      buf[1] = "";
-      gtk_clist_append(GTK_CLIST(uic->cwin.clist), buf);
-    }
-  }
-  gtk_clist_thaw( GTK_CLIST(uic->cwin.clist) );
+  uim_helper_send_message(im_uim_fd, msg->str);
+  g_string_free(msg, TRUE);
 }
 
 static void
 cand_activate_cb(void *ptr, int nr, int display_limit)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  init_candidate_win(uic);
-  uic->cwin.is_active = 1;
-  uic->cwin.candidate_index = -1;
-  uic->cwin.nr_candidates = nr;
-  uic->cwin.display_limit = display_limit;
-  uic->cwin.layout_begin = 0;
-  layout_candidate(uic);
-  gtk_widget_show(uic->cwin.top_win);
+  gint topwin_x, topwin_y;
+  GSList *list = NULL;
+  uim_candidate cand;
+  gint i;
+
+  uic->cwin_is_active = TRUE;
+
+  for (i = 0; i < nr; i++) {
+    cand = uim_get_candidate(uic->uc, i, i % display_limit);
+    list = g_slist_append(list, cand);
+  }
+
+  uim_cand_win_gtk_set_candidates(uic->cwin, display_limit, list);
+
+  g_slist_foreach(list, (GFunc)uim_candidate_free, NULL);
+  g_slist_free(list);
+
+  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
+  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  gtk_widget_show(GTK_WIDGET(uic->cwin));
 
   if (uic->win) {
     GdkWindow *toplevel;
+
     toplevel = gdk_window_get_toplevel(uic->win);
     gdk_window_add_filter(toplevel,
 			  toplevel_window_candidate_cb,
@@ -900,81 +765,76 @@ cand_activate_cb(void *ptr, int nr, int display_limit)
 }
 
 
+
 /* This function called by libuim.
    Selected by keyboard, etc. */
 static void
 cand_select_cb(void *ptr, int index)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  uic->cwin.candidate_index = index;
-  layout_candidate(uic);
+  gint topwin_x, topwin_y;
+
+  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
+  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  g_signal_handlers_block_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
+  uim_cand_win_gtk_set_index(uic->cwin, index);
+  g_signal_handlers_unblock_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
 }
 
 static void
 cand_shift_page_cb(void *ptr, int direction)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
+  gint topwin_x, topwin_y;
 
-  if(direction){
-    if(uic->cwin.nr_candidates < uic->cwin.candidate_index + uic->cwin.display_limit )
-      uic->cwin.candidate_index = 0;
-    else
-      uic->cwin.candidate_index += uic->cwin.display_limit;
-  }
-  else {
-    if(uic->cwin.candidate_index - uic->cwin.display_limit < 0)
-      uic->cwin.candidate_index = uic->cwin.nr_candidates - 1;
-    else
-      uic->cwin.candidate_index -= uic->cwin.display_limit;
-  }
-  layout_candidate(uic);
-  uim_set_candidate_index(uic->uc, uic->cwin.candidate_index);
+  gdk_window_get_origin(uic->win, &topwin_x, &topwin_y);
+  uim_cand_win_gtk_layout(uic->cwin, topwin_x, topwin_y);
+  g_signal_handlers_block_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
+  uim_cand_win_gtk_shift_page(uic->cwin, direction);
+  uim_set_candidate_index(uic->uc, uic->cwin->candidate_index);
+  g_signal_handlers_unblock_by_func(uic->cwin, (gpointer)index_changed_cb, uic);
 }
 
 static void
 cand_deactivate_cb(void *ptr)
 {
   IMUIMContext *uic = (IMUIMContext *)ptr;
-  if(uic->cwin.top_win) {
-    gtk_widget_hide(uic->cwin.top_win);
-  }
 
-  if (uic->win) {
+  uic->cwin_is_active = FALSE;
+
+  if (uic->cwin) {
     GdkWindow *toplevel;
+
+    gtk_widget_hide(GTK_WIDGET(uic->cwin));
+    uim_cand_win_gtk_clear_candidates(uic->cwin);
     toplevel = gdk_window_get_toplevel(uic->win);
     gdk_window_remove_filter(toplevel, toplevel_window_candidate_cb, uic);
   }
-
-  /*  uic->nr_psegs = 0; */
-  uic->cwin.is_active = 0;
 }
 
 
 GtkIMContext *
-im_module_create (const gchar *context_id)
+im_module_create(const gchar *context_id)
 {
-  GObject *obj = g_object_new(type_im_uim, NULL);
-  IMUIMContext *uic = IM_UIM_CONTEXT(obj);
+  GObject *obj;
+  IMUIMContext *uic;
+  const char *im_name;
 
-  uic->win = NULL;
-  uic->menu = NULL;
-  uic->pseg = 0;
-  uic->nr_psegs = 0;
+  g_return_val_if_fail(context_id, NULL);
+  g_return_val_if_fail(!strcmp(context_id, "uim"), NULL);
 
-  uic->cwin.top_win = NULL;
-  uic->cwin.clist = NULL;
-  uic->cwin.select_count = 0;
+  obj = g_object_new(type_im_uim, NULL);
+  uic = IM_UIM_CONTEXT(obj);
 
-  if (!strncmp(context_id, "uim-", 4)) {
-    context_id = &context_id[4];
+  if(uim_init() == -1) {
+    parent_class->finalize(obj);
+    return NULL;
   }
-
-  uim_init();
-
-  uic->uc =  uim_create_context(uic, "UTF-8",
-				NULL, (char *)context_id,
-				uim_iconv,
-				im_uim_commit_string);
+  im_name = uim_get_default_im_name(setlocale(LC_ALL, NULL));
+  uic->uc = uim_create_context(uic, "UTF-8",
+			       NULL, im_name,
+			       uim_iconv,
+			       im_uim_commit_string);
   if (uic->uc == NULL) {
     parent_class->finalize(obj);
     return NULL;
@@ -995,7 +855,7 @@ im_module_create (const gchar *context_id)
 				cand_select_cb,
 				cand_shift_page_cb,
 				cand_deactivate_cb);
-  
+
   uim_prop_list_update(uic->uc);
 
 
@@ -1014,61 +874,73 @@ im_module_create (const gchar *context_id)
 }
 
 void
-im_module_list(GtkIMContextInfo ***contexts,
+im_module_list(const GtkIMContextInfo ***contexts,
 	       int *n_contexts)
 {
-  uim_context uc;
-  int nr, i;
-  static GtkIMContextInfo **info_list;
-
-  uim_init();
-  uc = uim_create_context(NULL, "UTF-8",
-			  NULL, NULL, uim_iconv, NULL);
-  if (uc == NULL)
-    nr = 0;
-  else
-    nr = uim_get_nr_im(uc);
-
-  info_list = malloc(sizeof(GtkIMContextInfo*)*nr);
-  for (i = 0; i < nr; i++) {
-    const char *name = uim_get_im_name(uc, i);
-    const char *lang = uim_get_im_language(uc, i);
-    char s[100];
-
-    info_list[i] = malloc(sizeof(GtkIMContextInfo));
-
-    sprintf(s, "uim-%s", name);
-    info_list[i]->context_id = g_strdup(s);
-
-    sprintf(s, "uim-%s (%s)", name, lang);
-    info_list[i]->context_name = g_strdup(s);
-
-    info_list[i]->domain = "gtk+";
-    info_list[i]->domain_dirname = "";
-    info_list[i]->default_locales = g_strdup(lang);
-  }
-  uim_release_context(uc);
-  *contexts = info_list;
-  *n_contexts = nr;
+  *contexts = &im_uim_info_list;
+  *n_contexts = 1;
 }
 
 static void
-im_uim_parse_helper_str(char *str)
+im_uim_parse_helper_str_im_change(const char *str)
+{
+  IMUIMContext *cc;
+  gchar **lines = g_strsplit(str, "\n", -1);
+  gchar *im_name = lines[1];
+  if(g_str_has_prefix(str, "im_change_this_text_area_only") == TRUE) {
+    if(focused_context) {
+      uim_switch_im(focused_context->uc, im_name);
+      uim_prop_list_update(focused_context->uc);
+    }
+  } else if(g_str_has_prefix(str, "im_change_whole_desktop") == TRUE) {
+    for (cc = context_list.next; cc != &context_list; cc = cc->next) {
+      uim_switch_im(cc->uc, im_name);
+    }
+  } else if(g_str_has_prefix(str, "im_change_this_application_only") == TRUE) {
+    if(focused_context) {    
+      for (cc = context_list.next; cc != &context_list; cc = cc->next) {
+	uim_switch_im(cc->uc, im_name);
+      }
+    }
+  }
+  g_strfreev(lines);
+}
+
+static void
+im_uim_parse_helper_str(const char *str)
 {
   gchar **lines;
-  
-  if(focused_context && !disable_focused_context) {
-    if(strncmp("prop_list_get", str, 13) == 0){
+
+  if (g_str_has_prefix(str, "im_change") == TRUE) {
+    im_uim_parse_helper_str_im_change(str);
+  } else if (g_str_has_prefix(str, "prop_update_custom") == TRUE) {
+    IMUIMContext *cc;
+    lines = g_strsplit(str, "\n", 0);
+    if (lines && lines[0] && lines[1] && lines[2]) {
+      for (cc = context_list.next; cc != &context_list; cc = cc->next) {
+	uim_prop_update_custom(cc->uc, lines[1], lines[2]);
+	break;  /* all custom variables are global */
+      }
+      g_strfreev(lines);
+    }
+  } else if (focused_context && !disable_focused_context) {
+    if (g_str_has_prefix(str, "prop_list_get") == TRUE){
       uim_prop_list_update(focused_context->uc);
-    } else if(strncmp("prop_label_get", str, 14) == 0) {
+    } else if (g_str_has_prefix(str, "prop_label_get") == TRUE) {
       uim_prop_label_update(focused_context->uc);
-    } else if(strncmp("prop_activate", str, 13) == 0) {
+    } else if (g_str_has_prefix(str, "prop_activate") == TRUE) {
       lines = g_strsplit(str, "\n", 0);
-      if(lines && lines[0]) {
+      if (lines && lines[0]) {
 	uim_prop_activate(focused_context->uc, lines[1]);
 	g_strfreev(lines);
       }
-    } else if(strncmp("focus_in", str, 8) == 0) {
+    } else if (g_str_has_prefix(str, "im_list_get") == TRUE) {
+      im_uim_send_im_list();
+    } else if (g_str_has_prefix(str, "commit_string")) {
+      lines = g_strsplit(str, "\n", 0);
+      if (focused_context && lines && lines[0] && lines[1])
+	g_signal_emit_by_name(focused_context, "commit", lines[1]);
+    } else if (g_str_has_prefix(str, "focus_in") == TRUE) {
       disable_focused_context = TRUE;
       /* We shouldn't do "focused_context = NULL" here, because some
 	 window manager has some focus related bugs. */
@@ -1076,12 +948,10 @@ im_uim_parse_helper_str(char *str)
   }
 }
 
-
 static void
 helper_read_cb(gpointer p, int fd, GdkInputCondition c)
 {
   char *tmp;
-
   uim_helper_read_proc(fd);
   while ((tmp = uim_helper_get_message())) {
     im_uim_parse_helper_str(tmp);
@@ -1097,14 +967,14 @@ im_uim_helper_disconnect_cb(void)
 }
 
 /* XXX:This is not a recommended way!! */
-gboolean
+static gboolean
 uim_key_snoop(GtkWidget *grab_widget, GdkEventKey *key, gpointer data)
 {
   if (focused_context) {
     int rv;
     int kv = convert_keyval(key->keyval);
     int mod = convert_modifier(key->state);
-    
+
     if (key->type == GDK_KEY_RELEASE) {
       rv = uim_release_key(focused_context->uc, kv, mod);
     } else {
@@ -1128,13 +998,15 @@ im_module_init(GTypeModule *type_module)
 				GTK_TYPE_IM_CONTEXT,
 				"GtkIMContextUIM",
 				&class_info, 0);
+  uim_cand_win_gtk_register_type(type_module);
 
   /* XXX:This is not recommended way!! */
-  snooper_id = gtk_key_snooper_install( (GtkKeySnoopFunc)uim_key_snoop, NULL );
+  snooper_id = gtk_key_snooper_install((GtkKeySnoopFunc)uim_key_snoop, NULL );
+  snooper_installed = TRUE;
 }
 
 void
-im_module_exit()
+im_module_exit(void)
 {
   if (im_uim_fd != -1) {
     uim_helper_close_client_fd(im_uim_fd);
