@@ -105,15 +105,10 @@
 	   (desc (custom-symbol-rec-desc srec)))
       desc)))
 
-;; TODO
+;; only accepts single strict-key-str (not or'ed, not a variable reference)
 (define key-definition?
   (lambda (def)
-    #t))
-
-;; TODO
-(define prefix-tag-style-key-str?
-  (lambda (str)
-    (string? str)))
+    (valid-strict-key-str? def)))
 
 (define-record 'custom-group-rec
   '((sym   #f)
@@ -121,15 +116,36 @@
     (desc  "")))
 
 (define custom-define-group
-  (lambda (sym label desc)
-    (let ((grec (custom-group-rec-new sym label desc)))
-      (if (not (custom-group-rec sym))
+  (lambda (gsym label desc)
+    (let ((grec (custom-group-rec-new gsym label desc)))
+      (if (not (custom-group-rec gsym))
 	  (set! custom-group-rec-alist (cons grec
 					     custom-group-rec-alist))))))
 
 (define custom-group-rec
-  (lambda (sym)
-    (assq sym custom-group-rec-alist)))
+  (lambda (gsym)
+    (assq gsym custom-group-rec-alist)))
+
+;; API
+(define custom-group-label
+  (lambda (gsym)
+    (custom-group-rec-label (custom-group-rec gsym))))
+
+;; API
+(define custom-group-desc
+  (lambda (gsym)
+    (custom-group-rec-desc (custom-group-rec gsym))))
+
+;; API
+(define custom-group-subgroups
+  (lambda (gsym)
+    (let ((groups (filter-map (lambda (pair)
+				(let ((primary-grp (car pair))
+				      (subgrp (cdr pair)))
+				  (and (eq? gsym primary-grp)
+				       subgrp)))
+			      custom-subgroup-alist)))
+      (reverse groups))))
 
 ;; API
 (define custom-list-groups
@@ -301,27 +317,6 @@
   (lambda (sym)
     (custom-rec-desc (custom-rec sym))))
 
-;; API
-(define custom-group-label
-  (lambda (sym)
-    (custom-group-rec-label (custom-group-rec sym))))
-
-;; API
-(define custom-group-desc
-  (lambda (sym)
-    (custom-group-rec-desc (custom-group-rec sym))))
-
-;; API
-(define custom-group-subgroups
-  (lambda (sym)
-    (let ((groups (filter-map (lambda (pair)
-				(let ((primary-grp (car pair))
-				      (subgrp (cdr pair)))
-				  (and (eq? sym primary-grp)
-				       subgrp)))
-			      custom-subgroup-alist)))
-      (reverse groups))))
-
 ;; API
 (define custom-canonical-value-as-string
   (lambda (sym)
@@ -639,7 +634,7 @@
   "Canna server name"
   "long description will be here.")
 
-(define-custom 'canna-server-name #f
+(define-custom 'canna-server-name ""
   '(canna cannaserver)
   '(string ".*")
   "Canna server name"
