@@ -36,69 +36,53 @@
 (define generic-candidate-op-count 1)
 (define generic-nr-candidate-max 10)
 
-(define generic-context-rk-context
+(define generic-context-rec-spec
+  (append
+   context-rec-spec
+   '((rk-context         ()) ;; #f?
+     (rk-nth             0)
+     (on                 ()) ;; #f?
+     (candidate-op-count 0)
+     (raw-commit         #f)
+     (converting         #f))))
+(define-record 'generic-context generic-context-rec-spec)
+(define generic-context-new-internal generic-context-new)
+
+(define generic-context-new
+  (lambda (id im rule back)
+    (let ((pc (generic-context-new-internal id im)))
+      (generic-context-set-rk-context!
+       pc (rk-context-new rule #f back))
+      pc)))
+(define generic-context-flush
   (lambda (pc)
-    (car (nthcdr 0 pc))))
-(define generic-context-set-rk-context!
-  (lambda (pc rkc)
-    (set-car! (nthcdr 0 pc) rkc)))
-(define generic-context-rk-nth
-  (lambda (pc)
-    (car (nthcdr 1 pc))))
-(define generic-context-set-rk-nth!
-  (lambda (pc n)
-    (set-car! (nthcdr 1 pc) n)))
-(define generic-context-on
-  (lambda (pc)
-    (car (nthcdr 2 pc))))
-(define generic-context-set-on!
-  (lambda (pc n)
-    (set-car! (nthcdr 2 pc) n)))
-(define generic-context-candidate-op-count
-  (lambda (pc)
-    (car (nthcdr 3 pc))))
-(define generic-context-set-candidate-op-count!
-  (lambda (pc n)
-    (set-car! (nthcdr 3 pc) n)))
-(define generic-context-raw-commit
-  (lambda (pc)
-    (car (nthcdr 4 pc))))
-(define generic-context-set-raw-commit!
-  (lambda (pc n)
-    (set-car! (nthcdr 4 pc) n)))
+    (generic-context-set-rk-nth! pc 0)
+    (generic-context-set-candidate-op-count! pc 0)
+    (generic-context-set-converting! pc #f)))
+
 
 (define generic-update-preedit
-  (lambda (id pc)
+  (lambda (pc)
     (if (generic-context-raw-commit pc)
 	(generic-context-set-raw-commit! pc #f)
 	(let* ((rkc (generic-context-rk-context pc))
 	       (cs (rk-current-seq rkc))
 	       (n (generic-context-rk-nth pc)))
-	  (im-clear-preedit id)
+	  (im-clear-preedit pc)
 	  (im-pushback-preedit
-	   id preedit-reverse
+	   pc preedit-reverse
 	   (if cs
 	       (nth n (cadr cs))
 	       (rk-pending rkc)))
-	  (im-update-preedit id)))))
-
-;; (context nth)
-(define generic-context-new
-  (lambda (rule back)
-    (let ((c (copy-list '(() 0 () 0 #f))))
-      (generic-context-set-rk-context!
-       c (rk-context-new rule #f back))
-      (generic-context-set-rk-nth! c 0)
-      (generic-context-set-on! c #f)
-      c)))
+	  (im-update-preedit pc)))))
 
 (define generic-commit-raw
-  (lambda (pc id)
-    (im-commit-raw id)
+  (lambda (pc)
+    (im-commit-raw pc)
     (generic-context-set-raw-commit! pc #t)))
 
-(define generic-proc-on-mode
-  (lambda (id pc key state)
+(define generic-proc-input-state
+  (lambda (pc key state)
     (let* ((rkc (generic-context-rk-context pc))
 	   (n (generic-context-rk-nth pc))
 	   (cs (cadr (rk-current-seq rkc)))
@@ -108,10 +92,94 @@
 	   (begin
 	     (rk-flush rkc)
 	     (generic-context-set-on! pc #f)
-	     (generic-update-prop-label pc id)
-	     (im-update-mode id 0)
+	     (generic-update-prop-label pc)
+	     (generic-update-prop-list pc)
+	     (im-update-mode pc 0)
 	     #f)
 	   #t)
+       (if (generic-prev-candidate-key? key state)
+	   (begin
+	     (generic-context-set-converting! pc #t)
+	     (generic-proc-converting-state pc key state)
+	     #f)
+	   #t)
+       (if (generic-next-candidate-key? key state)
+	   (begin
+	     (generic-context-set-converting! pc #t)
+	     (generic-proc-converting-state pc key state)
+	     #f)
+	   #t)
+       (if (or
+	    (generic-prev-page-key? key state)
+	    (generic-next-page-key? key state))
+	   #f
+	   #t)
+       (if (generic-backspace-key? key state)
+	   (begin
+	     (if (not (rk-backspace rkc))
+		 (generic-commit-raw pc))
+	     (generic-context-set-rk-nth! pc 0)
+	     (im-deactivate-candidate-selector pc)
+	     #f)
+	   #t)
+       (if (generic-commit-key? key state)
+	   (let ((cs (rk-current-seq rkc)))
+	     (if (> (length (cadr cs)) 0)
+		 (begin
+		   (im-commit pc (nth (generic-context-rk-nth pc) (cadr cs)))
+		   (im-deactivate-candidate-selector pc)
+		   (rk-flush rkc)
+		   (generic-context-flush pc))
+		 (begin
+		   (im-commit-raw pc)
+		   (rk-flush rkc)
+		   (im-update-preedit pc)))
+	     #f)
+	   #t)
+       (if (symbol? key)
+	   (begin
+	     (rk-flush rkc)
+	     (generic-commit-raw pc)
+	     (generic-context-flush pc)
+	     #f)
+	   #t)
+       (if (and (modifier-key-mask state)
+		(not (shift-key-mask state)))
+	   (begin
+	     (generic-commit-raw pc)
+	     #f)
+	   #t)
+       (begin
+	 (set! res
+	       (rk-push-key!
+		rkc
+		(charcode->string key)))
+	 #t))
+      (if (not (rk-partial? rkc))
+	  (let ((cs (rk-current-seq rkc)))
+	    (if (= (length (cadr cs)) 1)
+		(begin
+		  (im-commit pc
+			     (nth (generic-context-rk-nth pc) (cadr cs)))
+		  (generic-context-set-rk-nth! pc 0)
+		  (generic-context-set-candidate-op-count! pc 0)
+		  (im-deactivate-candidate-selector pc)
+		  (rk-flush rkc)))))
+      (if res
+	  (begin
+	    (im-commit pc (nth (generic-context-rk-nth pc) res))
+	    (generic-context-set-rk-nth! pc 0)
+	    (generic-context-set-candidate-op-count! pc 0)
+	    (im-deactivate-candidate-selector pc))
+	  ))))
+
+(define generic-proc-converting-state
+  (lambda (pc key state)
+    (let* ((rkc (generic-context-rk-context pc))
+	   (n (generic-context-rk-nth pc))
+	   (cs (cadr (rk-current-seq rkc)))
+	   (res))
+      (and
        (if (generic-prev-candidate-key? key state)
 	   (if (> (length (cadr cs)) 0)
 	       (begin
@@ -126,12 +194,12 @@
 		  (+ 1 (generic-context-candidate-op-count pc)))
 		 (if (>= (generic-context-candidate-op-count pc)
 			 generic-candidate-op-count)
-		     (im-select-candidate id n))
+		     (im-select-candidate pc n))
 		 #f)
 	       (begin
-		 (im-commit-raw id)
+		 (im-commit-raw pc)
 		 (rk-flush rkc)
-		 (im-update-preedit id)
+		 (im-update-preedit pc)
 		 #f))
 	   #t)
        (if (generic-next-candidate-key? key state)
@@ -147,222 +215,204 @@
 		      (= (generic-context-candidate-op-count pc)
 			 generic-candidate-op-count)
 		      generic-use-candidate-window?)
-		     (im-activate-candidate-selector id (length cs) generic-nr-candidate-max))
+		     (im-activate-candidate-selector pc (length cs) generic-nr-candidate-max))
 		 (if (>= (generic-context-candidate-op-count pc)
 			 generic-candidate-op-count)
 		     (begin
 		       (if (>= (+ n 1) (length cs))
 			   (set! n -1))
-		       (im-select-candidate id (+ n 1))))
+		       (im-select-candidate pc (+ n 1))))
 		 #f)
 	       (begin
-		 (im-commit-raw id)
+		 (im-commit-raw pc)
 		 (rk-flush rkc)
-		 (im-update-preedit id)
+		 (im-update-preedit pc)
 		 #f))
 	   #t)
        (if (and (generic-prev-page-key? key state)
 		(<= generic-candidate-op-count (generic-context-candidate-op-count pc)))
 	   (begin
-	     (im-shift-page-candidate id #f)
+	     (im-shift-page-candidate pc #f)
 	     #f)
 	   #t)
        (if (and (generic-next-page-key? key state)
 		(<= generic-candidate-op-count (generic-context-candidate-op-count pc)))
 	   (begin
-	     (im-shift-page-candidate id #t)
+	     (im-shift-page-candidate pc #t)
 	     #f)
 	   #t)
        (if (generic-backspace-key? key state)
 	   (begin
 	     (if (not (rk-backspace rkc))
-		 (generic-commit-raw pc id))
+		 (generic-commit-raw pc))
 	     (generic-context-set-rk-nth! pc 0)
-	     (im-deactivate-candidate-selector id)
+	     (im-deactivate-candidate-selector pc)
 	     #f)
 	   #t)
        (if (generic-commit-key? key state)
 	   (let ((cs (rk-current-seq rkc)))
 	     (if (> (length (cadr cs)) 0)
 		 (begin
-		   (im-commit id (nth (generic-context-rk-nth pc) (cadr cs)))
-		   (generic-context-set-rk-nth! pc 0)
-		   (generic-context-set-candidate-op-count! pc 0)
+		   (im-commit pc (nth (generic-context-rk-nth pc) (cadr cs)))
+		   (generic-context-flush pc)
 		   (rk-flush rkc)
-		   (im-deactivate-candidate-selector id))
+		   (im-deactivate-candidate-selector pc))
 		 (begin
-		   (im-commit-raw id)
+		   (im-commit-raw pc)
 		   (rk-flush rkc)
-		   (im-update-preedit id)))
+		   (im-update-preedit pc)))
 	     #f)
 	   #t)
        (if (symbol? key)
 	   (begin
 	     (rk-flush rkc)
-	     (generic-commit-raw pc id)
+	     (generic-commit-raw pc)
 	     (generic-context-set-rk-nth! pc 0)
 	     #f)
 	   #t)
        (if (and (modifier-key-mask state)
 		(not (shift-key-mask state)))
 	   (begin
-	     (generic-commit-raw pc id)
+	     (generic-commit-raw pc)
 	     #f)
 	   #t)
-       (begin
-	 (set! res
-	       (rk-push-key!
-		rkc
-		(charcode->string key)))
-	 #t))
-      (if (not (rk-partial? rkc))
-	  (let ((cs (rk-current-seq rkc)))
-	    (if (= (length (cadr cs)) 1)
-		(begin
-		  (im-commit id
-			     (nth (generic-context-rk-nth pc) (cadr cs)))
-		  (generic-context-set-rk-nth! pc 0)
-		  (generic-context-set-candidate-op-count! pc 0)
-		  (im-deactivate-candidate-selector id)
-		  (rk-flush rkc)))))
-      (if res
-	  (begin
-	    (im-commit id (nth (generic-context-rk-nth pc) res))
-	    (generic-context-set-rk-nth! pc 0)
-	    (generic-context-set-candidate-op-count! pc 0)
-	    (im-deactivate-candidate-selector id))
-	  ))))
+       (let ((cs (rk-current-seq rkc)))
+	 (if (> (length (cadr cs)) 0)
+	     (im-commit pc
+			(nth (generic-context-rk-nth pc) (cadr cs))))
+	     (generic-context-flush pc)
+	     (im-deactivate-candidate-selector pc)
+	     (rk-flush rkc)
+	     (generic-proc-input-state pc key state)
+	 )))))
+  
 
 (define generic-proc-off-mode
-  (lambda (id pc key state)
+  (lambda (pc key state)
     (and
      (if (generic-on-key? key state)
 	 (begin
 	   (generic-context-set-on! pc #t)
-	   (generic-update-prop-label pc id)
-	   (im-update-mode id 1)
+	   (generic-update-prop-label pc)
+	   (generic-update-prop-list pc)
+	   (im-update-mode pc 1)
 	   #f)
 	 #t)
      ;;
-     (generic-commit-raw pc id))))
+     (generic-commit-raw pc))))
 
 (define generic-key-press-handler
-  (lambda (id key state)
-    (let* ((c (find-context id))
-	   (pc (context-data c)))
-      (if (control-char? key)
-	  (im-commit-raw id)
-	  (if (generic-context-on pc)
-	      (generic-proc-on-mode id pc key state)
-	      (generic-proc-off-mode id pc key state)))
-      (generic-update-preedit id pc)
-      ())))
-
-(define generic-key-release-handler
-  (lambda (id key state)
-    (let* ((c (find-context id))
-	   (pc (context-data c)))
-      (if (or (control-char? key)
-	      (not (generic-context-on pc)))
-	  ;; don't discard key release event for apps
-	  (generic-commit-raw pc id)))))
-
-(define generic-reset-handler
-  (lambda (id)
+  (lambda (pc key state)
+    (if (control-char? key)
+	(im-commit-raw pc)
+	(if (generic-context-on pc)
+	    (if (generic-context-converting pc)
+		(generic-proc-converting-state pc key state)
+		(generic-proc-input-state pc key state))
+	    (generic-proc-off-mode pc key state)))
+    (generic-update-preedit pc)
     ()))
 
+(define generic-key-release-handler
+  (lambda (pc key state)
+    (if (or (control-char? key)
+	    (not (generic-context-on pc)))
+	;; don't discard key release event for apps
+	(generic-commit-raw pc))))
+
+(define generic-reset-handler
+  (lambda (pc)
+    (let ((rkc (generic-context-rk-context pc)))
+      (rk-flush rkc))))
+
 (define generic-mode-handler
-  (lambda (id mode)
-    (let* ((c (find-context id))
-	   (pc (context-data c))
-	   (rkc (generic-context-rk-context pc)))
+  (lambda (pc mode)
+    (let ((rkc (generic-context-rk-context pc)))
       (if (= mode 0)
 	  (generic-context-set-on! pc #f)
 	  (generic-context-set-on! pc #t))
       (rk-flush rkc)
-      (generic-update-preedit id pc))))
+      (generic-update-preedit pc)
+      (generic-update-prop-label pc))))
 
 (define generic-get-candidate-handler
-  (lambda (id idx accel-enum-hint)
-    (let* ((c (find-context id))
-	   (pc (context-data c))
-	   (rkc (generic-context-rk-context pc))
+  (lambda (pc idx accel-enum-hint)
+    (let* ((rkc (generic-context-rk-context pc))
 	   (cs (cadr (rk-current-seq rkc)))
 	   (cand (car (nthcdr idx cs))))
       (list cand (digit->string (+ idx 1))))))
 
 (define generic-set-candidate-index-handler
-  (lambda (id idx)
-    (let* ((c (find-context id))
-	   (pc (context-data c))
-	   (rkc (generic-context-rk-context pc)))
+  (lambda (pc idx)
+    (let ((rkc (generic-context-rk-context pc)))
       (generic-context-set-rk-nth! pc idx))))
 
 (define generic-init-handler
-  (lambda (id init-handler)
-    (init-handler id nil)
-    (im-clear-mode-list id)
-    (im-pushback-mode-list id "RAW")
-    (im-pushback-mode-list
-     id
-     (im-name (context-im (find-context id))))
-    (im-update-mode-list id)
-    (im-update-mode id 0)
-    (generic-update-prop-list id)
-))
+  (lambda (id im init-handler)
+    (let ((pc (init-handler id im #f)))
+      (im-clear-mode-list pc)
+      (im-pushback-mode-list pc "RAW")
+      (im-pushback-mode-list pc (im-name im))
+      (im-update-mode-list pc)
+      (im-update-mode pc 0)
+      (generic-update-prop-list pc)
+      pc)))
 
 (define generic-prop-handler
-  (lambda (id message)
-    (let* ((c (find-context id))
-	   (pc (context-data c))
-	   (rkc (generic-context-rk-context pc)))
+  (lambda (pc message)
+    (let ((rkc (generic-context-rk-context pc)))
 
       (rk-flush rkc)
-      (generic-update-preedit id pc)
+      (generic-update-preedit pc)
       
-      (if (string-equal? message
-			 "prop_generic_raw")
+      (if (string=? message
+		    "prop_generic_raw")
 	  (begin
-	    (im-update-mode id 0)
+	    (im-update-mode pc 0)
 	    (generic-context-set-on! pc #f)))
-      (if (string-equal? message
-			 "prop_generic_on")
+      (if (string=? message
+		    "prop_generic_on")
 	  (begin
-	    (im-update-mode id 1)
+	    (im-update-mode pc 1)
 	    (generic-context-set-on! pc #t)))
-      (generic-update-prop-label pc id))))
+      (generic-update-prop-label pc id)
+      (generic-update-prop-list id))))
 
 (define generic-update-prop-label
-  (lambda (pc id)
+  (lambda (pc)
     (let*  ((state (generic-context-on pc))
 	    (str ""))
       (if state
 	  (set! str "O\ton\n")
 	  (set! str "o\toff\n"))
-      (im-update-prop-label id str))))
+      (im-update-prop-label pc str))))
 
 (define generic-update-prop-list
-  (lambda (id) 
-    (let* ((c (find-context id))
-	   (pc (context-data c))
-	   (state (generic-context-on pc))
-	   (name (im-name (context-im c)))
-	   (str "branch\t"))
+  (lambda (pc)
+    (let* ((state (generic-context-on pc))
+	   (name (im-name (generic-context-im pc)))
+	   (str ""))
       (if state
 	  (set! str "O\ton\n")
 	  (set! str "o\toff\n"))
       (set! str (string-append
 		 "branch\t" str
-		 "leaf\tO\tOn\tRaw input mode\tprop_generic_on\n"
-		 "leaf\to\t" name "\t " name " mode\t\tprop_generic_raw\n"))
-      (im-update-prop-list id str)
+		 "leaf\to\tRaw Input Mode\tRaw input mode\tprop_generic_raw\t"
+		 (if state
+		     "\n"
+		     "*\n")
+		 "leaf\tO\t" name " Mode\t " name " mode\tprop_generic_on\t"
+		 (if state
+		     "*\n"
+		     "\n")))
+      (im-update-prop-list pc str)
       )))
 
 
 (define generic-register-im
-  (lambda (name lang code init)
+  (lambda (name lang code short-desc init)
     (register-im
-     name lang code init generic-init-handler nil
+     name lang code short-desc init generic-init-handler #f
      generic-mode-handler
      generic-key-press-handler
      generic-key-release-handler
