@@ -34,47 +34,28 @@
 
 
 
-;; accessor
-
 ; to decrease of typing cost, internal procedure omit "context"
 ; table name is held in C code. Same table has same table id.
-(define hk-context-table-id
-  (lambda (hkc)
-    (car (nthcdr 0 hkc))))
-(define hk-context-set-table-id!
-  (lambda (hkc id)
-    (set-car! (nthcdr 0 hkc) id)))
-
-(define hk-left-string-list
-  (lambda (hkc)
-    (car (nthcdr 1 hkc))))
-
-(define hk-set-left-string-list!
-  (lambda (hkc str-list)
-    (set-car! (nthcdr 1 hkc) str-list)))
-
-(define hk-context-right-string-list
-  (lambda (hkc)
-    (car (nthcdr 2 hkc))))
-
-(define hk-context-set-right-string-list!
-  (lambda (hkc str-list)
-    (set-car! (nthcdr 2 hkc) str-list)))
+(define-record 'hk-context
+  '((table-id          ())
+    (left-string-list  ())
+    (right-string-list ())))
+(define hk-context-new-internal hk-context-new)
 
 (define hk-context-new
   (lambda (table)
-    (let ((res (copy-list '( ()()())))
+    (let ((res (hk-context-new-internal))
 	  (table-id (open-table table)))
     (if (> 0 table-id)
-	nil
+	#f
 	(begin
 	  (hk-context-set-table-id! res table-id)
 	  res)))))
 
 (define hk-left-string-append
   (lambda (hkc str-list)
-    (let ((l (car (nthcdr 1 hkc))))
-      (set-car! (nthcdr 1 hkc) '(l str-list))
+    (let ((l (car (hk-context-left-string-list hkc))))
+      (hk-context-set-left-string-list! hkc '(l str-list))
 )))
 
 
@@ -82,11 +63,11 @@
 
 (define hk-get-left-string
   (lambda (hkc mode)
-    (car (nthcdr 1 hkc))))
+    (car (hk-context-left-string-list hkc))))
 
 (define hk-get-right-string
   (lambda (hkc)
-    (car (nthcdr 2 hkc))))
+    (car (hk-context-right-string-list hkc))))
 
 (define hk-push-key!
   (lambda (hkc str state)
@@ -241,7 +222,7 @@
 		    str-list)
 	    (hk-make-str-list
 	     (find-entry-matched-complete table-id (string-append str "\t"))))
-	  nil
+	  #f
       ))))
 
 
@@ -257,7 +238,7 @@
 			  (set! str s)))
 		    str-list)
 	    (find-entry-matched-complete table-id (string-append str " ")))
-	  nil
+	  #f
 	  ))))
 
 (define hk-find-longer-entry-exist?
