@@ -96,13 +96,8 @@
 
 
 (define ipa-init-handler
-  (lambda (id arg)
-    (let* ((c (find-context id))
-	   (cc (context-data c)))
-      (set-context-data!
-       c
-       (generic-context-new ipa-rule #f))
-      #f)))
+  (lambda (id im arg)
+    (generic-context-new id im ipa-rule #f)))
 
 (generic-register-im
- 'ipa "" "UTF-8" ipa-init-handler)
+ 'ipa "" "UTF-8" (N_ "International Phonetic Alphabet") ipa-init-handler)
