@@ -29,7 +29,7 @@
 ;;;;
 
 ;; Japanese EUC
-(define ja-rk-rule
+(define ja-rk-rule-basic
   '(
     ((("-"). ())("" "" ""))
     (((","). ())("" "" ""))
@@ -63,7 +63,7 @@
     ((("?"). ())("" "" "?"))
     ((("/"). ())("" "" "/"))
     ((("_"). ())("" "" "_"))
-    
+
     ((("1"). ())("1" "1" "1"))
     ((("2"). ())("2" "2" "2"))
     ((("3"). ())("3" "3" "3"))
@@ -83,10 +83,18 @@
 
     ((("x" "a"). ())("" "" ""))
     ((("x" "i"). ())("" "" ""))
+    ((("x" "y" "i"). ())("" "" ""))
     ((("x" "u"). ())("" "" ""))
     ((("x" "e"). ())("" "" ""))
+    ((("x" "y" "e"). ())("" "" ""))
     ((("x" "o"). ())("" "" ""))
 
+    ((("l" "a"). ())("" "" ""))
+    ((("l" "i"). ())("" "" ""))
+    ((("l" "u"). ())("" "" ""))
+    ((("l" "e"). ())("" "" ""))
+    ((("l" "o"). ())("" "" ""))
+
     ((("k" "k"). ("k"))("" "" ""))
 
     ((("k" "a"). ())("" "" ""))
@@ -100,12 +108,6 @@
     ((("k" "y" "e"). ())("" "" ""))
     ((("k" "y" "o"). ())("" "" ""))
 
-    ((("k" "w" "a"). ())("" "" ""))
-    ((("k" "w" "i"). ())("" "" ""))
-    ((("k" "w" "u"). ())("" "" ""))
-    ((("k" "w" "e"). ())("" "" ""))
-    ((("k" "w" "o"). ())("" "" ""))
- 
     ((("g" "g"). ("g"))("" "" ""))
 
 
@@ -121,12 +123,6 @@
     ((("g" "y" "e"). ())("" "" "ގ"))
     ((("g" "y" "o"). ())("" "" "ގ"))
 
-    ((("g" "w" "a"). ())("" "" "ގ"))
-    ((("g" "w" "i"). ())("" "" "ގ")) 
-    ((("g" "w" "u"). ())("" "" "ގ")) 
-    ((("g" "w" "e"). ())("" "" "ގ")) 
-    ((("g" "w" "o"). ())("" "" "ގ"))
-
     ((("s" "s"). ("s"))("" "" ""))
 
     ((("s" "s"). ())("s" "s" "s"))
@@ -135,25 +131,13 @@
     ((("s" "u"). ())("" "" ""))
     ((("s" "e"). ())("" "" ""))
     ((("s" "o"). ())("" "" ""))
-    
+
     ((("s" "y" "a"). ())("" "" ""))
     ((("s" "y" "i"). ())("" "" ""))
     ((("s" "y" "u"). ())("" "" ""))
     ((("s" "y" "e"). ())("" "" ""))
     ((("s" "y" "o"). ())("" "" ""))
-    
-    ((("s" "h" "a"). ())("" "" ""))
-    ((("s" "h" "i"). ())("" "" ""))
-    ((("s" "h" "u"). ())("" "" ""))
-    ((("s" "h" "e"). ())("" "" ""))
-    ((("s" "h" "o"). ())("" "" ""))
 
-    ((("s" "w" "a"). ())("" "" ""))
-    ((("s" "w" "i"). ())("" "" ""))
-    ((("s" "w" "u"). ())("" "" ""))
-    ((("s" "w" "e"). ())("" "" ""))
-    ((("s" "w" "o"). ())("" "" "")) 
-    
     ((("z" "z"). ("z"))("" "" ""))
 
     ((("z" "a"). ())("" "" ""))
@@ -166,7 +150,7 @@
     ((("z" "y" "u"). ())("" "" "ގ"))
     ((("z" "y" "e"). ())("" "" "ގ"))
     ((("z" "y" "o"). ())("" "" "ގ"))
-    
+
     ((("j" "j"). ("j"))("" "" ""))
 
     ((("j" "a"). ())("" "" "ގ"))
@@ -196,30 +180,18 @@
     ((("t" "y" "e"). ())("" "" ""))
     ((("t" "y" "o"). ())("" "" ""))
 
-    ((("t" "w" "a"). ())("Ȥ" "ȥ" "Ď"))
-    ((("t" "w" "i"). ())("Ȥ" "ȥ" "Ď"))
-    ((("t" "w" "u"). ())("Ȥ" "ȥ" "Ď"))
-    ((("t" "w" "e"). ())("Ȥ" "ȥ" "Ď"))
-    ((("t" "w" "o"). ())("Ȥ" "ȥ" "Ď"))
-
-    ((("t" "h" "a"). ())("Ƥ" "ƥ" "Î"))
-    ((("t" "h" "i"). ())("Ƥ" "ƥ" "Î"))
-    ((("t" "h" "u"). ())("Ƥ" "ƥ" "Î"))
-    ((("t" "h" "e"). ())("Ƥ" "ƥ" "Î"))
-    ((("t" "h" "o"). ())("Ƥ" "ƥ" "Î"))
-    
     ((("t" "s" "a"). ())("Ĥ" "ĥ" ""))
     ((("t" "s" "i"). ())("Ĥ" "ĥ" ""))
     ((("t" "s" "u"). ())(""   ""   ""))
     ((("t" "s" "e"). ())("Ĥ" "ĥ" ""))
     ((("t" "s" "o"). ())("Ĥ" "ĥ" ""))    
 
-    ((("c" "h" "a"). ())("" "" ""))
-    ((("c" "h" "i"). ())("" "" ""))
-    ((("c" "h" "u"). ())("" "" ""))
-    ((("c" "h" "e"). ())("" "" ""))
-    ((("c" "h" "o"). ())("" "" ""))
-    
+    ((("c" "y" "a"). ())("" "" ""))
+    ((("c" "y" "i"). ())("" "" ""))
+    ((("c" "y" "u"). ())("" "" ""))
+    ((("c" "y" "e"). ())("" "" ""))
+    ((("c" "y" "o"). ())("" "" ""))
+
     ((("x" "t" "u"). ())("" "" ""))
     ((("x" "t" "s" "u"). ())("" "" ""))
     ((("c" "c"). ("c"))("" "" ""))
@@ -238,19 +210,6 @@
     ((("d" "y" "e"). ())("¤" "¥" "ގ"))
     ((("d" "y" "o"). ())("¤" "¥" "ގ"))
 
-    ((("d" "h" "a"). ())("Ǥ" "ǥ" "Îގ"))
-    ((("d" "h" "i"). ())("Ǥ" "ǥ" "Îގ"))
-    ((("d" "h" "u"). ())("Ǥ" "ǥ" "Îގ"))
-    ((("d" "h" "e"). ())("Ǥ" "ǥ" "Îގ"))
-    ((("d" "h" "o"). ())("Ǥ" "ǥ" "Îގ"))
-
-
-    ((("d" "w" "a"). ())("ɤ" "ɥ" " ގ"))
-    ((("d" "w" "i"). ())("ɤ" "ɥ" " ގ"))
-    ((("d" "w" "u"). ())("ɤ" "ɥ" "Ďގ"))
-    ((("d" "w" "e"). ())("ɤ" "ɥ" " ގ"))
-    ((("d" "w" "o"). ())("ɤ" "ɥ" " ގ"))
-
     ((("n" "n"). ())("" "" ""))
     ((("n"). ())("" "" ""))
     ((("n" ""). ())("" "" ""))
@@ -274,7 +233,7 @@
     ((("h" "u"). ())("" "" ""))
     ((("h" "e"). ())("" "" ""))
     ((("h" "o"). ())("" "" ""))
-    
+
     ((("h" "y" "a"). ())("Ҥ" "ҥ" "ˎ"))
     ((("h" "y" "i"). ())("Ҥ" "ҥ" "ˎ"))
     ((("h" "y" "u"). ())("Ҥ" "ҥ" "ˎ"))
@@ -288,15 +247,13 @@
     ((("f" "u"). ())("" "" ""))
     ((("f" "e"). ())("դ" "ե" "̎"))
     ((("f" "o"). ())("դ" "ե" "̎"))
-    
+
     ((("f" "y" "a"). ())("դ" "ե" "̎"))
     ((("f" "y" "i"). ())("դ" "ե" "̎"))
     ((("f" "y" "u"). ())("դ" "ե" "̎"))
     ((("f" "y" "e"). ())("դ" "ե" "̎"))
     ((("f" "y" "o"). ())("դ" "ե" "̎"))
 
-    ((("f" "w" "u"). ())("դ" "ե" "   ")) 
-
     ((("b" "b"). ("b"))("" "" ""))
 
     ((("b" "a"). ())("" "" "ʎ"))
@@ -304,7 +261,7 @@
     ((("b" "u"). ())("" "" "̎"))
     ((("b" "e"). ())("" "" "͎"))
     ((("b" "o"). ())("" "" "Ύ"))
-    
+
     ((("b" "y" "a"). ())("Ӥ" "ӥ" "ˎގ"))
     ((("b" "y" "i"). ())("Ӥ" "ӥ" "ˎގ"))
     ((("b" "y" "u"). ())("Ӥ" "ӥ" "ˎގ"))
@@ -349,6 +306,7 @@
     ((("y" "e"). ())("" "" ""))
     ((("y" "o"). ())("" "" ""))
    
+    ((("x" "c" "a"). ())("" "" ""))
     ((("x" "k" "a"). ())("" "" ""))
     ((("x" "k" "e"). ())("" "" ""))
  
@@ -363,7 +321,16 @@
     ((("r" "u"). ())("" "" ""))
     ((("r" "e"). ())("" "" ""))
     ((("r" "o"). ())("" "" ""))
-    
+
+    ((("l" "t" "u"). ())("" "" ""))
+    ((("l" "t" "s" "u"). ())("" "" ""))
+
+    ((("l" "y" "a"). ())("" "" "؎"))
+    ((("l" "y" "i"). ())("ꤣ" "ꥣ" "؎"))
+    ((("l" "y" "u"). ())("" "" "؎"))
+    ((("l" "y" "e"). ())("ꤧ" "ꥧ" "؎"))
+    ((("l" "y" "o"). ())("" "" "؎"))
+
     ((("r" "y" "a"). ())("" "" "؎"))
     ((("r" "y" "i"). ())("ꤣ" "ꥣ" "؎"))
     ((("r" "y" "u"). ())("" "" "؎"))
@@ -372,15 +339,15 @@
 
     ((("w" "w"). ("w"))("" "" ""))
 
-    ((("x" "w" "a"). ())("" "" ""))
     ((("w" "a"). ())("" "" ""))
     ((("w" "i"). ())("" "" ""))
-    ((("x" "w" "i"). ())("" "" ""))
     ((("w" "u"). ())("" "" ""))
     ((("w" "e"). ())("" "" ""))
-    ((("x" "w" "e"). ())("" "" ""))
     ((("w" "o"). ())("" "" ""))
-    ((("w" "h" "a"). ())("" "" "")) 
+    ((("w" "h" "a"). ())("" "" ""))
+    ((("w" "h" "i"). ())("" "" ""))
+    ((("w" "h" "u"). ())("" "" ""))
+    ((("w" "h" "e"). ())("" "" "")) 
     ((("w" "h" "o"). ())("" "" "")) 
 
     ((("v" "v"). ("v"))("" "" ""))
@@ -395,18 +362,107 @@
     ((("v" "y" "u"). ())("" "" "ގ")) 
     ((("v" "y" "o"). ())("" "" "ގ")) 
 
-    ((("z" "k"). ())("" ""))
-    ((("z" "j"). ())("" ""))
-    ((("z" "h"). ())("" ""))
-    ((("z" "l"). ())("" ""))
-    ((("z" "-"). ())("" ""))
-    ((("z" "["). ())("" ""))
-    ((("z" "]"). ())("" ""))
-    ((("z" ","). ())("" ""))
-    ((("z" "."). ())("" ""))
-    ((("z" "/"). ())("" ""))
+    ((("z" "k"). ())("" "" ""))
+    ((("z" "j"). ())("" "" ""))
+    ((("z" "h"). ())("" "" ""))
+    ((("z" "l"). ())("" "" ""))
+    ((("z" "-"). ())("" "" ""))
+    ((("z" "["). ())("" "" ""))
+    ((("z" "]"). ())("" "" ""))
+    ((("z" ","). ())("" "" ""))
+    ((("z" "."). ())("" "" ""))
+    ((("z" "/"). ())("" "" ""))
     ))
 
+(define ja-rk-rule-additional
+  '(
+    ((("d" "s" "u"). ())("" "" ""))
+
+    ((("d" "h" "a"). ())("Ǥ" "ǥ" "Îގ"))
+    ((("d" "h" "i"). ())("Ǥ" "ǥ" "Îގ"))
+    ((("d" "h" "u"). ())("Ǥ" "ǥ" "Îގ"))
+    ((("d" "h" "e"). ())("Ǥ" "ǥ" "Îގ"))
+    ((("d" "h" "o"). ())("Ǥ" "ǥ" "Îގ"))
+
+    ((("d" "w" "a"). ())("ɤ" "ɥ" "Ďގ"))
+    ((("d" "w" "i"). ())("ɤ" "ɥ" "Ďގ"))
+    ((("d" "w" "u"). ())("ɤ" "ɥ" "Ďގ"))
+    ((("d" "w" "e"). ())("ɤ" "ɥ" "Ďގ"))
+    ((("d" "w" "o"). ())("ɤ" "ɥ" "Ďގ"))
+
+    ((("k" "w" "a"). ())("" "" ""))
+    ((("k" "w" "i"). ())("" "" ""))
+    ((("k" "w" "u"). ())("" "" ""))
+    ((("k" "w" "e"). ())("" "" ""))
+    ((("k" "w" "o"). ())("" "" ""))
+
+    ((("s" "h" "a"). ())("" "" ""))
+    ((("s" "h" "i"). ())("" "" ""))
+    ((("s" "h" "u"). ())("" "" ""))
+    ((("s" "h" "e"). ())("" "" ""))
+    ((("s" "h" "o"). ())("" "" ""))
+
+    ((("s" "w" "a"). ())("" "" ""))
+    ((("s" "w" "i"). ())("" "" ""))
+    ((("s" "w" "u"). ())("" "" ""))
+    ((("s" "w" "e"). ())("" "" ""))
+    ((("s" "w" "o"). ())("" "" "")) 
+
+    ((("t" "w" "a"). ())("Ȥ" "ȥ" "Ď"))
+    ((("t" "w" "i"). ())("Ȥ" "ȥ" "Ď"))
+    ((("t" "w" "u"). ())("Ȥ" "ȥ" "Ď"))
+    ((("t" "w" "e"). ())("Ȥ" "ȥ" "Ď"))
+    ((("t" "w" "o"). ())("Ȥ" "ȥ" "Ď"))
+
+    ((("t" "h" "a"). ())("Ƥ" "ƥ" "Î"))
+    ((("t" "h" "i"). ())("Ƥ" "ƥ" "Î"))
+    ((("t" "h" "u"). ())("Ƥ" "ƥ" "Î"))
+    ((("t" "h" "e"). ())("Ƥ" "ƥ" "Î"))
+    ((("t" "h" "o"). ())("Ƥ" "ƥ" "Î"))
+
+    ((("h" "w" "a"). ())("դ" "ե" "̎"))
+    ((("h" "w" "i"). ())("դ" "ե" "̎"))
+    ((("h" "w" "e"). ())("դ" "ե" "̎"))
+    ((("h" "w" "o"). ())("դ" "ե" "̎"))
+
+    ((("f" "w" "a"). ())("դ" "ե" "̎"))
+    ((("f" "w" "i"). ())("դ" "ե" "̎"))
+    ((("f" "w" "u"). ())("դ" "ե" "̎"))
+    ((("f" "w" "e"). ())("դ" "ե" "̎"))
+    ((("f" "w" "o"). ())("դ" "ե" "̎"))
+
+    ((("x" "w" "a"). ())("" "" ""))
+    ((("x" "w" "i"). ())("" "" ""))
+    ((("x" "w" "e"). ())("" "" ""))
+
+    ((("c" "h" "a"). ())("" "" ""))
+    ((("c" "h" "i"). ())("" "" ""))
+    ((("c" "h" "u"). ())("" "" ""))
+    ((("c" "h" "e"). ())("" "" ""))
+    ((("c" "h" "o"). ())("" "" ""))
+
+    ((("q" "w" "a"). ())("" "" ""))
+    ((("q" "w" "i"). ())("" "" ""))
+    ((("q" "w" "u"). ())("" "" ""))
+    ((("q" "w" "e"). ())("" "" ""))
+    ((("q" "w" "o"). ())("" "" ""))
+
+    ((("q" "y" "a"). ())("" "" ""))
+    ((("q" "y" "i"). ())("" "" ""))
+    ((("q" "y" "u"). ())("" "" ""))
+    ((("q" "y" "e"). ())("" "" ""))
+    ((("q" "y" "o"). ())("" "" ""))
+
+    ((("g" "w" "a"). ())("" "" "ގ"))
+    ((("g" "w" "i"). ())("" "" "ގ")) 
+    ((("g" "w" "u"). ())("" "" "ގ")) 
+    ((("g" "w" "e"). ())("" "" "ގ")) 
+    ((("g" "w" "o"). ())("" "" "ގ"))
+
+    ))
+
+(define ja-rk-rule (append ja-rk-rule-basic ja-rk-rule-additional))
+
 (define ja-wide-rule
   '(("a" "")
     ("b" "")
@@ -506,6 +562,46 @@
     ("_"  "")
     ))
 
+;;
+;; 2004-08-30 Takuro Ashie <ashie@homa.ne.jp>
+;;
+;;   It's a ad-hoc way to detect vowel and consonant in roma string.
+;;   FIXME!
+;;
+(define ja-vowel-table
+ '(("a" "a")
+   ("i" "i")
+   ("u" "u")
+   ("e" "e")
+   ("o" "o")
+    ))
+
+(define ja-consonant-table
+ '(("b" "b")
+   ("c" "c")
+   ("d" "d")
+   ("f" "f")
+   ("g" "g")
+   ("h" "h")
+   ("j" "j")
+   ("k" "k")
+   ("l" "l")
+   ("m" "m")
+   ("n" "n")
+   ("p" "p")
+   ("q" "q")
+   ("r" "r")
+   ("s" "s")
+   ("t" "t")
+   ("v" "v")
+   ("w" "w")
+   ("x" "x")
+   ("y" "y")
+   ("z" "z")
+   ))
+
+(define ja-default-small-tsu-roma "xtu")
+
 ;(define ja-rk-rule ja-kana-rule)
 
 ;; What's intended? The name "ja-direct-rule" makes me confused
@@ -536,6 +632,110 @@
   (lambda (c)
     (ja-find-rec c ja-direct-rule)))
 
+(define ja-vowel?
+  (lambda (c)
+    (ja-find-rec c ja-vowel-table)))
+
+(define ja-consonant?
+  (lambda (c)
+    (ja-find-rec c ja-consonant-table)))
+
+;;
+;; 2004-08-30 Takuro Ashie <ashie@homa.ne.jp>
+;;
+;; ja-string-list-to-wide-alphabet
+;;
+;;   Convert alphabets in string list to wide alphabets.
+;;   This procedure is ad-hoc. Maybe more generalize is needed.
+;;
+(define ja-string-list-to-wide-alphabet
+  (lambda (char-list)
+    (if (not (null? char-list))
+        (string-append (ja-string-list-to-wide-alphabet (cdr char-list))
+                       (ja-wide (car char-list)))
+        "")))
+
+;;
+;; 2004-08-30 Takuro Ashie <ashie@homa.ne.jp>
+;;
+;; ja-raw-string-to-valid-roma
+;;
+;;   Convert a invalid roma consonant to a valid roma consonant
+;;   or valid roma string.
+;;   Please see the comment of ja-raw-string-list-to-valid-roma
+;;   for more detail.
+;;
+(define ja-raw-string-to-valid-roma
+  (lambda (head-str next-str)
+    (if (ja-consonant? head-str)
+        (if (string=? head-str "n")
+            (if (null? next-str)
+                head-str
+                ;; 2004-09-01 Takuro Ashie <ashie@homa.ne.jp>
+                ;;   FIXME!
+                ;;   It's a ad-hoc way.
+                ;;   Should we check matched record from ja-rk-rule?:
+                ;;   (if (hoge-find-rec (string-append head-str next-str))
+                (let ((next-head (car (reverse (string-to-list next-str)))))
+                 (if (or (ja-vowel? next-str)
+                         (string=? "y" next-head)
+                         (string=? "n" next-head))
+                    "nn"
+                    head-str)))
+            (if (null? next-str)
+                ja-default-small-tsu-roma
+                (if (ja-vowel? next-str)
+                    ja-default-small-tsu-roma
+                    (car (reverse (string-to-list next-str))))))
+        head-str)))
+
+;;
+;; 2004-08-30 Takuro Ashie <ashie@homa.ne.jp>
+;;
+;; ja-raw-string-list-to-valid-roma
+;;
+;;   This procedure converts a invalid roma string-list to valid one.
+;;   "Invalid roma string-list" will be generated while editing a preedit string:
+;;
+;;     Convert a "n" which is followed by a vowel (or "ya" "yu"...) to "nn":
+;;       1. at first, type a following string:
+;;          ("ka" "n" "ki")
+;;       2: press backspace (or move the cursor):
+;;          ("ka" "n")
+;;       3. type a vowel:
+;;          ("ka" "n" "i")
+;;       4. On this case, this procedure converts the list to:
+;;          ("ka" "nn" "i")
+;;
+;;     Fix a broken "":
+;;       1.  at fisrt, type a following string:
+;;             ("a" "t" "ti")
+;;       2.  press backspace (or move the cursor):
+;;             ("a" "t")
+;;       3.  type remaining strings:
+;;             ("a" "t" "ka" "nn" "be" "-")
+;;      (3'. On this case, this procedure converts the list to:
+;;             ("a" "t") -> ("a" "ltu"))
+;;       4.  On this case, this procedure converts the list to:
+;;             ("a" "k" "ka" "nn" "be" "-")
+;; 
+(define ja-raw-string-list-to-valid-roma
+  (lambda (raw-str-list)
+    (let* ((head-str (car raw-str-list))
+           (next-str (cadr raw-str-list)))
+      (if (null? next-str)
+          (list (ja-raw-string-to-valid-roma head-str next-str))
+          (cons (ja-raw-string-to-valid-roma head-str next-str)
+                (ja-raw-string-list-to-valid-roma (cdr raw-str-list)))))))
+
+
+;;
+;; 2004-08-30 Takuro Ashie <ashie@homa.ne.jp>
+;;
+;;   In the future, these APIs should be marked as "deprecated" because these
+;;   APIs breaks original "ja-rk-rule" table, and no longer can use it after
+;;   breaks it.
+;;
 (define using-kana-table? #f)
 
 (define load-kana-table
@@ -551,7 +751,10 @@
 (define load-azik-table
   (lambda ()
     (require "japanese-azik.scm")
-    (set! ja-rk-rule (append ja-rk-rule ja-azik-rule))
+    (set! ja-rk-rule (append ja-azik-rule ja-rk-rule-basic))
+    (define-key skk-prev-candidate-key? 'generic-prev-candidate-key?)
+    (define-key skk-kana-toggle-key? "@")
+    (define-key skk-kanji-mode-key? "`")
     ))
 
 ;;
