@@ -30,6 +30,108 @@
 
 (require "util.scm")
 
+;; config
+(define enable-emacs-like-key-prefix? #t)
+
+;;
+(define valid-key-symbols
+  '(backspace
+    delete
+    escape
+    return
+    tab
+    left
+    up
+    right
+    down
+    prior
+    next
+    home
+    end
+    zenkaku-hankaku
+    Multi_key
+    Mode_switch
+    Henkan_Mode
+    Muhenkan
+    F1
+    F2
+    F3
+    F4
+    F5
+    F6
+    F7
+    F8
+    F9
+    F10
+    F11
+    F12
+    F13
+    F14
+    F15
+    F16
+    F17
+    F18
+    F19
+    F20
+    F21
+    F22
+    F23
+    F24
+    F25
+    F26
+    F27
+    F28
+    F29
+    F30
+    F31
+    F32
+    F33
+    F34
+    F35
+    Private1
+    Private2
+    Private3
+    Private4
+    Private5
+    Private6
+    Private7
+    Private8
+    Private9
+    Private10
+    Private11
+    Private12
+    Private13
+    Private14
+    Private15
+    Private16
+    Private17
+    Private18
+    Private19
+    Private20
+    Private21
+    Private22
+    Private23
+    Private24
+    Private25
+    Private26
+    Private27
+    Private28
+    Private29
+    Private30
+    Shift_key
+    Alt_key
+    Control_key
+    Meta_key
+    Super_key
+    Hyper_key))
+
+(define intern-key-symbol
+  (lambda (key-str)
+    (let ((sym (string->symbol key-str)))
+      (if (memq sym valid-key-symbols)
+	  sym
+	  #f))))
+
 ;
 (define shift-key-mask
   (lambda (state)
@@ -157,8 +259,9 @@
 (define parse-key-prefix
   (lambda (str)
     (let* ((parsed-as-emacs (parse-emacs-like-prefix str))
-	   (prefix (car parsed-as-emacs)))
-      (or (and prefix
+	   (emacs-prefix (car parsed-as-emacs)))
+      (or (and enable-emacs-like-key-prefix?
+	       emacs-prefix
 	       parsed-as-emacs)
 	  (parse-tag-prefix str)))))
 
@@ -202,9 +305,13 @@
 				      translators)))
 	      (parse-key-str rest translators key key-state)))
 	   (else
-	    (let ((key (string->symbol str))	; keysym
-		  (rest ""))
-	      (parse-key-str rest translators key key-state))))))))))
+	    (let* ((key-symbol (intern-key-symbol str))
+		   (key (or key-symbol
+			    key))
+		   (rest (if key-symbol
+			     ""
+			     rest)))
+	      (list rest translators key key-state))))))))))
 
 (define apply-translators
   (lambda (translators key key-state)
@@ -280,7 +387,7 @@
 ;; Generates or'ed key predicate and bind it into
 ;; toplevel-environment. Use define-key rather than calling this
 ;; directly.
-;; (define-key-internal 'foo-key?  '("<Control>j" "<Alt>k" "<Control>L"))
+;; (define-key-internal 'foo-key? '("<Control>j" "<Alt>k" "<Control>L"))
 (define define-key-internal
   (lambda (key-predicate-sym key-strs)
     (let* ((modified-key-strs (modify-key-strs-implicitly key-strs))
@@ -288,6 +395,36 @@
       (eval (list 'define key-predicate-sym predicate)
 	    toplevel-env))))
 
+(define valid-key-str?
+  (lambda (key-str)
+    (let* ((parsed (parse-key-str key-str () -1 0))
+	   (rest        (nth 0 parsed))
+	   (translators (nth 1 parsed))
+	   (key         (nth 2 parsed))
+	   (key-state   (nth 3 parsed)))
+      (and (string? key-str)
+	   (string=? rest "")
+	   (not (= key -1))))))
+
+;; 'strict-key-str' stands for key-str without translator-prefixes and
+;; emacs like prefix
+(define valid-strict-key-str?
+  (lambda (key-str)
+    (let ((saved-enable-eprefix? enable-emacs-like-key-prefix?)
+	  (res #f))
+      (set! enable-emacs-like-key-prefix? #f)
+      (let* ((parsed (parse-key-str key-str () -1 0))
+	     (rest        (nth 0 parsed))
+	     (translators (nth 1 parsed))
+	     (key         (nth 2 parsed))
+	     (key-state   (nth 3 parsed)))
+	(set! res (and (string? key-str)
+		       (string=? rest "")
+		       (null? translators)
+		       (not (= key -1)))))
+      (set! enable-emacs-like-key-prefix? saved-enable-eprefix?)
+      res)))
+
 ;(define set-key-binding
 ;  (lambda (im state flavor)
 ;    #f
