@@ -4,6 +4,8 @@
 (use file.util)
 (use test.unit)
 
+(add-load-path ".")
+
 (if (symbol-bound? 'main)
     (define _main main))
 
@@ -12,8 +14,8 @@
     (for-each (lambda (test-script)
                 (load (string-join (list dir test-script) "/")))
               (directory-list dir
-                              :filter (lambda (x) (rxmatch #/^test-/ x)))
-              )
+                              :filter (lambda (x)
+                                        (rxmatch #/^test-.+\.scm$/ x))))
     (if (symbol-bound? '_main)
         (_main `(,(car args) "-vp" ,@(cdr args)))
         (run-all-test))))
