@@ -59,9 +59,9 @@ struct anthy_segment_stat{
 static struct anthy_api {
   void *diclib, *lib;
 
-  int (*init)();
-  int (*quit)();
-  struct anthy_context *(*create_context)();
+  int (*init)(void);
+  int (*quit)(void);
+  struct anthy_context *(*create_context)(void);
   void (*release_context)(struct anthy_context *);
   void (*set_string)(struct anthy_context *, char *);
   void (*get_stat)(struct anthy_context *, struct anthy_conv_stat *);
@@ -77,7 +77,7 @@ static struct context {
 } *context_slot;
 
 static int
-get_anthy_api()
+get_anthy_api(void)
 {
   api.diclib = dlopen(LIBANTHYDIC_SO, RTLD_GLOBAL |RTLD_NOW);
   if (!api.diclib) {
@@ -90,22 +90,22 @@ get_anthy_api()
     dlclose(api.diclib);
     return -1;
   }
-  api.init = dlsym(api.lib, "anthy_init");
-  api.quit = dlsym(api.lib, "anthy_quit");
-  api.create_context = dlsym(api.lib, "anthy_create_context");
-  api.release_context = dlsym(api.lib, "anthy_release_context");
-  api.set_string = dlsym(api.lib, "anthy_set_string");
-  api.get_stat = dlsym(api.lib, "anthy_get_stat");
-  api.get_segment_stat = dlsym(api.lib, "anthy_get_segment_stat");
-  api.get_segment = dlsym(api.lib, "anthy_get_segment");
-  api.resize_segment = dlsym(api.lib, "anthy_resize_segment");
-  api.commit_segment = dlsym(api.lib, "anthy_commit_segment");
+  *(int  **)(&api.init) = dlsym(api.lib, "anthy_init");
+  *(int  **)(&api.quit) = dlsym(api.lib, "anthy_quit");
+  *(struct anthy_context **)(&api.create_context) = dlsym(api.lib, "anthy_create_context");
+  *(void **)(&api.release_context) = dlsym(api.lib, "anthy_release_context");
+  *(void **)(&api.set_string) = dlsym(api.lib, "anthy_set_string");
+  *(void **)(&api.get_stat) = dlsym(api.lib, "anthy_get_stat");
+  *(void **)(&api.get_segment_stat) = dlsym(api.lib, "anthy_get_segment_stat");
+  *(int  **)(&api.get_segment) = dlsym(api.lib, "anthy_get_segment");
+  *(void **)(&api.resize_segment) = dlsym(api.lib, "anthy_resize_segment");
+  *(void **)(&api.commit_segment) = dlsym(api.lib, "anthy_commit_segment");
   if (api.init &&
       api.quit &&
       api.create_context &&
       api.release_context && api.set_string &&
       api.get_stat && api.get_segment_stat &&
-      api.resize_segment && api.commit_segment) {
+      api.get_segment && api.resize_segment && api.commit_segment) {
     return 0;
   }
   return -1;
@@ -121,20 +121,23 @@ get_anthy_context(int id)
 }
 
 static LISP
-init_anthy_lib()
+init_anthy_lib(void)
 {
   int i;
   if (context_slot) {
     return siod_true_value();
   }
   if (get_anthy_api() == -1) {
-    return NIL;
+    return siod_false_value();
   }
   if (api.init() == -1) {
-    return NIL;
+    return siod_false_value();
   }
   context_slot = malloc(sizeof(struct context) *
 			MAX_CONTEXT);
+  if(!context_slot) {
+    return siod_false_value();
+  }
   for (i = 0; i < MAX_CONTEXT; i++) {
     context_slot[i].ac = NULL;
   }
@@ -142,57 +145,61 @@ init_anthy_lib()
 }
 
 static LISP
-create_context()
+create_context(void)
 {
   int i;
+  if (!context_slot) {
+    return siod_false_value();
+  }
+
   for (i = 0; i < MAX_CONTEXT; i++) {
     if (!context_slot[i].ac) {
       struct anthy_context *ac = api.create_context();
       if (!ac) {
-	return NIL;
+	return siod_false_value();
       }
       context_slot[i].ac = ac;
       return intcons(i);
     }
   }
-  return NIL;
+  return siod_false_value();
 }
 
 
 static LISP
 release_context(LISP id_)
 {
-  int id = get_c_long(id_);
+  int id = get_c_int(id_);
   if (context_slot[id].ac) {
     api.release_context(context_slot[id].ac);
     context_slot[id].ac = NULL;
   }
-  return NIL;
+  return siod_false_value();
 }
 
 static LISP
 set_string(LISP id_, LISP str_)
 {
-  int id = get_c_long(id_);
+  int id = get_c_int(id_);
   char *str;
   struct anthy_context *ac = get_anthy_context(id);
   if (!ac) {
-    return NIL;
+    return siod_false_value();
   }
   str = uim_get_c_string(str_);
   api.set_string(ac, str);
   free(str);
-  return NIL;
+  return siod_false_value();
 }
 
 static LISP
 get_nr_segments(LISP id_)
 {
-  int id = get_c_long(id_);
+  int id = get_c_int(id_);
   struct anthy_conv_stat acs;
   struct anthy_context *ac = get_anthy_context(id);
   if (!ac) {
-    return NIL;
+    return siod_false_value();
   }
   api.get_stat(ac, &acs);
 
@@ -205,11 +212,11 @@ get_nr_candidates(LISP id_, LISP nth_)
   int id, nth;
   struct anthy_context *ac;
   struct anthy_conv_stat cs;
-  id = get_c_long(id_);
-  nth = get_c_long(nth_);
+  id = get_c_int(id_);
+  nth = get_c_int(nth_);
   ac = get_anthy_context(id);
   if (!ac) {
-    return NIL;
+    return siod_false_value();
   }
   api.get_stat(ac, &cs);
   if (nth < cs.nr_segment) {
@@ -217,55 +224,57 @@ get_nr_candidates(LISP id_, LISP nth_)
     api.get_segment_stat(ac, nth, &ss);
     return intcons(ss.nr_candidate);
   }
-  return NIL;
+  return siod_false_value();
 }
 
 static LISP
 get_nth_candidate(LISP id_, LISP seg_, LISP nth_)
 {
-  int id = get_c_long(id_);
-  int seg = get_c_long(seg_);
-  int nth  = get_c_long(nth_);
+  int id = get_c_int(id_);
+  int seg = get_c_int(seg_);
+  int nth  = get_c_int(nth_);
   int buflen;
   char *buf;
+  LISP buf_;
   struct anthy_context *ac = get_anthy_context(id);
-  struct anthy_conv_stat cs;
   if (!ac) {
-    return NIL;
+    return siod_false_value();
   }
   buflen = api.get_segment(ac, seg, nth, NULL, 0);
   if (buflen == -1) {
-    return NIL;
+    return siod_false_value();
   }
-  buf = alloca(buflen+1);
+  buf = malloc(buflen+1);
   api.get_segment(ac, seg, nth, buf, buflen+1);
-  return strcons(buflen, buf);
+  buf_ = strcons(buflen, buf);
+  free(buf);
+  return buf_;
 }
 
 static LISP
 resize_segment(LISP id_, LISP seg_, LISP cnt_)
 {
-  int id = get_c_long(id_);
-  int seg = get_c_long(seg_);
-  int cnt = get_c_long(cnt_);
+  int id = get_c_int(id_);
+  int seg = get_c_int(seg_);
+  int cnt = get_c_int(cnt_);
   struct anthy_context *ac = get_anthy_context(id);
   api.resize_segment(ac, seg, cnt);
-  return NIL;
+  return siod_false_value();
 }
 
 static LISP
 commit_segment(LISP id_, LISP s_, LISP nth_)
 {
-  int id = get_c_long(id_);
-  int s = get_c_long(s_);
-  int nth = get_c_long(nth_);
+  int id = get_c_int(id_);
+  int s = get_c_int(s_);
+  int nth = get_c_int(nth_);
   struct anthy_context *ac = get_anthy_context(id);
   api.commit_segment(ac, s, nth);
-  return NIL;
+  return siod_false_value();
 }
 
 void
-uim_init_anthy()
+uim_init_anthy(void)
 {
   init_subr_0("anthy-lib-init", init_anthy_lib);
 
@@ -281,15 +290,17 @@ uim_init_anthy()
 }
 
 void
-uim_quit_anthy()
+uim_quit_anthy(void)
 {
   int i;
 
-  if (context_slot) {
-    for (i = 0; i < MAX_CONTEXT; i++) {
-      if(context_slot[i].ac) {
-	api.release_context(context_slot[i].ac);
-      }
+  if (!context_slot) {
+    return;
+  }
+
+  for (i = 0; i < MAX_CONTEXT; i++) {
+    if(context_slot[i].ac) {
+      api.release_context(context_slot[i].ac);
     }
   }
 
