@@ -1,4 +1,6 @@
 /*
+  $Id$
+
   canna.c: Canna for uim.
 
   Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
@@ -75,17 +77,17 @@ static struct canna_api {
   int (*RkBgnBun)(int, char *, int, int);
   int (*RkEndBun)(int, int);
   int (*RkFinalize)(void);
-  int (*RkGetDicList)(int, char*, int);
-  int (*RkMountDic)(int, char*, int);
+  int (*RkGetDicList)(int, char *, int);
+  int (*RkMountDic)(int, char *, int);
   int (*RkNfer)(int);
   int (*RkXfer)(int, int);
   int (*RkEnlarge)(int);
   int (*RkShorten)(int);
-  int (*RkGetStat)(int, RkStat*);
+  int (*RkGetStat)(int, RkStat *);
   int (*RkGoTo)(int, int);
-  int (*RkGetKanji)(int, char*, int);
-  int (*RkGetKanjiList)(int, char*, int);
-  int (*RkSetServerName)(char*);
+  int (*RkGetKanji)(int, char *, int);
+  int (*RkGetKanjiList)(int, char *, int);
+  int (*RkSetServerName)(char *);
 } api;
 
 static int
@@ -95,22 +97,22 @@ get_canna_api(void)
   if(!api.lib)
     return -1;
 
-  api.RkInitialize = dlsym(api.lib, "RkInitialize");
-  api.RkFinalize = dlsym(api.lib, "RkFinalize");
-  api.RkGetDicList = dlsym(api.lib, "RkGetDicList");
-  api.RkMountDic = dlsym(api.lib, "RkMountDic");
-  api.RkCreateContext = dlsym(api.lib, "RkCreateContext");
-  api.RkCloseContext = dlsym(api.lib, "RkCloseContext");
-  api.RkBgnBun = dlsym(api.lib, "RkBgnBun");
-  api.RkEndBun = dlsym(api.lib, "RkEndBun");
-  api.RkGoTo = dlsym(api.lib, "RkGoTo");
-  api.RkGetKanji = dlsym(api.lib, "RkGetKanji");
-  api.RkGetKanjiList = dlsym(api.lib, "RkGetKanjiList");
-  api.RkGetStat = dlsym(api.lib, "RkGetStat");
-  api.RkXfer = dlsym(api.lib, "RkXfer");
-  api.RkNfer = dlsym(api.lib, "RkXfer");
-  api.RkEnlarge = dlsym(api.lib, "RkEnlarge");
-  api.RkShorten = dlsym(api.lib, "RkShorten");
+  *(int **)(&api.RkInitialize) = dlsym(api.lib, "RkInitialize");
+  *(int **)(&api.RkFinalize) = dlsym(api.lib, "RkFinalize");
+  *(int **)(&api.RkGetDicList) = dlsym(api.lib, "RkGetDicList");
+  *(int **)(&api.RkMountDic) = dlsym(api.lib, "RkMountDic");
+  *(int **)(&api.RkCreateContext) = dlsym(api.lib, "RkCreateContext");
+  *(int **)(&api.RkCloseContext) = dlsym(api.lib, "RkCloseContext");
+  *(int **)(&api.RkBgnBun) = dlsym(api.lib, "RkBgnBun");
+  *(int **)(&api.RkEndBun) = dlsym(api.lib, "RkEndBun");
+  *(int **)(&api.RkGoTo) = dlsym(api.lib, "RkGoTo");
+  *(int **)(&api.RkGetKanji) = dlsym(api.lib, "RkGetKanji");
+  *(int **)(&api.RkGetKanjiList) = dlsym(api.lib, "RkGetKanjiList");
+  *(int **)(&api.RkGetStat) = dlsym(api.lib, "RkGetStat");
+  *(int **)(&api.RkXfer) = dlsym(api.lib, "RkXfer");
+  *(int **)(&api.RkNfer) = dlsym(api.lib, "RkXfer");
+  *(int **)(&api.RkEnlarge) = dlsym(api.lib, "RkEnlarge");
+  *(int **)(&api.RkShorten) = dlsym(api.lib, "RkShorten");
   if(api.RkInitialize && api.RkFinalize && api.RkGetDicList &&
      api.RkMountDic && api.RkCreateContext && api.RkCloseContext &&
      api.RkBgnBun && api.RkEndBun && api.RkGoTo && api.RkGetKanji &&
@@ -151,7 +153,7 @@ init_canna_lib(LISP str_)
     return NIL;
 
   if(str_ != NIL)
-    cannaserver = uim_get_c_string(str_);
+    cannaserver = uim_scm_c_str(str_);
   else
     cannaserver = NULL;
 
@@ -224,7 +226,7 @@ create_context() {
 static LISP
 release_context(LISP id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   struct canna_context *cc = get_canna_context(id);
 
   if(cc == NULL)
@@ -306,7 +308,7 @@ _update_segment (const int id, const int segment_num)
 static LISP
 begin_conversion(LISP id_, LISP str_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   char *str;
   int len, segment_num, mode;
   struct canna_context *cc = get_canna_context(id);
@@ -318,7 +320,7 @@ begin_conversion(LISP id_, LISP str_)
     return NIL;
 
   mode = cc->rk_mode;
-  str = uim_get_c_string(str_);
+  str = uim_scm_c_str(str_);
   len = strlen(str);
 
   segment_num = api.RkBgnBun(cc->rk_context_id, str, len, mode);
@@ -342,9 +344,9 @@ begin_conversion(LISP id_, LISP str_)
 static LISP
 get_nth_candidate(LISP id_, LISP seg_, LISP nth_)
 {
-  int id = get_c_int(id_);
-  int seg = get_c_int(seg_);
-  int nth = get_c_int(nth_);
+  int id = uim_scm_c_int(id_);
+  int seg = uim_scm_c_int(seg_);
+  int nth = uim_scm_c_int(nth_);
   struct canna_context *cc = get_canna_context(id);
   char buf[BUFSIZE];
   int len;
@@ -366,7 +368,7 @@ get_nth_candidate(LISP id_, LISP seg_, LISP nth_)
 static LISP
 get_nr_segments(LISP id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
   RkStat stat;
   struct canna_context *cc = get_canna_context(id);
 
@@ -382,8 +384,8 @@ get_nr_segments(LISP id_)
 static LISP
 get_nr_candidate(LISP id_, LISP nth_)
 {
-  int id = get_c_int(id_);
-  int nth = get_c_int(nth_);
+  int id = uim_scm_c_int(id_);
+  int nth = uim_scm_c_int(nth_);
   RkStat stat;
   struct canna_context *cc = get_canna_context(id);
 
@@ -404,9 +406,9 @@ get_nr_candidate(LISP id_, LISP nth_)
 static LISP
 resize_segment(LISP id_, LISP s_, LISP nth_)
 {
-  int id = get_c_int(id_);
-  int s = get_c_int(s_);
-  int nth = get_c_int(nth_);
+  int id = uim_scm_c_int(id_);
+  int s = uim_scm_c_int(s_);
+  int nth = uim_scm_c_int(nth_);
   struct canna_context *cc = get_canna_context(id);
 
   if(cc == NULL)
@@ -431,9 +433,9 @@ resize_segment(LISP id_, LISP s_, LISP nth_)
 static LISP
 commit_segment(LISP id_, LISP s_, LISP nth_)
 {
-  int id = get_c_int(id_);
-  int s = get_c_int(s_);
-  int nth = get_c_int(nth_);
+  int id = uim_scm_c_int(id_);
+  int s = uim_scm_c_int(s_);
+  int nth = uim_scm_c_int(nth_);
   struct canna_context *cc = get_canna_context(id);
 
   if(cc == NULL)
@@ -449,7 +451,7 @@ commit_segment(LISP id_, LISP s_, LISP nth_)
 static LISP
 reset_conversion(LISP id_)
 {
-  int id = get_c_int(id_);
+  int id = uim_scm_c_int(id_);
 
   _reset_conversion(id);
 }
@@ -479,8 +481,10 @@ uim_quit_canna(void)
      cannaserver = NULL;
   }
 
-  if(api.RkFinalize)
+  if(api.RkFinalize && rk_initialized == 1) {
      api.RkFinalize();
+     rk_initialized = -1;
+  }
 
   if(api.lib) {
      dlclose(api.lib);
