@@ -32,15 +32,11 @@
 
 #include <signal.h>
 #include <stdlib.h>
-#include <sys/types.h>
-#include <unistd.h>
-#include <fcntl.h>
 #include <errno.h>
+#include <string.h>
 
 #include "context.h"
 
-char * uim_ipc_send_command(int *, FILE **, FILE **, char*, char *);
-
 static FILE *primer = NULL, *primew = NULL;
 static int prime_pid = 0;
 
@@ -89,8 +85,11 @@ uim_quit_prime(void)
   if(primew) {
     uim_ipc_send_command(&prime_pid, &primer, &primew, prime_command, "close\n");
     fclose(primew);
+    primew = NULL;
   }
 
-  if(primer)
+  if(primer) {
     fclose(primer);
+    primer = NULL;
+  }
 }
