@@ -53,57 +53,87 @@
  */
 
 
-/*
+/* candidate array for each okurigana
+ *
  * |C0|C1| .. |Cnr_real_cands| ..              |Cnr_cands|
  * <-------should be saved --><-- cache of master dict -->
  */
 struct skk_cand_array {
+  /* okurigana string */
   char *okuri;
+
   int nr_cands;/* length of cands array allocated */
   int nr_real_cands;/* length of read from file part */
+  /* candidate string */
   char **cands;
 
+  /* this array was used and merged with okuri-nasi entry array */
   int is_used;
+  /* link to next entry in the list */
   struct skk_line *line;
 };
 
+/* skk dictionary line */
 struct skk_line {
+  /* line index. head part */
   char *head;
+  /* line index. okurigana part. value will be 0,
+     if it is okuri-nasi entry */
   char okuri_head;
-  /**/
+  /* array of candidate array for different okuri-gana */
   int nr_cand_array;
   struct skk_cand_array *cands;
-  /**/
+  /* modified or read from file */
   int need_save;
-  /**/
+  /* link to next entry in the list */
   struct skk_line *next;
 };
 
+/* skk dictionary file */
 static struct dic_info {
+  /* address of mmap'ed dictionary file */
   void *addr;
+  /* byte offset of first valid entry in mmap'ed region */
   int first;
+  /* byte offset of first okuri-nasi entry */
   int border;
+  /* size of dictionary file */
   int size;
+  /* head of cached skk dictionary line list. LRU ordered */
   struct skk_line head;
 } *skk_dic;
 
+/* completion */
+struct skk_comp_array {
+  /* index of completion */
+  char *head;
+  /* array of completion string */
+  int nr_comps;
+  char **comps;
+  /**/
+  int refcount;
+  /**/
+  struct skk_comp_array *next;
+} *skk_comp;
+
 static int
-calc_line_len(char *s)
+calc_line_len(const char *s)
 {
   int i;
   for (i = 0; s[i] != '\n'; i++);
   return i;
 }
 
-
 static int
-is_okuri(char *str)
+is_okuri(const char *line_str)
 {
-  char *b;
-  b = strchr(str, ' ');
+  const char *b;
+  /* find first white space */
+  b = strchr(line_str, ' ');
   if (!b) {
     return 0;
   }
+  /* check previous character */
   b--;
   if (isalpha(*b)) {
     return 1;
@@ -125,11 +155,11 @@ find_first_line(struct dic_info *di)
 }
 
 static int
-find_border(struct dic_info *di)
+find_border(struct dic_info *di, int size)
 {
   char *s = di->addr;
   int off = 0;
-  while (1) {
+  while (off < size) {
     int l = calc_line_len(&s[off]);
     if (s[off] == ';') {
       off += l + 1;
@@ -140,7 +170,8 @@ find_border(struct dic_info *di)
     }
     off += l + 1;
   }
-  return 0;
+  /* every entry is okuri-ari, it may not happen. */
+  return size - 1;
 }
 
 static struct dic_info *
@@ -166,7 +197,7 @@ open_dic(const char *fn)
   di->addr = addr;
   di->size = st.st_size;
   di->first = find_first_line(di);
-  di->border = find_border(di);
+  di->border = find_border(di, st.st_size);
   di->head.next = NULL;
   return di;
 }
@@ -185,7 +216,7 @@ find_line(struct dic_info *di, int off)
 }
 
 static char *
-extract_entry(struct dic_info *di, int off, char *buf, int len)
+extract_line_index(struct dic_info *di, int off, char *buf, int len)
 {
   char *p = find_line(di, off);
   int i;
@@ -200,18 +231,18 @@ extract_entry(struct dic_info *di, int off, char *buf, int len)
 }
 
 static int
-do_search(struct dic_info *di, char *s, int min,
-	  int max, int d)
+do_search_line(struct dic_info *di, char *s, int min,
+	       int max, int d)
 {
   char buf[256];
   char *r;
-  int idx = (min + max) / 2;
+  int idx = (min + max)/2;
   int c = 0;
 
   if (abs(max-min) < 4) {
     return -1;
   }
-  r = extract_entry(di, idx, buf, 256);
+  r = extract_line_index(di, idx, buf, 256);
   if (r) {
     c = strcmp(s, r);
   } else {
@@ -222,13 +253,23 @@ do_search(struct dic_info *di, char *s, int min,
     return idx;
   }
   if (c * d> 0) {
-    return do_search(di, s, idx, max, d);
+    return do_search_line(di, s, idx, max, d);
   } else {
-    return do_search(di, s, min, idx, d);
+    return do_search_line(di, s, min, idx, d);
   }
   return -1;
 }
 
+/* This funciton name is temporary. I want a better name. */
+static char *
+first_space(char *str)
+{
+  while (*str && (*str != ' ')) {
+    str++;
+  }
+  return str;
+}
+
 static char *
 next_slash(char *str)
 {
@@ -250,6 +291,9 @@ nth_candidate(char *str, int nth)
 {
   char *p , *term;
   int i;
+
+  str = first_space(str);
+  
   for (i = 0; i <= nth; i++) {
     str = next_slash(str);
     if (*str == '/') {
@@ -366,7 +410,7 @@ static void
 compose_line_parts(struct dic_info *di, struct skk_line *sl,
 		   char *okuri, char *line)
 {
-  int i, nth;
+  int nth;
   char *tmp;
   struct skk_cand_array *ca = find_candidate_array_from_line(sl, okuri, 1);
 
@@ -424,12 +468,46 @@ compose_line(struct dic_info *di, char *word, char okuri_head, char *entry)
 }
 
 static void
-add_line_to_cache(struct dic_info *di, struct skk_line *sl)
+add_line_to_cache_head(struct dic_info *di, struct skk_line *sl)
 {
   sl->next = di->head.next;
   di->head.next = sl;
 }
 
+static void
+move_line_to_cache_head(struct dic_info *di, struct skk_line *sl)
+{
+  struct skk_line *prev;
+
+  if (di->head.next == sl)
+    return;
+
+  prev = di->head.next;
+  while (prev->next != sl) {
+    prev = prev->next;
+  }
+  prev->next = sl->next;
+  sl->next = di->head.next;
+  di->head.next = sl;
+}
+
+static void
+add_line_to_cache_last(struct dic_info *di, struct skk_line *sl)
+{
+  struct skk_line *prev;
+
+  if (di->head.next == NULL)
+    di->head.next = sl;
+  else {
+    prev = di->head.next;
+    while (prev->next) {
+      prev = prev->next;
+    }	
+    prev->next = sl;
+  }
+  sl->next = NULL;
+}
+
 static struct skk_line *
 skk_search_line_from_file(struct dic_info *di, char *s, char okuri_head)
 {
@@ -445,9 +523,9 @@ skk_search_line_from_file(struct dic_info *di, char *s, char okuri_head)
   }
   sprintf(idx, "%s%c",s, okuri_head);
   if (okuri_head) {
-    n = do_search(di, idx, di->first, di->border - 1, -1);
+    n = do_search_line(di, idx, di->first, di->border - 1, -1);
   } else {
-    n = do_search(di, idx, di->border, di->size - 1, 1);
+    n = do_search_line(di, idx, di->border, di->size - 1, 1);
   }
   if (n == -1) {
     return NULL;
@@ -502,7 +580,7 @@ find_cand_array(struct dic_info *di, char *s,
       sl = alloc_skk_line(s, okuri_head);
     }
     from_file = 1;
-    add_line_to_cache(di, sl);
+    add_line_to_cache_head(di, sl);
   }
 
   ca = find_candidate_array_from_line(sl, okuri, create_if_not_found);
@@ -525,7 +603,6 @@ static struct skk_cand_array *
 find_cand_array_lisp(LISP head_, LISP okuri_head_, LISP okuri_,
 		     int create_if_not_found)
 {
-  int n;
   char o;
   char *hs;
   char *okuri = NULL;
@@ -568,7 +645,7 @@ skk_get_nth_candidate(LISP nth_, LISP head_,
   char *str;
 
   ca = find_cand_array_lisp(head_, okuri_head_, okuri_, 0);
-  n = get_c_long(nth_);
+  n = get_c_int(nth_);
   if (ca && ca->nr_cands > n) {
     str = ca->cands[n];
     return strcons(strlen(str), str);
@@ -590,6 +667,153 @@ skk_get_nr_candidates(LISP head_, LISP okuri_head_, LISP okuri_)
   return intcons(n);
 }
 
+static struct skk_comp_array *
+skk_make_comp_array_from_cache(struct dic_info *di, char *s)
+{
+  struct skk_line *sl;
+  struct skk_comp_array *ca;
+
+  if (!di) {
+    return NULL;
+  }
+  ca = malloc(sizeof(struct skk_comp_array));
+  ca->nr_comps = 0;
+  ca->refcount = 0;
+  ca->comps = NULL;
+  ca->head = NULL;
+  ca->next = NULL;
+
+  /* search from cache */
+  for (sl = di->head.next; sl; sl = sl->next) {
+    if (/* string 's' is part of sl->head */
+	!strncmp(sl->head, s, strlen(s)) && strcmp(sl->head, s) &&
+	/* and sl is okuri-nasi line */
+        (sl->okuri_head == '\0')) {
+      ca->nr_comps++;
+      ca->comps = realloc(ca->comps, sizeof(char *) * ca->nr_comps);
+      ca->comps[ca->nr_comps - 1] = strdup(sl->head);
+    }
+  }
+
+  if (ca->nr_comps == 0) {
+    free(ca);
+    ca = NULL;
+  } else {
+    ca->head = strdup(s);
+    ca->next = skk_comp;
+    skk_comp = ca;
+  }
+  return ca;
+}
+
+static struct skk_comp_array *
+find_comp_array(struct dic_info *di, char *s)
+{
+  struct skk_comp_array *ca;
+
+  if (strlen(s) == 0)
+    return NULL;
+
+  for (ca = skk_comp; ca; ca = ca->next) {
+    if (!strcmp(ca->head, s))
+      break;
+  }
+  if (ca == NULL) {
+    ca = skk_make_comp_array_from_cache(skk_dic, s);
+  }
+
+  return ca;
+}
+
+static struct skk_comp_array *
+find_comp_array_lisp(LISP head_)
+{
+  char *hs;
+  struct skk_comp_array *ca;
+  
+  hs = get_c_string(head_);
+  ca = find_comp_array(skk_dic, hs);
+  return ca;
+}
+
+static LISP
+skk_get_completion(LISP head_)
+{
+  struct skk_comp_array *ca;
+  ca = find_comp_array_lisp(head_);
+  if (ca) {
+    ca->refcount++;
+    return siod_true_value();
+  }
+  return NIL;
+}
+
+static LISP
+skk_get_nth_completion(LISP nth_, LISP head_)
+{
+  int n;
+  struct skk_comp_array *ca;
+  char *str;
+
+  ca = find_comp_array_lisp(head_);
+  n = get_c_int(nth_);
+  if (ca && ca->nr_comps > n) {
+    str = ca->comps[n];
+    return strcons(strlen(str), str);
+  }
+  return NIL;
+}
+
+static LISP
+skk_get_nr_completions(LISP head_)
+{
+  int n = 0;
+  struct skk_comp_array *ca;
+
+  ca = find_comp_array_lisp(head_);
+  if (ca) {
+    n = ca->nr_comps;
+  }
+  return intcons(n);
+}
+
+static LISP
+skk_clear_completions(LISP head_)
+{
+  int i;
+  struct skk_comp_array *ca, *ca_prev;
+  char *hs;
+
+  hs = get_c_string(head_);
+  for (ca = skk_comp; ca; ca = ca->next) {
+    if (!strcmp(ca->head, hs)) {
+      ca->refcount--;
+      break;
+    }
+  }
+
+  if (ca && ca->refcount == 0) {
+    for (i = 0; i < ca->nr_comps; i++) {
+      free(ca->comps[i]);
+    }
+    free(ca->comps);
+    free(ca->head);
+
+    if (ca == skk_comp) {
+      skk_comp = ca->next;
+      free(ca);
+    } else {
+      ca_prev = skk_comp;
+      while (ca_prev->next != ca) {
+        ca_prev = ca_prev->next;
+      }
+      ca_prev->next = ca->next;
+      free(ca);
+    }
+  }
+  return siod_true_value();
+}
+
 static void
 reorder_candidate(struct skk_cand_array *ca, char *str)
 {
@@ -625,8 +849,10 @@ skk_commit_candidate(LISP head_, LISP okuri_head_,
   struct skk_cand_array *ca;
   char *str;
 
-  nth = get_c_long(nth_);
+  nth = get_c_int(nth_);
   if (nth == 0) {
+    ca = find_cand_array_lisp(head_, okuri_head_, okuri_, 0);
+    move_line_to_cache_head(skk_dic, ca->line);
     return NIL;
   }
 
@@ -647,6 +873,7 @@ skk_commit_candidate(LISP head_, LISP okuri_head_,
   }
 
   ca->line->need_save = 1;
+  move_line_to_cache_head(skk_dic, ca->line);
 
   return NIL;
 }
@@ -692,7 +919,7 @@ skk_learn_word(LISP head_, LISP okuri_head_, LISP okuri_, LISP word_)
   word = sanitize_word(tmp);
   free(tmp);
   if (!word) {
-    return ;
+    return NIL;
   }
 
   ca = find_cand_array_lisp(head_, okuri_head_, okuri_, 1);
@@ -732,7 +959,7 @@ parse_dic_line(char *line)
     return ;
   }
   *sep = 0;
-  if (islower(sep[-1])) {
+  if (!islower(buf[0]) && islower(sep[-1])) { /* okuri-ari entry */
     char okuri_head = sep[-1];
     sep[-1] = 0;
     sl = compose_line(skk_dic, buf, okuri_head, line);
@@ -745,7 +972,7 @@ parse_dic_line(char *line)
     sl->cands[i].nr_real_cands = sl->cands[i].nr_cands;
   }
   /**/
-  add_line_to_cache(skk_dic, sl);
+  add_line_to_cache_last(skk_dic, sl);
 }
 
 static void
@@ -790,6 +1017,10 @@ skk_lib_save_personal_dictionary(LISP fn_)
   char *fn = uim_get_c_string(fn_);
   struct skk_line *sl;
 
+  if (!skk_dic) {
+    return NIL;
+  }
+
   if (fn) {
     fp = fopen(fn, "w");
     free(fn);
@@ -866,6 +1097,10 @@ uim_init_skk_dic(void)
   init_subr_4("skk-lib-commit-candidate", skk_commit_candidate);
   init_subr_4("skk-lib-learn-word", skk_learn_word);
   init_subr_1("skk-lib-remove-annotation", skk_lib_remove_annotation);
+  init_subr_1("skk-lib-get-completion", skk_get_completion);
+  init_subr_2("skk-lib-get-nth-completion", skk_get_nth_completion);
+  init_subr_1("skk-lib-get-nr-completions", skk_get_nr_completions);
+  init_subr_1("skk-lib-clear-completions", skk_clear_completions);
 }
 
 void
@@ -883,4 +1118,5 @@ uim_quit_skk_dic(void)
     free_skk_line(tmp);
   }
   free(skk_dic);
+  skk_dic = NULL;
 }
