@@ -74,6 +74,11 @@
   removed copygc (Nov-03) Yusuke TABATA
   removed many unneeded functionality (03-04) Yusuke TABATA
   removed sliba.c (Feb-04) Yusuke TABATA
+  added second arg "LISP env" to undefine() (Jul-04-2004) YamaKen
+  added 'case' special form (Sep-09-2004) Jun Inoue
+  added 'else' symbol definition (Sep-21-2004) YamaKen
+  fix broken feature? and provide (Sep-28-2004) YamaKen
+  removed non-standard _"str" syntax for i18n (Sep-30-2004) YamaKen
  */
 #include <stdio.h>
 #include <stdarg.h>
@@ -90,56 +95,55 @@
 #endif
 
 #include "siod.h"
+#include "intl.h"
 
 /* strcut */
 
-struct catch_frame
-  {
-    LISP tag;
-    LISP retval;
-    jmp_buf cframe;
-    struct catch_frame *next;
-  };
+struct catch_frame {
+  LISP tag;
+  LISP retval;
+  jmp_buf cframe;
+  struct catch_frame *next;
+};
 
 
-struct gen_readio
-  {
-    int (*getc_fcn) (void *);
-    void (*ungetc_fcn) (int, void *);
-    void *cb_argument;
-  };
+struct gen_readio {
+  int (*getc_fcn) (void *);
+  void (*ungetc_fcn) (int, void *);
+  void *cb_argument;
+};
 
-struct gen_printio
-  {
-    int (*putc_fcn) (int, void *);
-    int (*puts_fcn) (char *, void *);
-    void *cb_argument;
-  };
+struct gen_printio {
+  int (*putc_fcn) (int, void *);
+  int (*puts_fcn) (char *, void *);
+  void *cb_argument;
+};
 
-struct user_type_hooks
-  {
-      LISP (*gc_mark) (LISP);
-    void (*gc_free) (LISP);
-    void (*prin1) (LISP, struct gen_printio *);
-      LISP (*leval) (LISP, LISP *, LISP *);
-      LISP (*equal) (LISP, LISP);
-  };
+struct user_type_hooks {
+  LISP (*gc_mark) (LISP);
+  void (*gc_free) (LISP);
+  void (*prin1) (LISP, struct gen_printio *);
+  LISP (*leval) (LISP, LISP *, LISP *);
+  LISP (*equal) (LISP, LISP);
+};
 
-struct repl_hooks
-  {
-    void (*repl_puts) (char *);
-      LISP (*repl_read) (void);
-      LISP (*repl_eval) (LISP);
-    void (*repl_print) (LISP);
-  };
+struct repl_hooks {
+  void (*repl_puts) (char *);
+  LISP (*repl_read) (void);
+  LISP (*repl_eval) (LISP);
+  void (*repl_print) (LISP);
+};
 
+struct gc_protected {
+  LISP *location;
+  long length;
+  struct gc_protected *next;
+};
 
-struct gc_protected
-  {
-    LISP *location;
-    long length;
-    struct gc_protected *next;
-  };
+struct func_frame {
+  struct func_frame *prev;
+  LISP obj;
+};
 
 /* forward declaration of static functions */
 static void gc_for_newcell (void);
@@ -158,7 +162,7 @@ static void gc_for_newcell (void);
 #define STACK_CHECK(_ptr) \
   if (((char *) (_ptr)) < stack_limit_ptr) err_stack((char *) _ptr);
 
-#define NEWCELL(_into,_type)          \
+#define _NEWCELL(_into, _type)        \
 {  if NULLP(freelist)                 \
       gc_for_newcell();               \
     _into = freelist;                 \
@@ -167,9 +171,16 @@ static void gc_for_newcell (void);
  (*_into).gc_mark = 0;                \
  (*_into).type = (short) _type;}
 
-#ifndef SIOD_LIB_DEFAULT
-#define SIOD_LIB_DEFAULT "/usr/local/lib/siod"
-#endif
+#if DEBUG_SCM
+#define NEWCELL(_into,_type)          \
+{  _NEWCELL (_into, _type);           \
+   (*_into).dbg_info = car (dbg_pos);}
+#else  /* not DEBUG_SCM */
+#define NEWCELL(_into, _type) _NEWCELL (_into, _type)
+#define dbg_readini(f)
+#define dbg_readend()
+#define dbg_readabrt()
+#endif /* DEBUG_SCM */
 
 /* exported global symbol */
 long siod_verbose_level;
@@ -195,17 +206,16 @@ static LISP *stack_start_ptr;
 static LISP freelist;
 static jmp_buf errjmp;
 static long errjmp_ok;
-static long nointerrupt;
 static LISP oblistvar;
 static LISP eof_val;
 static LISP sym_errobj;
 static LISP sym_catchall;
 static LISP sym_progn;
 static LISP sym_lambda;
+static LISP sym_else;
 static LISP sym_quote;
 static LISP sym_dot;
 static LISP sym_after_gc;
-static LISP sym_eval_history_ptr;
 static LISP sym_features;
 static LISP unbound_marker;
 static LISP *obarray;
@@ -219,7 +229,6 @@ static void (*repl_print) (LISP);
 static LISP *inums;
 static long inums_dim;
 static struct user_type_hooks *user_types;
-static long user_tc_next;
 static struct gc_protected *protected_registers;
 static jmp_buf save_regs_gc_mark;
 static double gc_rt;
@@ -230,11 +239,30 @@ static LISP (*user_readm) (int, struct gen_readio *);
 static LISP (*user_readt) (char *, long, int *);
 static void (*fatal_exit_hook) (void);
 static long stack_size;
+static struct func_frame *func_trace;
+
+#if DEBUG_SCM
+static LISP dbg_pos = NIL;
+
+static LISP orig_readtl (struct gen_readio * f);
+static int dbg_getc (struct gen_readio * f);
+static void dbg_ungetc (int c, struct gen_readio * f);
+static void dbg_readini (char *file);
+static void dbg_readend (void);
+static void dbg_lineinc (void);
+static void dbg_linedec (void);
+#if 0
+static LISP dbg_curpos (void);
+#endif
+static void init_dbg (void);
+#endif /* DEBUG_SCM */
 
 static LISP lreadparen (struct gen_readio * f);
 static LISP lreadr (struct gen_readio *f);
 static LISP my_err(char *message, LISP obj);
 static LISP lprint (LISP exp, LISP);
+static void gc_protect (LISP * location);
+static LISP provide (LISP name);
 
 #define ENVLOOKUP_TRICK 1
 
@@ -293,6 +321,29 @@ setvar (LISP var, LISP val, LISP env)
   return (CAR (tmp) = val);
 }
 
+static void
+show_backtrace(void)
+{
+  struct func_frame *fr;
+  fprintf(siod_output, "*backtrace*\n");
+  for (fr = func_trace; fr; fr = fr->prev) {
+    fprintf(siod_output, ">>");
+    lprin1f(fr->obj, siod_output);
+#if DEBUG_SCM
+    if NNULLP
+      (fr->obj->dbg_info)
+      {
+	fprintf (siod_output,
+		 " at %s:%d",
+		 CAR (fr->obj->dbg_info)->storage_as.string.data,
+		 INTNM (CDR (fr->obj->dbg_info)));
+      }
+#endif
+    fprintf(siod_output, "\n");
+  }
+  fprintf(siod_output, "\n");
+}
+
 static LISP
 my_err (char *message, LISP x)
 {
@@ -300,7 +351,6 @@ my_err (char *message, LISP x)
   long was_inside = inside_err;
   LISP retval, nx;
   char *msg, *eobj;
-  nointerrupt = 1;
   if ((!message) && CONSP (x) && TYPEP (CAR (x), tc_string))
     {
       msg = get_c_string (CAR (x));
@@ -329,8 +379,12 @@ my_err (char *message, LISP x)
       fprintf (siod_output, "%s\n", siod_err_msg);
       fflush (siod_output);
     }
+  if (siod_verbose_level >= 5)
+    show_backtrace();
   if (errjmp_ok == 1)
     {
+      /* prevent recording of bogus debug info */
+      dbg_readend ();
       inside_err = 1;
       setvar (sym_errobj, nx, NIL);
       for (l = catch_framep; l; l = (*l).next)
@@ -342,7 +396,6 @@ my_err (char *message, LISP x)
 	    (*l).retval = (NNULLP (retval) ? retval :
 			   (was_inside) ? NIL :
 			   cons (strcons (strlen (msg), msg), nx));
-	    nointerrupt = 0;
 	    inside_err = 0;
 	    longjmp ((*l).cframe, 2);
 	  }
@@ -386,16 +439,6 @@ get_user_type_hooks (long type)
   return (NULL);
 }
 
-long
-no_interrupt (long n)
-{
-  long x;
-  x = nointerrupt;
-  nointerrupt = n;
-
-  return (x);
-}
-
 int
 get_c_int (LISP x)
 {
@@ -446,16 +489,13 @@ myrealtime (void)
 static void
 fput_st (FILE * f, char *st)
 {
-  long flag;
-  flag = no_interrupt (1);
   if (siod_verbose_level >= 1)
     {
-      //      fprintf (stdout, "%s", st);
-      //      fflush (siod_output);
+      /*      fprintf (stdout, "%s", st); */
+      /*      fflush (siod_output); */
       fprintf (f, "%s", st);
       fflush (f);
     }
-  no_interrupt (flag);
 }
 
 static void
@@ -546,7 +586,7 @@ setcdr (LISP cell, LISP value)
   return (CDR (cell) = value);
 }
 
-LISP
+static LISP
 newcell (long type)
 {
   LISP z;
@@ -558,9 +598,7 @@ static LISP
 fopen_cg (FILE * (*fcn) (const char *, const char *), char *name, char *how)
 {
   LISP sym;
-  long flag;
   char errmsg[80];
-  flag = no_interrupt (1);
   sym = newcell (tc_c_file);
   sym->storage_as.c_file.f = (FILE *) NULL;
   sym->storage_as.c_file.name = (char *) NULL;
@@ -571,7 +609,6 @@ fopen_cg (FILE * (*fcn) (const char *, const char *), char *name, char *how)
     }
   sym->storage_as.c_file.name = (char *) must_malloc (strlen (name) + 1);
   strcpy (sym->storage_as.c_file.name, name);
-  no_interrupt (flag);
   return (sym);
 }
 
@@ -599,12 +636,9 @@ file_gc_free (LISP ptr)
 static LISP
 fclose_l (LISP p)
 {
-  long flag;
-  flag = no_interrupt (1);
   if NTYPEP
     (p, tc_c_file) my_err ("not a file", p);
   file_gc_free (p);
-  no_interrupt (flag);
   return (NIL);
 }
 
@@ -633,11 +667,6 @@ vload (char *fname, long cflag, long rflag)
   char  *key = "parser:", *start, *end, *ftype = ".scm";
   if (rflag)
     {
-      int iflag;
-      iflag = no_interrupt (1);
-      /*      if ((f = fopen (fname, "r")))
-	fclose (f);
-	else*/
       if ((fname[0] != '/'))
 	{
 	  fnbuf = alloca(strlen(siod_lib) + strlen(fname) + 2);
@@ -650,7 +679,6 @@ vload (char *fname, long cflag, long rflag)
 	      fclose (f);
 	    }
 	}
-      no_interrupt (iflag);
     }
   if (siod_verbose_level >= 3)
     {
@@ -660,6 +688,7 @@ vload (char *fname, long cflag, long rflag)
     }
   lf = fopen_c (fname, "r");
   f = lf->storage_as.c_file.f;
+  dbg_readini (fname);
   result = NIL;
   tail = NIL;
   j = 0;
@@ -668,13 +697,18 @@ vload (char *fname, long cflag, long rflag)
   while ((c == '#') || (c == ';'))
     {
       while (((c = getc (f)) != EOF) && (c != '\n'))
-	if ((j + 1) < sizeof (buffer))
+	if ((j + 1) < (int)sizeof (buffer))
 	  {
 	    buffer[j] = c;
 	    buffer[++j] = 0;
 	  }
       if (c == '\n')
-	c = getc (f);
+	{
+	  c = getc (f);
+#ifdef DEBUG_SCM
+	  dbg_lineinc ();
+#endif
+	}
     }
   if (c != EOF)
     ungetc (c, f);
@@ -719,6 +753,7 @@ vload (char *fname, long cflag, long rflag)
 	leval (form, NIL);
     }
   fclose_l (lf);
+  dbg_readend ();
   if (siod_verbose_level >= 3)
     put_st ("done.\n");
   return (result);
@@ -730,6 +765,7 @@ repl_driver (long want_init, struct repl_hooks *h)
   int k;
   struct repl_hooks hd;
   LISP stack_start;
+  func_trace = NULL;
   stack_start_ptr = &stack_start;
   stack_limit_ptr = STACK_LIMIT (stack_start_ptr, stack_size);
   k = setjmp (errjmp);
@@ -737,7 +773,6 @@ repl_driver (long want_init, struct repl_hooks *h)
     return (2);
   catch_framep = (struct catch_frame *) NULL;
   errjmp_ok = 1;
-  nointerrupt = 0;
   if (want_init && init_file && (k == 0))
     vload (init_file, 0, 1);
   if (!h)
@@ -815,9 +850,7 @@ flush_ws (struct gen_readio *f, char *eoferr)
 LISP
 strcons (long length, const char *data)
 {
-  long flag;
   LISP s;
-  flag = no_interrupt (1);
   s = cons (NIL, NIL);
   s->type = tc_string;
   if (length == -1)
@@ -827,7 +860,6 @@ strcons (long length, const char *data)
   if (data)
     memcpy (s->storage_as.string.data, data, length);
   s->storage_as.string.data[length] = 0;
-  no_interrupt (flag);
   return (s);
 }
 
@@ -984,6 +1016,11 @@ lreadr (struct gen_readio *f)
 	  UNGETC_FCN (c, f);
 	}
       return (cons (rintern (p), lreadr (f)));
+      /* We use the form (_ "str") to represent gettext string to be
+	 compatible with other lisp implementations, so the strange
+	 _"str" syntax handled below are removed -- YamaKen 2004-09-30
+      */
+#if 0
     case '_':  /*  might be a string marked for translation using _(...)  */
       c = GETC_FCN (f);
       if (c == '"')
@@ -991,6 +1028,7 @@ lreadr (struct gen_readio *f)
       else
 	UNGETC_FCN (c, f);
       break;
+#endif
     case '"':
       return (lreadstring (f));
     case '#':
@@ -1057,6 +1095,99 @@ lreadparen (struct gen_readio * f)
 
 static LISP
 readtl (struct gen_readio * f)
+#if DEBUG_SCM
+{
+  /* wrapper that prepares debugging information */
+  if NNULLP
+    (dbg_pos)
+    {
+      struct gen_readio s;
+
+      s.getc_fcn = (int (*)(void *)) dbg_getc;
+      s.ungetc_fcn = (void (*)(int, void *)) dbg_ungetc;
+      s.cb_argument = (void *) f;
+      f = &s;
+    }
+  return orig_readtl (f);
+}
+
+static int
+dbg_getc (struct gen_readio * f)
+{
+  int c;
+  c = GETC_FCN (f);
+  if (c == '\n')
+    dbg_lineinc ();
+  return c;
+}
+
+static void
+dbg_ungetc (int c, struct gen_readio * f)
+{
+  UNGETC_FCN (c, f);
+  if (c == '\n')
+    dbg_linedec ();
+}
+
+static void
+dbg_lineinc (void)
+{
+  LISP file, line;
+  if CONSP
+    (dbg_pos)
+    {
+      file = CAR (CAR (dbg_pos));
+      line = CDR (CAR (dbg_pos));
+      line = intcons (INTNM (line) + 1);
+      CAR (dbg_pos) = cons (file, line);
+      line->dbg_info = CAR (dbg_pos)->dbg_info = NIL;
+    }
+  /* else: we have given up debugging information */
+}
+
+static void
+dbg_linedec (void)
+{
+  LISP file, line;
+  if CONSP
+    (dbg_pos)
+    {
+      file = CAR (CAR (dbg_pos));
+      line = CDR (CAR (dbg_pos));
+      line = intcons (INTNM (line) - 1);
+      CAR (dbg_pos) = cons (file, line);
+      line->dbg_info = CAR (dbg_pos)->dbg_info = NIL;
+    }
+  /* else: we have given up debugging information */
+}
+
+static void
+dbg_readini (char *filename)
+{
+  LISP tmp;
+  tmp = cons (strcons (-1, filename), intcons (1));
+  tmp->dbg_info = CAR(tmp)->dbg_info = CDR(tmp)->dbg_info = NIL;
+  dbg_pos = cons (tmp, dbg_pos);
+  dbg_pos->dbg_info = NIL;
+}
+
+static void
+dbg_readend (void)
+{
+  dbg_pos = cdr (dbg_pos);
+}
+
+static void
+init_dbg (void)
+{
+  dbg_pos = NIL;
+  gc_protect (&dbg_pos);
+  provide (rintern ("debug"));
+}
+
+static LISP
+orig_readtl (struct gen_readio * f)
+#endif /* DEBUG_SCM */
 {
   int c;
   c = flush_ws (f, (char *) NULL);
@@ -1141,13 +1272,13 @@ set_repl_hooks (void (*puts_f) (char *),
   repl_print = print_f;
 }
 
-void
+static void
 gput_st (struct gen_printio *f, char *st)
 {
   PUTS_FCN (st, f);
 }
 
-int
+static int
 fputs_fcn (char *st, void *cb)
 {
   fput_st ((FILE *) cb, st);
@@ -1198,7 +1329,7 @@ stack_limit (LISP amount, LISP silent)
     (silent)
     {
       sprintf (tkbuffer, "Stack_size = %ld bytes, [%p,%p]\n",
-	       stack_size, stack_start_ptr, stack_limit_ptr);
+	       stack_size, (void *)stack_start_ptr, stack_limit_ptr);
       put_st (tkbuffer);
       return (NIL);
     }
@@ -1406,7 +1537,7 @@ Quotient (LISP x, LISP y)
     }
 }
 
-LISP
+static LISP
 lllabs (LISP x)
 {
   double v;
@@ -1443,7 +1574,7 @@ lessp (LISP x, LISP y)
   return (NIL);
 }
 
-LISP
+static LISP
 greaterEp (LISP x, LISP y)
 {
   if NINTNUMP
@@ -1455,7 +1586,7 @@ greaterEp (LISP x, LISP y)
   return (NIL);
 }
 
-LISP
+static LISP
 lessEp (LISP x, LISP y)
 {
   if NINTNUMP
@@ -1467,7 +1598,7 @@ lessEp (LISP x, LISP y)
   return (NIL);
 }
 
-LISP
+static LISP
 lmax (LISP x, LISP y)
 {
   if NULLP
@@ -1479,7 +1610,7 @@ lmax (LISP x, LISP y)
   return ((INTNM (x) > INTNM (y)) ? x : y);
 }
 
-LISP
+static LISP
 lmin (LISP x, LISP y)
 {
   if NULLP
@@ -1541,7 +1672,7 @@ loop:
     }
 }
 
-LISP
+static LISP
 eq (LISP x, LISP y)
 {
   if EQ
@@ -1620,7 +1751,7 @@ symbolp (LISP x)
     return (NIL);
 }
 
-LISP
+static LISP
 err_ubv (LISP v)
 {
   return (my_err ("unbound variable", v));
@@ -1697,8 +1828,7 @@ gen_intern (const char *name)
 {
   LISP l, sym, sl;
   char *cname;
-  long hash = 0, flag;
-  flag = no_interrupt (1);
+  long hash = 0;
   if (obarray_dim > 1)
     {
       hash = name_hash(name);
@@ -1709,7 +1839,6 @@ gen_intern (const char *name)
   for (l = sl; NNULLP (l); l = CDR (l))
     if (strcmp (name, PNAME (CAR (l))) == 0)
       {
-	no_interrupt (flag);
 	return (CAR (l));
       }
 
@@ -1720,7 +1849,6 @@ gen_intern (const char *name)
   if (obarray_dim > 1)
     obarray[hash] = cons (sym, sl);
   oblistvar = cons (sym, oblistvar);
-  no_interrupt (flag);
   return (sym);
 }
 
@@ -1745,7 +1873,7 @@ unlink_from_sym_list(const char *name, LISP *lst)
   }
 }
 
-void
+static void
 do_undefine(const char *name)
 {
   int hash;
@@ -1756,8 +1884,8 @@ do_undefine(const char *name)
   }
 }
 
-LISP
-undefine(LISP name)
+static LISP
+undefine(LISP name, LISP env)
 {
   /* not yet tested */
   if SYMBOLP(name) {
@@ -1822,6 +1950,7 @@ gc_protect_sym (LISP * location, char *st)
   gc_protect (location);
 }
 
+/* This function will resurrect when we implement module loading
 static void
 gc_unprotect (LISP * location)
 {
@@ -1848,6 +1977,7 @@ gc_unprotect (LISP * location)
       reg = reg->next;
     }
 }
+*/
 
 static LISP
 string_gc_mark (LISP ptr)
@@ -1959,7 +2089,7 @@ file_prin1 (LISP ptr, struct gen_printio *f)
   char *name;
   name = ptr->storage_as.c_file.name;
   gput_st (f, "#<FILE ");
-  sprintf (tkbuffer, " %p", ptr->storage_as.c_file.f);
+  sprintf (tkbuffer, " %p", (void *)ptr->storage_as.c_file.f);
   gput_st (f, tkbuffer);
   if (name)
     {
@@ -2009,12 +2139,12 @@ init_storage_1 (void)
   gc_protect_sym (&sym_catchall, "all");
   gc_protect_sym (&sym_progn, "begin");
   gc_protect_sym (&sym_lambda, "lambda");
+  gc_protect_sym (&sym_else, "else");
+  setvar (sym_else, sym_t, NIL);
   gc_protect_sym (&sym_quote, "quote");
   gc_protect_sym (&sym_dot, ".");
   gc_protect_sym (&sym_after_gc, "*after-gc*");
   setvar (sym_after_gc, NIL, NIL);
-  gc_protect_sym (&sym_eval_history_ptr, "*eval-history-ptr*");
-  setvar (sym_eval_history_ptr, NIL, NIL);
   gc_protect_sym (&sym_features, "features");
   setvar (sym_features, NIL, NIL);
   if (inums_dim > 0)
@@ -2124,25 +2254,14 @@ assq (LISP x, LISP alist)
   return (my_err ("improper list to assq", alist));
 }
 
-static long
-allocate_user_tc (void)
-{
-  long x = user_tc_next;
-  if (x > tc_user_max)
-    my_err ("ran out of user type codes", NIL);
-  ++user_tc_next;
-  return (x);
-}
-
 static LISP
 allocate_aheap (void)
 {
-  long j, flag;
+  long j;
   LISP ptr, end, next;
   for (j = 0; j < nheaps; ++j)
     if (!heaps[j])
       {
-	flag = no_interrupt (1);
 	if (gc_status_flag && (siod_verbose_level >= 4))
 	  fprintf (siod_output, "[allocating heap %ld]\n", j);
 	heaps[j] = (LISP) must_malloc (sizeof (struct obj) * heap_size);
@@ -2164,7 +2283,6 @@ allocate_aheap (void)
 	      }
 	  }
 	freelist = heaps[j];
-	flag = no_interrupt (flag);
 	return (sym_t);
       }
   return (NIL);
@@ -2196,6 +2314,9 @@ gc_mark_loop:
   if ((*ptr).gc_mark)
     return;
   (*ptr).gc_mark = 1;
+#if DEBUG_SCM
+  gc_mark ((*ptr).dbg_info);
+#endif
   switch ((*ptr).type)
     {
     case tc_intnum:
@@ -2377,7 +2498,7 @@ gc_mark_and_sweep (void)
 static void
 gc_for_newcell (void)
 {
-  long flag, n;
+  long n;
   LISP l;
   if (heap < heap_end)
     {
@@ -2388,11 +2509,9 @@ gc_for_newcell (void)
     }
   if (errjmp_ok == 0)
     gc_fatal_error ();
-  flag = no_interrupt (1);
   errjmp_ok = 0;
   gc_mark_and_sweep ();
   errjmp_ok = 1;
-  no_interrupt (flag);
   for (n = 0, l = freelist; (n < 100) && NNULLP (l); ++n)
     l = CDR (l);
   if (n == 0)
@@ -2410,8 +2529,7 @@ gc_for_newcell (void)
 static LISP
 user_gc (LISP args)
 {
-  long old_status_flag, flag;
-  flag = no_interrupt (1);
+  long old_status_flag;
   errjmp_ok = 0;
   old_status_flag = gc_status_flag;
   if NNULLP (args)
@@ -2424,7 +2542,6 @@ user_gc (LISP args)
   gc_mark_and_sweep ();
   gc_status_flag = old_status_flag;
   errjmp_ok = 1;
-  no_interrupt (flag);
   return (NIL);
 }
 
@@ -2538,32 +2655,36 @@ err_closure_code (LISP tmp)
   return (my_err ("closure code type not valid", tmp));
 }
 
+/* main evaluator */
 LISP
 leval (LISP x, LISP env)
 {
   LISP tmp, arg1;
+  LISP rval;
   struct user_type_hooks *p;
+  struct func_frame this_frame;
   STACK_CHECK (&x);
+  this_frame.prev = func_trace;
+  this_frame.obj = x;
+  func_trace = &this_frame;
 loop:
   INTERRUPT_CHECK ();
-  tmp = VCELL (sym_eval_history_ptr);
-  if TYPEP
-    (tmp, tc_cons)
-    {
-      CAR (tmp) = x;
-      VCELL (sym_eval_history_ptr) = CDR (tmp);
-    }
   switch TYPE
     (x)
     {
     case tc_symbol:
       tmp = envlookup (x, env);
       if NNULLP
-	(tmp) return (CAR (tmp));
+	(tmp) {
+	rval = (CAR (tmp));
+	goto ret;
+      }
       tmp = VCELL (x);
       if EQ
 	(tmp, unbound_marker) err_ubv (x);
-      return (tmp);
+      rval = tmp;
+      goto ret;
+
     case tc_cons:
       tmp = CAR (x);
       switch TYPE
@@ -2589,15 +2710,18 @@ loop:
 	(tmp)
 	{
 	case tc_subr_0:
-	  return (SUBR0 (tmp) ());
+	  rval = (SUBR0 (tmp) ());
+	  goto ret;
 	case tc_subr_1:
-	  return (SUBR1 (tmp) (leval (car (CDR (x)), env)));
+	  rval = (SUBR1 (tmp) (leval (car (CDR (x)), env)));
+	  goto ret;
 	case tc_subr_2:
 	  x = CDR (x);
 	  arg1 = leval (car (x), env);
 	  x = NULLP (x) ? NIL : CDR (x);
-	  return (SUBR2 (tmp) (arg1,
+	  rval = (SUBR2 (tmp) (arg1,
 			       leval (car (x), env)));
+	  goto ret;
 	case tc_subr_2n:
 	  x = CDR (x);
 	  arg1 = leval (car (x), env);
@@ -2606,41 +2730,50 @@ loop:
 			      leval (car (x), env));
 	  for (x = cdr (x); CONSP (x); x = CDR (x))
 	    arg1 = SUBR2 (tmp) (arg1, leval (CAR (x), env));
-	  return (arg1);
+	  rval = (arg1);
+	  goto ret;
 	case tc_subr_3:
 	  x = CDR (x);
 	  arg1 = leval (car (x), env);
 	  x = NULLP (x) ? NIL : CDR (x);
-	  return (SUBR3 (tmp) (arg1,
+	  rval = (SUBR3 (tmp) (arg1,
 			       leval (car (x), env),
 			       leval (car (cdr (x)), env)));
+	  goto ret;
 
 	case tc_subr_4:
 	  x = CDR (x);
 	  arg1 = leval (car (x), env);
 	  x = NULLP (x) ? NIL : CDR (x);
-	  return (SUBR4 (tmp) (arg1,
+	  rval = (SUBR4 (tmp) (arg1,
 			       leval (car (x), env),
 			       leval (car (cdr (x)), env),
 			       leval (car (cdr (cdr (x))), env)));
+	  goto ret;
 
 	case tc_subr_5:
 	  x = CDR (x);
 	  arg1 = leval (car (x), env);
 	  x = NULLP (x) ? NIL : CDR (x);
-	  return (SUBR5 (tmp) (arg1,
-			       leval (car (x), env),
-			       leval (car (cdr (x)), env),
-			       leval (car (cdr (cdr (x))), env),
-			       leval (car (cdr (cdr (cdr (x)))), env)));
+	  rval = (SUBR5 (tmp) (arg1,
+			      leval (car (x), env),
+			      leval (car (cdr (x)), env),
+			      leval (car (cdr (cdr (x))), env),
+			      leval (car (cdr (cdr (cdr (x)))), env)));
+	  goto ret;
 
 	case tc_lsubr:
-	  return (SUBR1 (tmp) (leval_args (CDR (x), env)));
+	  rval = (SUBR1 (tmp) (leval_args (CDR (x), env)));
+	  goto ret;
 	case tc_fsubr:
-	  return (SUBR2 (tmp) (CDR (x), env));
+	  rval = (SUBR2 (tmp) (CDR (x), env));
+	  goto ret;
 	case tc_msubr:
 	  if NULLP
-	    (SUBRM (tmp) (&x, &env)) return (x);
+	    (SUBRM (tmp) (&x, &env)) {
+	    rval = x;
+	    goto ret;
+	  }
 	  goto loop;
 	case tc_closure:
 	  switch TYPE
@@ -2653,45 +2786,51 @@ loop:
 	      x = CDR ((*tmp).storage_as.closure.code);
 	      goto loop;
 	    case tc_subr_1:
-	      return (SUBR1 (tmp->storage_as.closure.code)
+	      rval = (SUBR1 (tmp->storage_as.closure.code)
 		      (tmp->storage_as.closure.env));
+	      goto ret;
 	    case tc_subr_2:
 	      x = CDR (x);
 	      arg1 = leval (car (x), env);
-	      return (SUBR2 (tmp->storage_as.closure.code)
+	      rval = (SUBR2 (tmp->storage_as.closure.code)
 		      (tmp->storage_as.closure.env, arg1));
+	      goto ret;
 	    case tc_subr_3:
 	      x = CDR (x);
 	      arg1 = leval (car (x), env);
 	      x = NULLP (x) ? NIL : CDR (x);
-	      return (SUBR3 (tmp->storage_as.closure.code)
+	      rval = (SUBR3 (tmp->storage_as.closure.code)
 		      (tmp->storage_as.closure.env,
 		       arg1,
 		       leval (car (x), env)));
+	      goto ret;
 	    case tc_subr_4:
 	      x = CDR (x);
 	      arg1 = leval (car (x), env);
 	      x = NULLP (x) ? NIL : CDR (x);
-	      return (SUBR4 (tmp->storage_as.closure.code)
+	      rval = (SUBR4 (tmp->storage_as.closure.code)
 		      (tmp->storage_as.closure.env,
 		       arg1,
 		       leval (car (x), env),
 		       leval (car (cdr (x)), env)));
+	      goto ret;
 	    case tc_subr_5:
 	      x = CDR (x);
 	      arg1 = leval (car (x), env);
 	      x = NULLP (x) ? NIL : CDR (x);
-	      return (SUBR5 (tmp->storage_as.closure.code)
+	      rval = (SUBR5 (tmp->storage_as.closure.code)
 		      (tmp->storage_as.closure.env,
 		       arg1,
 		       leval (car (x), env),
 		       leval (car (cdr (x)), env),
 		       leval (car (cdr (cdr (x))), env)));
+	      goto ret;
 
 	    case tc_lsubr:
-	      return (SUBR1 (tmp->storage_as.closure.code)
+	      rval = (SUBR1 (tmp->storage_as.closure.code)
 		      (cons (tmp->storage_as.closure.env,
 			     leval_args (CDR (x), env))));
+	      goto ret;
 	    default:
 	      err_closure_code (tmp);
 	    }
@@ -2705,15 +2844,21 @@ loop:
 	  if (p->leval)
 	    {
 	      if NULLP
-		((*p->leval) (tmp, &x, &env)) return (x);
-	      else
+		((*p->leval) (tmp, &x, &env)) {
+		rval = x;
+		goto ret;
+	      } else
 		goto loop;
 	    }
 	  my_err ("bad function", tmp);
 	}
     default:
-      return (x);
+      rval = x;
+      goto ret;
     }
+ ret:
+  func_trace = this_frame.prev;
+  return rval;
 }
 
 LISP
@@ -2962,14 +3107,17 @@ static LISP
 leval_catch (LISP args, LISP env)
 {
   struct catch_frame frame;
+  struct func_frame *cur_func;
   int k;
   frame.tag = leval (car (args), env);
   frame.next = catch_framep;
+  cur_func = func_trace;
   k = setjmp (frame.cframe);
   catch_framep = &frame;
   if (k == 2)
     {
       catch_framep = frame.next;
+      func_trace = cur_func;
       return (frame.retval);
     }
   return (leval_catch_1 (cdr (args), env));
@@ -3001,7 +3149,7 @@ leval_let (LISP * pform, LISP * penv)
   return (sym_t);
 }
 
-LISP
+static LISP
 letstar_macro (LISP form)
 {
   LISP bindings = cadr (form);
@@ -3015,7 +3163,7 @@ letstar_macro (LISP form)
   return (form);
 }
 
-static LISP
+LISP
 reverse (LISP l)
 {
   LISP n, p;
@@ -3114,7 +3262,7 @@ leval_tenv (LISP args, LISP env)
   return (env);
 }
 
-LISP
+static LISP
 leval_while (LISP args, LISP env)
 {
   LISP l;
@@ -3247,7 +3395,7 @@ lprin1g (LISP exp, struct gen_printio * f)
 	(*p->prin1) (exp, f);
       else
 	{
-	  sprintf (tkbuffer, "#<UNKNOWN %d %p>", TYPE (exp), exp);
+	  sprintf (tkbuffer, "#<UNKNOWN %d %p>", TYPE (exp), (void *)exp);
 	  gput_st (f, tkbuffer);
 	}
     }
@@ -3341,11 +3489,8 @@ lread (LISP f)
 static int
 f_getc (FILE * f)
 {
-  long iflag;
   int c;
-  iflag = no_interrupt (1);
   c = getc (f);
-  no_interrupt (iflag);
   return (c);
 }
 
@@ -3365,17 +3510,6 @@ lreadf (FILE * f)
   return (readtl (&s));
 }
 
-static void
-set_read_hooks (char *all_set, char *end_set,
-		LISP (*fcn1) (int, struct gen_readio *),
-		LISP (*fcn2) (char *, long, int *))
-{
-  user_ch_readm = all_set;
-  user_te_readm = end_set;
-  user_readm = fcn1;
-  user_readt = fcn2;
-}
-
 LISP
 lreadtk (char *buffer, long j)
 {
@@ -3484,13 +3618,13 @@ memq (LISP x, LISP il)
   return (my_err ("improper list to memq", il));
 }
 
-LISP
+static LISP
 featurep (LISP name)
 {
-  return memq (name, sym_features);
+  return memq (name, leval (sym_features, NIL));
 }
 
-LISP
+static LISP
 provide (LISP name)
 {
   if NSYMBOLP
@@ -3499,7 +3633,7 @@ provide (LISP name)
     return NIL;
   }
   if (featurep(name) == NIL)
-    setvar (sym_features, cons (name, sym_features), NIL);
+    setvar (sym_features, cons (name, leval (sym_features, NIL)), NIL);
   return (NIL);
 }
 
@@ -3571,7 +3705,7 @@ get_c_file (LISP p, FILE * deflt)
   return (p->storage_as.c_file.f);
 }
 
-LISP
+static LISP
 lgetc (LISP p)
 {
   int i;
@@ -3579,7 +3713,7 @@ lgetc (LISP p)
   return ((i == EOF) ? NIL : intcons ((double) i));
 }
 
-LISP
+static LISP
 lungetc (LISP ii, LISP p)
 {
   int i;
@@ -3592,10 +3726,9 @@ lungetc (LISP ii, LISP p)
   return (NIL);
 }
 
-LISP
+static LISP
 lputc (LISP c, LISP p)
 {
-  long flag;
   int i;
   FILE *f;
   f = get_c_file (p, siod_output);
@@ -3604,13 +3737,11 @@ lputc (LISP c, LISP p)
       i = (int) INTNM (c);
   else
     i = *get_c_string (c);
-  flag = no_interrupt (1);
   putc (i, f);
-  no_interrupt (flag);
   return (NIL);
 }
 
-LISP
+static LISP
 lputs (LISP str, LISP p)
 {
   fput_st (get_c_file (p, siod_output), get_c_string (str));
@@ -3670,14 +3801,14 @@ siod_verbose (LISP arg)
   return (intcons (siod_verbose_level));
 }
 
-LISP
+static LISP
 lruntime (void)
 {
   return (cons (intcons (myruntime ()),
 		cons (intcons (gc_time_taken), NIL)));
 }
 
-LISP
+static LISP
 lrealtime (void)
 {
   return (intcons (myrealtime ()));
@@ -3707,7 +3838,7 @@ cddr (LISP x)
   return (cdr (cdr (x)));
 }
 
-LISP
+static LISP
 letrec_macro (LISP form)
 {
   LISP letb, setb, l;
@@ -3721,7 +3852,7 @@ letrec_macro (LISP form)
   return (form);
 }
 
-LISP
+static LISP
 lrand (LISP m)
 {
   long res;
@@ -3733,7 +3864,7 @@ lrand (LISP m)
     return (intcons (res % get_c_int (m)));
 }
 
-LISP
+static LISP
 lsrand (LISP s)
 {
   srand (get_c_int (s));
@@ -3777,7 +3908,7 @@ llast_c_errmsg (int num)
   return (rintern ((char *) errmsg));
 }
 
-LISP
+static LISP
 lllast_c_errmsg (void)
 {
   return (llast_c_errmsg (-1));
@@ -3880,13 +4011,13 @@ llist (LISP l)
 }
 
 
-LISP
+static LISP
 lstrspn (LISP str1, LISP str2)
 {
   return (intcons (strspn (get_c_string (str1), get_c_string (str2))));
 }
 
-LISP
+static LISP
 lstrcspn (LISP str1, LISP str2)
 {
   return (intcons (strcspn (get_c_string (str1), get_c_string (str2))));
@@ -3927,7 +4058,63 @@ butlast (LISP l)
   return (my_err ("not a list", l));
 }
 
-LISP
+
+/* Note: relies on car(), cdr () and predicates for nil-check */
+static LISP
+leval_case (LISP * pform, LISP * penv)
+{
+  LISP args, env, clause, key, next, data;
+  args = cdr (*pform);
+  env = *penv;
+  key = leval (car (args), env);
+  args = cdr (args);
+  next = cdr (args);
+  while NNULLP
+    (next)
+    {
+      clause = car (args);
+      data = car (clause);
+      while NNULLP
+	(data)
+	{
+	  if (eql (key, car (data)))
+	    goto progn;
+	  data = cdr (data);
+	}
+      args = next;
+      next = cdr (next);
+    }
+  /* last clause; might be `else' */
+  clause = car (args);
+  data = car (clause);
+  if (eq (data, sym_else))
+    goto progn;
+  while NNULLP
+    (data)
+    {
+      if (eql (key, car (data)))
+	goto progn;
+      data = cdr (data);
+    }
+  *pform = NIL;
+  return (NIL);
+
+ progn:
+  clause = cdr (clause);
+  next = cdr (clause);
+  while NNULLP
+    (next)
+    {
+      leval (car (clause), env);
+      clause = next;
+      next = cdr (next);
+    }
+  *pform = car (clause);
+  return (sym_t);
+}
+
+
+static LISP
 leval_cond (LISP * pform, LISP * penv)
 {
   LISP args, env, clause, value, next;
@@ -4093,7 +4280,7 @@ memv (LISP x, LISP il)
   return (my_err ("improper list to memv", il));
 }
 
-LISP
+static LISP
 lsubset (LISP fcn, LISP l)
 {
   LISP result = NIL, v;
@@ -4198,6 +4385,7 @@ init_subrs (void)
   init_subr_2 ("=", eql);
   init_subr_2 ("assq", assq);
   init_msubr ("cond", leval_cond);
+  init_msubr ("case", leval_case);
   init_subr_2 ("delq", delq);
   init_subr_1 ("read", lread);
   init_subr_1 ("parser_read", parser_read);
@@ -4302,32 +4490,6 @@ init_subrs (void)
   init_slib_version ();
 }
 
-
-/* err0,pr,prp are convenient to call from the C-language debugger */
-
-void
-err0 (void)
-{
-  my_err ("0", NIL);
-}
-
-void
-pr (LISP p)
-{
-  if (looks_pointerp (p))
-    lprint (p, NIL);
-  else
-    put_st ("invalid\n");
-}
-
-void
-prp (LISP * p)
-{
-  if (!p)
-    return;
-  pr (*p);
-}
-
 void
 siod_quit (void)
 {
@@ -4383,17 +4545,16 @@ siod_init (int argc, char **argv, int warnflag, FILE *fp)
   stack_start_ptr = NULL;
   freelist = NIL;
   errjmp_ok = 0;
-  nointerrupt = 1;
   oblistvar = NIL;
   eof_val = NIL;
   sym_errobj = NIL;
   sym_catchall = NIL;
   sym_progn = NIL;
   sym_lambda = NIL;
+  sym_else = NIL;
   sym_quote = NIL;
   sym_dot = NIL;
   sym_after_gc = NIL;
-  sym_eval_history_ptr = NIL;
   sym_features = NIL;
   unbound_marker = NIL;
   obarray = NULL;
@@ -4407,7 +4568,6 @@ siod_init (int argc, char **argv, int warnflag, FILE *fp)
   inums = NULL;
   inums_dim = 256;
   user_types = NULL;
-  user_tc_next = tc_user_min;
   protected_registers = NULL;
   gc_rt = 0;
   gc_cells_collected = 0;
@@ -4417,6 +4577,7 @@ siod_init (int argc, char **argv, int warnflag, FILE *fp)
   user_readt = NULL;
   fatal_exit_hook = NULL;
   stack_size = 50000;
+  func_trace = NULL;
 
   /* parse arguments */
   for (k = 1; k < argc; ++k)
@@ -4461,4 +4622,8 @@ siod_init (int argc, char **argv, int warnflag, FILE *fp)
     }
   init_storage ();
   init_subrs ();
+  init_intl ();
+#if DEBUG_SCM
+  init_dbg ();
+#endif
 }
