@@ -1,3 +1,4 @@
+
 /*
 
   Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
@@ -40,8 +41,9 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <uim/uim.h>
-#include <uim/uim-helper.h>
+#include <unistd.h>
+#include "uim.h"
+#include "uim-helper.h"
 #include "context.h"
 
 static int uim_fd = -1;
@@ -49,15 +51,14 @@ static void (*uim_disconnect_cb)(void);
 
 static void uim_helper_client_focus(uim_context uc, int flg);
 
-#define BUFFER_SIZE 2048
+#define BUFFER_SIZE (32 * 1024)
 /*Common buffer for some functions's temporary buffer.
   Pay attention for use.*/
 static char uim_help_buf[BUFFER_SIZE];
-static int uim_read_buf_count;
-static int uim_read_buf_error;
-static char uim_read_buf[BUFFER_SIZE];
+static int uim_read_buf_size;
+static char *uim_read_buf;
 
-char*
+static char *
 get_server_command(void)
 {
   return "uim-helper-server";
@@ -66,7 +67,6 @@ get_server_command(void)
 
 int uim_helper_init_client_fd(void (*disconnect_cb)(void))
 {
-  struct stat lst;
   int fd;
   struct sockaddr_un server;
   char *path = uim_helper_get_pathname();
@@ -76,10 +76,10 @@ int uim_helper_init_client_fd(void (*disconnect_cb)(void))
   if(!path)
     return -1;
 
-  memset(&server, 0, sizeof(server));
-  
+  bzero(&server, sizeof(server));
   server.sun_family = PF_UNIX;
   strcpy(server.sun_path, path);
+
   free(path);
   
   fd = socket(PF_UNIX, SOCK_STREAM, 0);
@@ -122,10 +122,9 @@ int uim_helper_init_client_fd(void (*disconnect_cb)(void))
     close(fd);
     return -1;
   }
- 
+
+  uim_read_buf = strdup("");
   uim_disconnect_cb = disconnect_cb;
-  uim_read_buf_count = 0;
-  uim_read_buf_error = 0;
   uim_fd = fd;
   return fd;
 }
@@ -184,54 +183,52 @@ uim_helper_client_get_prop_list(void)
 void
 uim_helper_read_proc(int fd)
 {
-  int rc, count;
+  char buf[BUFFER_SIZE];
+  int rc;
 
-  count = BUFFER_SIZE - uim_read_buf_count;
-  if (count == 0) {
-    return ;
-  }
-  rc = read(fd, uim_read_buf, count);
-  if (rc == 0) {
-    close(fd);
-    if (uim_disconnect_cb) {
-      uim_disconnect_cb();
+  while(uim_helper_fd_readable(fd) > 0) {
+    
+    rc = read(fd, buf, sizeof(buf)-1);
+    buf[rc] = '\0';
+    
+    if(rc == 0) {
+      if (uim_disconnect_cb) {
+	uim_disconnect_cb();
+      }
+      uim_fd = -1;
+      return;
     }
-    uim_fd = -1;
-    return ;
+    uim_read_buf = (char *)realloc(uim_read_buf, strlen(uim_read_buf) + strlen(buf)+1);
+    strcat(uim_read_buf, buf);
   }
-  uim_read_buf_count += rc;
+  uim_read_buf_size = strlen(uim_read_buf);
+  return;
 }
 
 static void
 shift_read_buffer(int count)
 {
   memmove(uim_read_buf, &uim_read_buf[count],
-	  uim_read_buf_count - count);
-  uim_read_buf_count -= count;
+	  uim_read_buf_size - count);
+  uim_read_buf_size -= count;
+  uim_read_buf[uim_read_buf_size] = '\0';
 }
 
 char *
-uim_helper_get_message()
+uim_helper_get_message(void)
 {
   int i;
-  char buf[BUFFER_SIZE];
+  char *buf;
 
-  for (i = 0; i < uim_read_buf_count - 1; i++) {
+  for (i = 0; i < uim_read_buf_size - 1; i++) {
     if (uim_read_buf[i] == '\n' &&
 	uim_read_buf[i+1] == '\n') {
-      memcpy(buf, uim_read_buf, i+2);
-      buf[i+2] = 0;
+      buf = (char *)malloc(i+2);
+      memcpy(buf, uim_read_buf, i+1);
+      buf[i+1] = '\0';
       shift_read_buffer(i+2);
-      if (uim_read_buf_error) {
-	uim_read_buf_error = 0;
-	return NULL;
-      }
-      return strdup(buf);
+      return buf;
     }
   }
-  if (uim_read_buf_count == BUFFER_SIZE) {
-    uim_read_buf_count = 0;
-    uim_read_buf_error = 1;
-  }
   return NULL;
 }
