@@ -43,12 +43,13 @@
 #include <string.h>
 #include <pwd.h>
 #include <errno.h>
-#include <uim/context.h>
+#include "context.h"
+#include "uim-helper.h"
 
 enum RorW
   {
     READ,
-    WRITE,
+    WRITE
   };
 
 static int
@@ -115,7 +116,7 @@ uim_helper_get_pathname(void)
     login = strdup(pw->pw_name);
   }
 
-  path = (char*)malloc(strlen(login)+ 20);
+  path = (char *)malloc(strlen(login)+ 20);
   sprintf(path, "/tmp/uimhelper-%s",login);
   if (pw) {
     free(login);
@@ -123,12 +124,12 @@ uim_helper_get_pathname(void)
   return path;
 }
 
-int uim_helpler_fd_readable(int fd)
+int uim_helper_fd_readable(int fd)
 {
   return uim_helper_fd(fd, READ);
 }
 
-int uim_helpler_fd_writable(int fd)
+int uim_helper_fd_writable(int fd)
 {
   return uim_helper_fd(fd, WRITE);
 }
