@@ -33,6 +33,7 @@
 
 #include <signal.h>
 #include <stdlib.h>
+#include <sys/wait.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -40,8 +41,11 @@
 #include <stdio.h>
 #include <string.h>
 
+#include "context.h"
+#include "uim-helper.h"
+
 /* This function is come from the GNU C Library manual */
-int
+static int
 set_cloexec(int fd)
 {
   int oldflags = fcntl (fd, F_GETFD, 0);
@@ -136,23 +140,26 @@ open_pipe_rw(FILE **fr, FILE **fw)
 }
 
 int
-uim_ipc_open_command(int old_pid, FILE **read_handler,
-		     FILE **write_handler, char *command)
+uim_ipc_open_command(int old_pid, FILE **read_fp,
+		     FILE **write_fp, char *command)
 {
   int new_pid, result;
 
-  if (*read_handler  != NULL)
-    fclose(*read_handler);
-  if (*write_handler != NULL)
-    fclose(*write_handler);
+  if (*read_fp != NULL) {
+    fclose(*read_fp);
+  }
+  if (*write_fp != NULL) {
+    fclose(*write_fp);
+  }
   
-  *read_handler = *write_handler = NULL;
+  *read_fp = *write_fp = NULL;
   
   /* kill child process if exists */
-  if (old_pid)
+  if (old_pid) {
     kill(old_pid, SIGKILL);
+  }
 
-  new_pid = open_pipe_rw(read_handler, write_handler);
+  new_pid = open_pipe_rw(read_fp, write_fp);
 
   if (new_pid < 0)
     return 0;
@@ -180,27 +187,26 @@ uim_ipc_open_command(int old_pid, FILE **read_handler,
 
 }
 
-
-
 char *
-uim_ipc_send_command(int *pid, FILE **read_handler,
-		     FILE **write_handler, char* command, char *str)
+uim_ipc_send_command(int *pid,
+		     FILE **read_fp, FILE **write_fp,
+		     char *command, char *str)
 {
   char *tmp = strdup("");
   char buf[8192];
 
-  if (*read_handler == NULL || *write_handler == NULL) {
-    *pid = uim_ipc_open_command(*pid, read_handler, write_handler, command);
+  if (*read_fp == NULL || *write_fp == NULL) {
+    *pid = uim_ipc_open_command(*pid, read_fp, write_fp, command);
   }
   if(*pid == 0) {
     free(tmp);
     return NULL;
   }
 
-  fputs(str, *write_handler);
+  fputs(str, *write_fp);
 
  again:
-  if (fflush(*write_handler) != 0) {
+  if (fflush(*write_fp) != 0) {
     switch (errno) {
     case EINTR:
       goto again;
@@ -209,9 +215,19 @@ uim_ipc_send_command(int *pid, FILE **read_handler,
     }
   }
 
-  while (fgets (buf, sizeof(buf), *read_handler ) != NULL ) {
-    if(strcmp( buf, "\n" ) == 0)
+  if (feof(*read_fp)) {
+    fclose(*read_fp);
+    fclose(*write_fp);
+    *read_fp = NULL;
+    *write_fp = NULL;
+    free(tmp);
+    return NULL;
+  }
+
+  while (fgets (buf, sizeof(buf), *read_fp) != NULL) {
+    if(strcmp( buf, "\n" ) == 0) {
       break;
+    }
 
     tmp = realloc(tmp, strlen(tmp) + strlen(buf) + 1);
     strcat(tmp, buf);
@@ -221,6 +237,6 @@ uim_ipc_send_command(int *pid, FILE **read_handler,
 
  err:
   free(tmp);
-  *pid = uim_ipc_open_command(*pid, read_handler, write_handler, command );
+  *pid = uim_ipc_open_command(*pid, read_fp, write_fp, command);
   return NULL;
 }
