@@ -96,6 +96,20 @@ close_table(struct table_info *table)
   free(table);
 }
 
+static char *
+find_line_head(struct table_info *table, char *from)
+{
+  while(table->addr <= (void *)from){
+
+    if( *from == '\n') {
+      from++;
+      return from;
+    }
+    from--;
+  }
+  return from;
+}
+
 static int
 find_table(const char *fn)
 {
@@ -107,24 +121,10 @@ find_table(const char *fn)
   return -1;
 }
 
-char *
-find_line_head(struct table_info *table, char *from)
-{
-  while(table->addr <= (void*)from){
-
-    if( *from == '\n') {
-      from++;
-      return from;
-    }
-    from--;
-  }
-  return from;
-}
-
-char *
+static char *
 find_next_linefeed(struct table_info *table, char *from)
 {
-  while((table->addr + table->size) >= (void*)from){
+  while(((char *)(table->addr) + table->size) >= from){
 
     if( *from == '\n') {
       return from;
@@ -134,7 +134,7 @@ find_next_linefeed(struct table_info *table, char *from)
   return from;
 }
 
-char *
+static char *
 cut_off_line_from_table(struct table_info *table, char *line_head)
 {
   char *line_tail;
@@ -145,7 +145,7 @@ cut_off_line_from_table(struct table_info *table, char *line_head)
 
   line_size = sizeof(char) * (line_tail - line_head);
 
-  line = (char*)malloc( line_size + 1 );
+  line = (char *)malloc( line_size + 1 );
 
   memcpy(line, line_head, line_size);
 
@@ -167,7 +167,7 @@ find_matched_lines_from_table_internal(int table_id, char *str, int flag)
       char *tmp = cut_off_line_from_table(table,
 					  find_line_head(table, p));
       LISP res = strcons(strlen(tmp), tmp);
-      //      free(tmp);
+      /*      free(tmp); */
       return res;
     }
 
