@@ -1,14 +1,55 @@
+/*
+
+  Copyright (c) 2003,2004 uim Project http://uim.freedesktop.org/
+
+  All rights reserved.
+
+  Redistribution and use in source and binary forms, with or without
+  modification, are permitted provided that the following conditions
+  are met:
+
+  1. Redistributions of source code must retain the above copyright
+     notice, this list of conditions and the following disclaimer.
+  2. Redistributions in binary form must reproduce the above copyright
+     notice, this list of conditions and the following disclaimer in the
+     documentation and/or other materials provided with the distribution.
+  3. Neither the name of authors nor the names of its contributors
+     may be used to endorse or promote products derived from this software
+     without specific prior written permission.
+
+  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+  SUCH DAMAGE.
+
+*/
+
 #include <stdlib.h>
 #include <string.h>
+#include <locale.h>
+#include <langinfo.h>
 #include "context.h"
 #include "siod.h"
- 
+#include "config.h"
+#include "gettext.h"
+#include "uim-util.h"
+
+
 #define TRUEP(x) EQ(x, true_sym)
 #define FALSEP(x) EQ(x, false_sym)
  
 #define NTRUEP(x) NEQ(x, true_sym)
 #define NFALSEP(x) NEQ(x, false_sym)
 
+extern char *uim_return_str;
+
 static LISP true_sym;
 static LISP false_sym;
 static LISP quote_sym;
@@ -39,7 +80,7 @@ charcode2string(LISP x)
     buf[0] = 0;
   }
   buf[1] = 0;
-    return strcons (1,buf);
+  return strcons (1,buf);
 }
 
 static LISP
@@ -47,10 +88,10 @@ string2charcode(LISP x)
 {
   char *buf = get_c_string(x);
 
-  if(buf) {
+  if (buf) {
     return intcons(*buf);
   }
-  return NIL;
+  return false_sym;
 }
 
 static LISP
@@ -71,9 +112,9 @@ nthcdr(LISP nth_, LISP lst)
   int nth = get_c_int(nth_);
   int i;
   for (i = 0; i < nth; i++) {
-    if (!lst) {
+    if NULLP(lst) {
       /* something bad happened */
-      return NIL;
+      return false_sym;
     }
     lst = CDR(lst);
   }
@@ -145,42 +186,48 @@ str_seq_partial(LISP seq, LISP rule)
 }
 
 static LISP
-rk_find_seq(LISP s, LISP rule)
+rk_find_seq(LISP seq, LISP rules)
 {
-  for (; rule != NIL; rule = CDR(rule)) {
-    if (str_seq_equal(s, CAR(CAR(CAR(rule))))) {
-      return CAR(rule);
+  for (; NNULLP(rules); rules = CDR(rules)) {
+    LISP rule = CAR(rules);
+    LISP key = CAR(CAR(rule));
+    if NFALSEP(str_seq_equal(seq, key)) {
+      return rule;
     }
   }
-  return NIL;
+  return false_sym;
 }
 
 static LISP
-rk_find_partial_seq(LISP s, LISP rule)
+rk_find_partial_seq(LISP seq, LISP rules)
 {
-  for (; rule != NIL; rule = CDR(rule)) {
-    if NFALSEP(str_seq_partial(s, CAR(CAR(CAR(rule))))) {
-      return CAR(rule);
+  for (; NNULLP(rules); rules = CDR(rules)) {
+    LISP rule = CAR(rules);
+    LISP key = CAR(CAR(rule));
+    if NFALSEP(str_seq_partial(seq, key)) {
+      return rule;
     }
   }
-  return NIL;
+  return false_sym;
 }
 
+/*
+ * returns possible next characters
+ * (rk-lib-expect-seq '("k" "y") ja-rk-rule) -> ("o" "e" "u" "i" "a")
+ */
 static LISP
 rk_expect_seq(LISP seq, LISP rules)
 {
-  long flag;
   LISP cur, res = NIL;
-  flag = no_interrupt(1);
-  for (cur = rules; cur; cur = CDR(cur)) {
+  for (cur = rules; NNULLP(cur); cur = CDR(cur)) {
     LISP rule = CAR(cur);
-    LISP e = str_seq_partial(seq, CAR(CAR(rule)));
+    LISP key = CAR(CAR(rule));
+    LISP e = str_seq_partial(seq, key);
     if NFALSEP(e) {
       res = cons(e, res);
     }
   }
-  no_interrupt(flag);
-  return res;
+  return res;  /* don't return false_sym */
 }
 
 static LISP
@@ -190,10 +237,41 @@ c_getenv(LISP str_)
   char *val;
 
   if (!str) {
-    return NIL;
+    return false_sym;
   }
   val = getenv(str);
-  return strcons(strlen(val), val);
+  if (val) {
+    return strcons(strlen(val), val);
+  } else {
+    return false_sym;
+  }
+}
+
+static LISP
+c_setenv(LISP name_, LISP val_, LISP overwrite_)
+{
+  char *name = get_c_string(name_);
+  char *val = get_c_string(val_);
+  int overwrite = NFALSEP(overwrite_);
+  int err;
+
+  if (!name || !val) {
+    return false_sym;
+  }
+  err = setenv(name, val, overwrite);
+  return (err) ? false_sym : true_sym;
+}
+
+static LISP
+c_unsetenv(LISP name_)
+{
+  char *name = get_c_string(name_);
+
+  if (!name) {
+    return false_sym;
+  }
+  unsetenv(name);
+  return true_sym;
 }
 
 static char **
@@ -265,19 +343,19 @@ uim_split_string(LISP _splittee, LISP _splitter)
   int n_strs;
 
   if(_splittee == NULL || _splitter == NULL)
-    return NIL;
+    return false_sym;
 
   if(splittee == NULL || splitter == NULL)
-    return NIL;
+    return false_sym;
 
   strs = uim_strsplit(splittee, splitter);
 
   if(!strs || !*strs)
-    return NIL;
+    return false_sym;
 
   for (n_strs = 0; strs[n_strs] != '\0'; n_strs++);
 
-  l = uim_scm_c_strs_into_list(n_strs, strs);
+  l = uim_scm_c_strs_into_list(n_strs, (const char *const *)strs);
   for (i = n_strs - 1; i >= 0; i--) {
     free(strs[i]);
   }
@@ -289,7 +367,7 @@ static LISP
 eucjp_string_to_list(LISP str_)
 {
   char *str = get_c_string(str_);
-  unsigned char *cur = str;
+  unsigned char *cur = (unsigned char *)str;
   LISP res = NIL;
   while (*cur) {
     char buf[3];
@@ -306,7 +384,7 @@ eucjp_string_to_list(LISP str_)
       buf[1] = 0;
       len = 1;
     }
-    res = cons (strcons(len, buf), res);
+    res = cons (strcons(len, (char *)buf), res);
     cur ++;
   }
   return res;
@@ -329,7 +407,7 @@ uim_scm_symbol_value_int(const char *symbol_str)
   int val;
   val_ = uim_scm_symbol_value(symbol_str);
 
-  if( val_ != NIL) {
+  if NNULLP(val_) {
     val = uim_scm_c_int(val_);
   } else {
     val = 0;
@@ -354,7 +432,7 @@ uim_scm_symbol_value_str(const char *symbol_str)
   char *val;
   val_ = uim_scm_symbol_value(symbol_str);
 
-  if( val_ != NIL) {
+  if NNULLP(val_) {
     val = uim_scm_c_str(val_);
   } else {
     val = NULL;
@@ -394,10 +472,10 @@ uim_scm_symbol_value(const char *symbol_str)
 {
   LISP symbol_str_ = rintern(symbol_str);
   
-  if(symbol_boundp(symbol_str_, NIL) == true_sym) {
+  if TRUEP(symbol_boundp(symbol_str_, NIL)) {
     return (uim_lisp)symbol_value(symbol_str_, NIL);         
   } else {
-    return (uim_lisp)NIL;
+    return (uim_lisp)false_sym;
   }
 }
 
@@ -432,15 +510,10 @@ uim_scm_set_verbose_level(long new_value)
 void
 uim_scm_load_file(const char *fn)
 {
-  long want_init = 0, want_print = 0;
-  char *ln = malloc(strlen(fn)+40);
-
-  if(!fn || !ln)
+  if(!fn)
     return;
 
-  snprintf(ln, strlen(fn)+40, "(*catch 'errobj (load \"%s\" #f #f))", fn);
-  uim_scm_repl_c_string(ln, want_init, want_print);
-  free(ln);
+  UIM_EVAL_FSTRING1(NULL, "(*catch 'errobj (load \"%s\" #f #f))", fn);
 }
 
 uim_lisp
@@ -450,7 +523,7 @@ uim_scm_t(void) {
 
 uim_lisp
 uim_scm_f(void) {
-  return (uim_lisp)NIL;
+  return (uim_lisp)false_sym;
 }
 
 uim_lisp
@@ -475,7 +548,7 @@ uim_scm_string_equal(uim_lisp a, uim_lisp b) {
 		       a,
 		       b);
   p = uim_scm_eval(form);
-  return NFALSEP(p);
+  return TRUEP(p);
 }
 
 uim_lisp
@@ -570,6 +643,19 @@ uim_scm_list5(uim_lisp elm1, uim_lisp elm2, uim_lisp elm3, uim_lisp elm4,
   return lst;
 }
 
+uim_lisp
+uim_scm_reverse(uim_lisp cell)
+{
+  return (uim_lisp)reverse((LISP)cell);
+}
+
+uim_lisp
+uim_scm_nreverse(uim_lisp cell)
+{
+  return (uim_lisp)nreverse((LISP)cell);
+}
+
+
 /* Customize interface functions: They are not appropriate to be
  * here. More discussion is required.
  */
@@ -652,6 +738,8 @@ uim_custom_value_as_symbol(uim_lisp custom_sym) {
 		 uim_scm_intern_c_str("symbol")))
   {
     val = uim_custom_value(custom_sym);
+  } else {
+    val = uim_scm_f();
   }
 
   return val;
@@ -691,6 +779,26 @@ uim_custom_symbol_desc(uim_lisp custom_sym, uim_lisp val_sym) {
   return uim_scm_c_str(desc);
 }
 
+uim_lisp
+uim_custom_label(uim_lisp custom_sym) {
+  uim_lisp form;
+
+  form = uim_scm_list2(uim_scm_intern_c_str("custom-label"),
+		       uim_scm_quote(custom_sym));
+
+  return uim_scm_eval(form);
+}
+
+uim_lisp
+uim_custom_desc(uim_lisp custom_sym) {
+  uim_lisp form;
+
+  form = uim_scm_list2(uim_scm_intern_c_str("custom-desc"),
+		       uim_scm_quote(custom_sym));
+
+  return uim_scm_eval(form);
+}
+
 uim_lisp
 uim_custom_type(uim_lisp custom_sym) {
   uim_lisp form;
@@ -701,6 +809,16 @@ uim_custom_type(uim_lisp custom_sym) {
   return uim_scm_eval(form);
 }
 
+uim_lisp
+uim_custom_default_value(uim_lisp custom_sym) {
+  uim_lisp form;
+
+  form = uim_scm_list2(uim_scm_intern_c_str("custom-default-value"),
+		       uim_scm_quote(custom_sym));
+
+  return uim_scm_eval(form);
+}
+
 int
 uim_custom_ctype(uim_lisp custom_sym) {
   uim_lisp type;
@@ -731,6 +849,8 @@ uim_custom_ctype(uim_lisp custom_sym) {
 			uim_scm_intern_c_str("key")))
   {
     result = UCustom_Key;
+  } else {
+    result = UCustom_Bool;
   }
 
   return result;
@@ -832,6 +952,57 @@ uim_custom_definition_as_string(uim_lisp sym) {
   return uim_scm_c_str(definition);
 }
 
+
+/* Following is utility functions for C world */
+struct _locale_language_table {
+  char *locale;
+  char *language;
+};
+
+static struct _locale_language_table locale_language_table[] = {
+#include "iso-639-1.def"
+};
+#define NR_LOCALE_LANGUAGE \
+        (sizeof(locale_language_table) / sizeof(struct _locale_language_table))
+
+static const char *
+get_language_name_from_locale(const char *localename)
+{
+  unsigned int i;
+  for(i = 0; i < NR_LOCALE_LANGUAGE; i++) {
+    if(strcmp(locale_language_table[i].locale, localename) == 0) {
+      return locale_language_table[i].language;
+    }
+  }
+  return NULL;
+}
+
+const char *
+uim_get_language_name_from_locale(const char *localename)
+{
+#if 1
+  /* performs adhoc "zh_TW:zh_HK" style locale handling as temporary
+     specification of this function for backward compatibility
+  */
+  UIM_EVAL_FSTRING1(NULL, "(langgroup-primary-lang-code \"%s\")", localename);
+  localename = uim_return_str;  /* will be free() automatically */
+#endif
+  return get_language_name_from_locale(localename);
+}
+
+static LISP
+lang_code_to_lang_name_raw(LISP code_)
+{
+  const char *code = get_c_string(code_);
+  const char *name;
+  int unknown_strlen = -1;
+
+  if (!code)
+    return false_sym;
+  name = get_language_name_from_locale(code);
+  return (name) ? strcons(unknown_strlen, name) : false_sym;
+}
+
 void
 uim_init_util_subrs()
 {
@@ -856,6 +1027,9 @@ uim_init_util_subrs()
   init_subr_2("rk-lib-find-partial-seq", rk_find_partial_seq);
   init_subr_2("rk-lib-expect-seq", rk_expect_seq);
   init_subr_1("getenv", c_getenv);
+  init_subr_3("setenv", c_setenv);
+  init_subr_1("unsetenv", c_unsetenv);
   init_subr_2("string-split", uim_split_string);
   init_subr_1("string-to-list", eucjp_string_to_list);
+  init_subr_1("lang-code->lang-name-raw", lang_code_to_lang_name_raw);
 }
