@@ -39,8 +39,9 @@
 #include <string.h>
 #include <ctype.h>
 #include "config.h"
-#include <uim/uim.h>
-#include <uim/uim-scm.h>
+#include "uim.h"
+#include "uim-im-switcher.h"
+#include "uim-scm.h"
 #include "context.h"
 #include "gettext.h"
 
@@ -51,12 +52,7 @@ extern char *uim_return_str_list[10];
 #define UIM_RETURN_STR_LIST_SIZE ((sizeof(uim_return_str_list) \
                                    / sizeof(uim_return_str_list[0])) - 2)
 
-extern void uim_init_skk_dic();
-extern void uim_quit_skk_dic();
-extern void uim_init_anthy();
-extern void uim_quit_anthy();
-extern void uim_init_prime();
-extern void uim_quit_prime();
+char *uim_last_client_encoding;
 
 #ifdef HAVE_CANNA_RK_H
 extern void uim_init_canna();
@@ -100,22 +96,22 @@ uim_set_preedit_cb(uim_context uc,
 }
 
 static void
-get_context_id(uim_context is)
+get_context_id(uim_context uc)
 {
   int i;
   for (i = 0; i < CONTEXT_ARRAY_SIZE; i++) {
     if (!context_array[i]) {
-      context_array[i] = is;
-      is->id = i;
+      context_array[i] = uc;
+      uc->id = i;
       return;
     }
   }
 }
 
 static void
-put_context_id(uim_context is)
+put_context_id(uim_context uc)
 {
-  context_array[is->id] = NULL;
+  context_array[uc->id] = NULL;
 }
 
 uim_context
@@ -127,8 +123,6 @@ uim_create_context(void *ptr,
 		   void (*commit_cb)(void *ptr, const char *str))
 {
   uim_context uc;
-  int buf_len;
-  char *buf;
 
   if (!conv) {
     conv = uim_iconv;
@@ -149,6 +143,7 @@ uim_create_context(void *ptr,
   if (!enc) {
     enc = "UTF-8";
   }
+  uc->short_desc = NULL;
   uc->encoding = strdup(enc);
   uc->conv_if = conv;
   uc->conv = NULL;
@@ -176,58 +171,106 @@ uim_create_context(void *ptr,
   uc->candidate_selector_shift_page_cb = NULL;
   uc->candidate_selector_deactivate_cb = NULL;
   /**/
+  uc->request_surrounding_text_cb = NULL;
+  uc->delete_surrounding_text_cb = NULL;
+  /**/
   uc->nr_candidates = 0;
   uc->candidate_index = 0;
   /**/
   uc->cb_q.first_cb = NULL;
   uc->cb_q.tail_cb = NULL;
   uc->cb_q.flushing = 0;
+  /**/
+  uc->psegs = NULL;
+  uc->nr_psegs = 0;
 
   if (!lang) {
-    lang = "()";
+    lang = "#f";
   }
   if (!engine) {
-    engine = "()";
+    engine = "#f";
+    uc->current_im_name = NULL;
+  } else {
+    uc->current_im_name = strdup(engine);
+  }
+  if (uim_last_client_encoding) {
+    free(uim_last_client_encoding);
+    uim_last_client_encoding = NULL;
+  }
+  if (enc) {
+    /* Save client encoding for bind_textdomain_codeset(). The value
+       is used when UIM_EVAL_STRING() is called without
+       uim_context. The specification assumes client always uses
+       consistent encoding
+    */
+    uim_last_client_encoding = strdup(enc);
   }
-  buf_len = strlen(lang) + strlen(engine) + 40;
-  buf = malloc(buf_len);
-  snprintf(buf, buf_len - 1,
-	   "(create-context %d '%s '%s)", uc->id, lang, engine);
-  uim_eval_string(uc, buf);
 
-  free(buf);
+  UIM_EVAL_FSTRING3(uc, "(create-context %d '%s '%s)", uc->id, lang, engine);
 
   return uc;
 }
 
 void
 uim_reset_context(uim_context uc)
- {
-   char buf[30];
-   snprintf(buf, 29, "(reset-handler %d)", uc->id);
-   uim_eval_string(uc, buf);
+{
+   UIM_EVAL_FSTRING1(uc, "(reset-handler %d)", uc->id);
+
+   /* delete all preedit segments */
+   uim_release_preedit_segments(uc);
+}
+
+void
+uim_switch_im(uim_context uc, const char *engine) {
+  /* related to the commit log of r1400:
+
+     We should not add the API uim_destroy_context(). We should move
+     IM-switching feature into Scheme instead. It removes the context
+     management code related to IM-switching duplicated in several IM
+     bridges. Refer the implementation of GtkIMMulticontext as example
+     of proxy-like IM-switcher. It does IM-switching without extending
+     immodule API. We should follow its design to make our API simple.
+     -- 2004-10-05 YamaKen
+  */
+  uim_reset_context(uc);
+  UIM_EVAL_FSTRING2(uc, "(create-context %d #f '%s)", uc->id, engine);
+  if (uc->current_im_name)
+    free(uc->current_im_name);
+  uc->current_im_name = strdup(engine);
+#if 0
+  /*
+    I don't know when uc->short_desc is used, so this part is
+    disabled. -- 2004-10-04 YamaKen
+  */
+  if (uc->short_desc)
+    free(uc->short_desc);
+  UIM_EVAL_FSTRING1(uc, "(uim-get-im-short-desc '%s)", im->name);
+  uc->short_desc = uim_return_str;
+  uim_return_str = NULL;  /* ownership has been transferred */
+#endif
 }
 
 void
 uim_release_context(uim_context uc)
 {
-  char buf[30];
   int i;
 
   if(!uc)
     return;
 
-  snprintf(buf, 29, "(release-context %d)", uc->id);
-  uim_eval_string(uc, buf);
+  UIM_EVAL_FSTRING1(uc, "(release-context %d)", uc->id);
   put_context_id(uc);
   if (uc->conv) {
     uc->conv_if->release(uc->conv);
   }
+  uim_release_preedit_segments(uc);
   for (i = 0; i < uc->nr_modes; i++) {
     free(uc->modes[i]);
   }
   free(uc->propstr);
+  free(uc->proplabelstr);
   free(uc->modes);
+  free(uc->short_desc);
   free(uc->encoding);
   free(uc);
 }
@@ -247,7 +290,7 @@ uim_get_nr_modes(uim_context uc)
 const char *
 uim_get_mode_name(uim_context uc, int nth)
 {
-  if (nth < uc->nr_modes) {
+  if (nth >= 0 && nth < uc->nr_modes) {
     return uc->modes[nth];
   }
   return NULL;
@@ -280,13 +323,11 @@ uim_set_prop_label_update_cb(uim_context uc,
 void
 uim_prop_activate(uim_context uc, const char *str)
 {
-  char buf[100];
-  
   if(!str)
     return;
       
-  snprintf(buf, 99, "(prop-handler %d 'prop-activate \"%s\")", uc->id, str);
-  uim_eval_string(uc, buf);
+  UIM_EVAL_FSTRING2(uc, "(prop-handler %d 'prop-activate \"%s\")",
+		    uc->id, str);
 }
 
 /** Update custom value from property message.
@@ -315,10 +356,8 @@ uim_get_current_mode(uim_context uc)
 void
 uim_set_mode(uim_context uc, int mode)
 {
-  char buf[100];
   uc->mode = mode;
-  snprintf(buf, 99, "(mode-handler %d %d)", uc->id, mode);
-  uim_eval_string(uc, buf);
+  UIM_EVAL_FSTRING2(uc, "(mode-handler %d %d)", uc->id, mode);
 }
 
 void
@@ -373,6 +412,15 @@ get_nth_im(uim_context uc, int nth)
   return NULL;
 }
 
+const char *
+uim_get_current_im_name(uim_context uc)
+{
+  if(uc) {
+    return uc->current_im_name;
+  }
+  return NULL;
+}
+
 const char *
 uim_get_im_name(uim_context uc, int nth)
 {
@@ -395,6 +443,22 @@ uim_get_im_language(uim_context uc, int nth)
   return NULL;
 }
 
+const char *
+uim_get_im_short_desc(uim_context uc, int nth)
+{
+  struct uim_im *im = get_nth_im(uc, nth);
+
+  if (im) {
+    if (im->short_desc)
+      free(im->short_desc);
+    UIM_EVAL_FSTRING1(uc, "(uim-get-im-short-desc '%s)", im->name);
+    im->short_desc = uim_return_str;
+    uim_return_str = NULL;  /* ownership has been transferred */
+    return im->short_desc;
+  }
+  return NULL;
+}
+
 const char *
 uim_get_im_encoding(uim_context uc, int nth)
 {
@@ -406,11 +470,50 @@ uim_get_im_encoding(uim_context uc, int nth)
   return NULL;
 }
 
+static const char *
+uim_check_im_exist(const char *im_engine_name)
+{
+  int i;
+  for(i = 0; i < uim_nr_im; i++) {
+    struct uim_im *im = &uim_im_array[i];
+    if (strcmp(im_engine_name, im->name) == 0) {
+      return im->name;
+    }
+  }
+  return NULL;
+}
+
+const char *
+uim_get_default_im_name(const char *localename)
+{
+  const char *valid_default_im_name;
+  UIM_EVAL_FSTRING1(NULL, "(uim-get-default-im-name \"%s\")", localename);
+  valid_default_im_name = uim_check_im_exist(uim_return_str);
+
+  if (!valid_default_im_name)
+    valid_default_im_name = "default";  /* never happen */
+  return valid_default_im_name;
+}
+
+const char *
+uim_get_im_name_for_locale(const char *localename)
+{
+  const char *valid_im_name;
+  UIM_EVAL_FSTRING1(NULL, "(uim-get-im-name-for-locale \"%s\")", localename);
+  valid_im_name = uim_check_im_exist(uim_return_str);
+
+  if (!valid_im_name)
+    valid_im_name = "default";  /* never happen */
+  return valid_im_name;
+}
+
+
 static int
 load_conf()
 {
   struct passwd *pw;
   char *fn, *cmd;
+  int cmd_size;
   FILE *fp;
   long svl = siod_verbose_level;
   long ret;
@@ -419,7 +522,7 @@ load_conf()
 
   if (!fn) {
     pw = getpwuid(getuid());
-    fn = malloc(strlen(pw->pw_dir)+10);
+    fn = malloc(strlen(pw->pw_dir) + sizeof("/.uim"));
     sprintf(fn, "%s/.uim", pw->pw_dir);
   } else {
     fn = strdup(fn);
@@ -434,7 +537,8 @@ load_conf()
   if (siod_verbose_level < 1) {
     siod_verbose_level = 1;	/* to show error message of SIOD */
   }
-  cmd = malloc(strlen(fn) + 40);
+  cmd_size = uim_sizeof_sexp_str("(*catch 'errobj (load \"%s\" #f #f))", fn);
+  cmd = malloc(cmd_size);
   sprintf(cmd, "(*catch 'errobj (load \"%s\" #f #f))", fn);
   ret = repl_c_string(cmd, 0, 0);
   siod_verbose_level = svl;
@@ -460,10 +564,9 @@ int uim_set_candidate_selector_cb(uim_context uc,
 uim_candidate
 uim_get_candidate(uim_context uc, int index, int accel_enumeration_hint)
 {
-  char buf[100];
   uim_candidate cand = malloc(sizeof(*cand));
-  snprintf(buf, 99, "(get-candidate %d %d %d)", uc->id, index, accel_enumeration_hint);
-  uim_eval_string(uc, buf);
+  UIM_EVAL_FSTRING3(uc, "(get-candidate %d %d %d)",
+		    uc->id, index, accel_enumeration_hint);
 
   if (uim_return_str_list[0] && uim_return_str_list[1]) {
     cand->str = uc->conv_if->convert(uc->conv, uim_return_str_list[0]);
@@ -505,12 +608,25 @@ int uim_get_candidate_index(uim_context uc)
 void
 uim_set_candidate_index(uim_context uc, int nth)
 {
-  char buf[100];
-  snprintf(buf, 99, "(set-candidate-index %d %d)", uc->id, nth);
-  uim_eval_string(uc, buf);
+  UIM_EVAL_FSTRING2(uc, "(set-candidate-index %d %d)", uc->id, nth);
   return ;
 }
 
+void
+uim_set_surrounding_text_cb(uim_context uc,
+			    void (*request_cb)(void *ptr),
+			    int (*delete_cb)(void *ptr, int offset, int len))
+{
+  uc->request_surrounding_text_cb = request_cb;
+  uc->delete_surrounding_text_cb = delete_cb;
+}
+
+void
+uim_set_surrounding_text(uim_context uc, const char *text,
+			 int cursor_pos, int len)
+{
+}
+
 static void
 exit_hook(void)
 {
@@ -525,11 +641,11 @@ uim_init_scm()
   char *siod_argv[] =
     {
       "siod",
-      "-h100000:10",
-      "-o1000",
-      "-s200000",
-      "-n128",
-      "-v0",
+      "-h100000:10",  /* heap_size(unit: lisp objects):nheaps */
+      "-o1000",       /* obarray_dim */
+      "-s200000",     /* stack_size (unit: bytes) */
+      "-n128",        /* inums_dim (preallocated fixnum objects) */
+      "-v0",          /* siod_verbose_level */
     };
   char verbose_argv[4] = "-v4";
   char *env;
@@ -570,16 +686,14 @@ uim_init_scm()
 
   uim_scm_load_file("im.scm");
   uim_scm_load_file("loader.scm");
+  uim_scm_load_file("direct.scm");  /* must be loaded at last of IMs */
   if (getenv("LIBUIM_VANILLA") ||
       load_conf() == -1) {
     uim_scm_load_file("default.scm");
   }
-  if (getenv("LIBUIM_VERBOSE")) {
-    siod_verbose_level = 10;
-  }
 
   uim_return_str = NULL;
-  for (i = 0; i < UIM_RETURN_STR_LIST_SIZE; i++){
+  for (i = 0; i < (int)UIM_RETURN_STR_LIST_SIZE; i++){
     uim_return_str_list[i] = NULL;
   }
 }
@@ -593,6 +707,7 @@ uim_init(void)
   if (!uim_output) {
     uim_output = stderr;
   }
+  uim_last_client_encoding = NULL;
   uim_im_array = NULL;
   uim_nr_im = 0;
   uim_init_scm();
@@ -626,6 +741,7 @@ uim_quit(void)
   uim_quit_m17nlib();
 #endif /* HAVE_M17NLIB */
   siod_quit();
+  uim_last_client_encoding = NULL;
   uim_output = NULL;
   uim_initialized = 0;
 }
