@@ -36,34 +36,66 @@
 #include <X11/Xutil.h>
 #include <X11/extensions/shape.h>
 #include <stdlib.h>
+#include <locale.h>
 #include "xim.h"
 #include "ximserver.h"
 #include "convdisp.h"
 #include "canddisp.h"
 #include "xdispatch.h"
 
+#define UNDERLINE_HEIGHT	2
+// Temporal hack for flashplayer plugin's broken over-the-spot XIM style
+#define FLASHPLAYER_WORKAROUND
+
 //
-// PeWin: Base class for preedit window
-//  PeLineWin  root window style
-//  PeOvWin  over the spot
+// PeWin:     Base class for preedit window
+// PeLineWin: Child class for root window style preedit window
+// PeOvWin:   Child class for over the spot style preedit window
 
-static XFontSet gFontset;
+#ifdef FLASHPLAYER_WORKAROUND
+static Window getTopWindow(Display *, Window);
+#endif
 
-void init_convdisp()
-{
-    char **missing, *def;
-    int nr_missing;
-    gFontset = 
-	XCreateFontSet(XimServer::gDpy,
-		       "-misc-fixed-medium-r-normal--14-110-100-100-c-70-iso8859-1,"
-		       "-misc-fixed-medium-r-normal--14-130-75-75-c-140-jisx0208.1983-0,"
-		       "-misc-fixed-medium-r-normal--14-130-75-75-c-70-jisx0201.1976-0",
-		       &missing, &nr_missing, &def);
-    // same fontset as kterm
+const char *fontset_zhCN = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -isas-fangsong ti-medium-r-normal--16-160-72-72-c-160-gb2312.1980-0";
+const char *fontset_zhTW = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -taipei-fixed-medium-r-normal--16-150-75-75-c-160-big5-0";
+const char *fontset_ja = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -jis-fixed-medium-r-normal--16-*-75-75-c-160-jisx0208.1983-0, -sony-fixed-medium-r-normal--16-*-*-*-c-80-jisx0201.1976-0";
+const char *fontset_ko = "-sony-fixed-medium-r-normal--16-*-*-*-c-80-iso8859-1, -daewoo-gothic-medium-r-normal--16-120-100-100-c-160-ksc5601.1987-0";
 
-    if (missing) {
-	XFreeStringList(missing);
-    }
+
+static XFontSet
+create_default_fontset(const char *im_lang, const char *encoding, const char *locale) {
+    char *orig_locale;
+    const char *name;
+    XFontSet ret;
+
+    orig_locale = strdup(setlocale(LC_CTYPE, NULL));
+
+    if (strcmp(locale, orig_locale))
+	setlocale(LC_CTYPE, locale);
+
+    if (!strcmp(im_lang, "ja"))
+	name = fontset_ja;
+    else if (!strcmp(im_lang, "ko"))
+	name = fontset_ko;
+    else if (!strcmp(im_lang, "zh_CN"))
+	name = fontset_zhCN;
+    else if (!strcmp(im_lang, "zh_TW:zh_HK"))
+	name = fontset_zhTW;
+    else
+	name = fontset_ja; // XXX fallback fontset
+
+    ret = get_font_set(name, locale);
+
+    if (strcmp(locale, orig_locale))
+	setlocale(LC_CTYPE, orig_locale);
+    free(orig_locale);
+
+    return ret;
+}
+
+static XFontSet
+choose_default_fontset(const char *im_lang, const char *encoding, const char *locale) {
+    return create_default_fontset(im_lang, encoding, locale);
 }
 
 struct char_ent {
@@ -77,21 +109,21 @@ struct char_ent {
 // Preedit Window
 class PeWin : public WindowIf {
 public:
-    PeWin(Window pw);
+    PeWin(Window pw, const char *im_lang, const char *encoding, const char *locale);
     virtual ~PeWin();
 
     virtual void release();
     
-    virtual void destroy(Window);
-    void expose(Window w);
+    virtual void destroy(Window w);
+    virtual void expose(Window w);
 
     void draw_char(int x, int y, uchar ch, int stat);
     void set_back(unsigned long p);
     void set_fore(unsigned long p);
     void set_fontset(XFontSet f);
     
-    virtual void set_size(int w,int h);
-    void set_pos(int x,int y);
+    virtual void set_size(int w, int h);
+    void set_pos(int x, int y);
 
     void do_map();
 
@@ -102,18 +134,19 @@ protected:
     Window mParentWin;
     Window mWin;
     Pixmap mPixmap;
-    GC mGC,mClearGC;
+    GC mGC, mClearGC;
     GC mHilitGC;
-    unsigned int mFore,mBack;
+    unsigned int mFore, mBack;
     XFontSet mFontset;
-    int mWidth,mHeight;
+    const char *mEncoding;
+    int mWidth, mHeight;
     bool mIsMapped;
 };
 
 // one line preedit window for RootWindowStyle
 class PeLineWin : public PeWin {
 public:
-    PeLineWin(Window w);
+    PeLineWin(Window w, const char *im_lang, const char *encoding, const char *locale);
     virtual ~PeLineWin();
 
     void draw_pe(pe_stat *p);
@@ -129,10 +162,10 @@ private:
 // window for over the spot style
 class PeOvWin : public PeWin {
 public:
-    PeOvWin(Window w);
+    PeOvWin(Window w, const char *im_lang, const char *encoding, const char *locale);
     
-    void draw_ce(char_ent *ce,int len);
-    virtual void set_size(int w,int h);
+    void draw_ce(char_ent *ce, int len);
+    virtual void set_size(int w, int h);
     virtual ~PeOvWin();
 private:
     void draw_a_ce(char_ent *ce);
@@ -140,14 +173,15 @@ private:
     GC m_mask_pix_gc;
 };
 
-class ConvdispOv : public Convdisp{
+class ConvdispOv : public Convdisp {
 public:
-    ConvdispOv(InputContext *,icxatr *);
+    ConvdispOv(InputContext *, icxatr *);
     virtual ~ConvdispOv();
-    void update_preedit();
+    virtual void update_preedit();
+    virtual void clear_preedit();
     virtual void update_icxatr();
-    virtual void set_focus();
-    virtual void unset_focus();
+    virtual void move_candwin();
+    virtual void set_im_lang(const char *im_lang);
 private:
     bool check_win();
     bool check_atr();
@@ -155,43 +189,51 @@ private:
     void make_ce_array();
     void draw_preedit();
     void do_draw_preedit();
-    int calc_ce_width(int b,int e);
+    int calc_ce_width(int b, int e);
+#ifdef FLASHPLAYER_WORKAROUND
+    int get_ce_font_height(char_ent *ce, int len);
+    int revised_spot_y;
+#endif
 
     void validate_area();
 
     char_ent *m_ce;
     int m_ce_len;
     PeOvWin *m_ov_win;
+    XFontSet m_initial_fontset;
+    char *m_initial_lang;
+    bool m_lang_changed;
 };
 
 // Preedit window for RootWindowStyle
-class ConvdispRw : public Convdisp{
+class ConvdispRw : public Convdisp {
 public:
-    ConvdispRw(InputContext *,icxatr *);
+    ConvdispRw(InputContext *, icxatr *);
     virtual ~ConvdispRw();
 
-    void update_preedit();
+    virtual void update_preedit();
+    virtual void clear_preedit();
     virtual void update_icxatr();
+    virtual void move_candwin();
 private:
     PeLineWin *mPeWin;
 };
 
-class Connection;
-class ConvdispOs : public Convdisp{
+class ConvdispOs : public Convdisp {
 public:
-    ConvdispOs(InputContext *,icxatr *,Connection *);
+    ConvdispOs(InputContext *, icxatr *, Connection *);
     virtual ~ConvdispOs();
     virtual void update_preedit();
+    virtual void clear_preedit();
     virtual void update_icxatr();
-    virtual void set_focus();
-    virtual void unset_focus();
+    virtual void move_candwin();
 
 private:
     void compose_preedit_array(TxPacket *);
     void compose_feedback_array(TxPacket *);
   
     Connection *mConn;
-    int mImid,mIcid;
+    int mImid, mIcid;
     int mPrevLen;
 };
 
@@ -214,7 +256,7 @@ Convdisp *create_convdisp(int style, InputContext *k,
 //
 // PeWin(PreEdit Window)
 //
-PeWin::PeWin(Window pw)
+PeWin::PeWin(Window pw, const char *im_lang, const char *encoding, const char *locale)
 {
     mParentWin = pw;
     int scr_num = DefaultScreen(XimServer::gDpy);
@@ -243,17 +285,18 @@ PeWin::PeWin(Window pw)
     XSetBackground(XimServer::gDpy, mClearGC, BlackPixel(XimServer::gDpy, scr_num));
 
     add_window_watch(mWin, this, EXPOSE_MASK|STRUCTURE_NOTIFY_MASK);
-    mIsMapped = false;//not mapped now
-    mFontset = gFontset;
+    mIsMapped = false; //not mapped now
+    
+    mFontset = choose_default_fontset(im_lang, encoding, locale);
+    mEncoding = encoding;
     
     XFlush(XimServer::gDpy);
 }
 
 PeWin::~PeWin()
 {
-    if (mWin) {
-        release();
-    }
+    if (mWin)
+	release();
 
     XFreePixmap(XimServer::gDpy, mPixmap);
     
@@ -273,15 +316,14 @@ void PeWin::release()
 
 void PeWin::destroy(Window w)
 {
-    if (mWin && mWin == w) {
-        mWin = 0;
-    }
+    if (mWin && mWin == w)
+	mWin = 0;
 }
 
 void PeWin::expose(Window w)
 {
     XCopyArea(XimServer::gDpy, mPixmap, mWin, mGC,
-              0, 0, mWidth, mHeight, 0, 0);
+	      0, 0, mWidth, mHeight, 0, 0);
 
     XFlush(XimServer::gDpy);
 }
@@ -289,24 +331,30 @@ void PeWin::expose(Window w)
 void PeWin::draw_char(int x, int y, uchar ch, int stat)
 {
     GC gc = mGC;
-    if (stat & PE_REVERSE) {
+    if (stat & PE_REVERSE)
 	gc = mClearGC;
-    }
-    if (stat & PE_HILIGHT) {
-        //gc = mHilitGC;
-    }
-    
-    if (ch < 256) {
-	char c = ch;
-	XmbDrawImageString(XimServer::gDpy, mPixmap,
-			   mFontset, gc, x, y, &c, 1);
-    } else {
-	unsigned char c[2];
-	c[0] = ((ch >>8)&255)|0x80;
-	c[1] = (ch & 255)|0x80;
 
+#if 0
+    if (stat & PE_HILIGHT)
+	gc = mHilitGC;
+#endif
+    char utf8[6];
+    int len = utf8_wctomb((unsigned char *)utf8, ch);
+    utf8[len] = '\0';
+
+    if (!strcmp(mEncoding, "UTF-8"))
+    	XwcDrawImageString(XimServer::gDpy, mPixmap, mFontset,
+			gc, x, y, &ch, 1);
+    else {
+	char *native_str;
+	
+	native_str = utf8_to_native_str(utf8, mEncoding);
+	if (!native_str)
+	    return;
+	int len = strlen(native_str);
 	XmbDrawImageString(XimServer::gDpy, mPixmap, mFontset,
-			   gc, x, y, (char *)c, 2);
+			   gc, x, y, native_str, len);
+	free(native_str);
     }
 }
 
@@ -331,13 +379,13 @@ void PeWin::set_fontset(XFontSet f)
 
 void PeWin::set_size(int w, int h)
 {
-    if (w == mWidth && h == mHeight) {
-        return ;
-    }
+    if (w == mWidth && h == mHeight)
+	return;
+
     XResizeWindow(XimServer::gDpy, mWin, w, h);
     XFreePixmap(XimServer::gDpy, mPixmap);
-    mPixmap =XCreatePixmap(XimServer::gDpy, DefaultRootWindow(XimServer::gDpy), w, h,
-                            DefaultDepth(XimServer::gDpy, DefaultScreen(XimServer::gDpy)));
+    mPixmap = XCreatePixmap(XimServer::gDpy, DefaultRootWindow(XimServer::gDpy), w, h,
+			    DefaultDepth(XimServer::gDpy, DefaultScreen(XimServer::gDpy)));
     mWidth = w;
     mHeight = h;
     clear();
@@ -354,9 +402,9 @@ void PeWin::set_pos(int x, int y)
 void PeWin::do_map()
 {
     if (!mIsMapped) {
-        XFlush(XimServer::gDpy);
-        XMapRaised(XimServer::gDpy, mWin);
-        mIsMapped = true;
+	XFlush(XimServer::gDpy);
+	XMapRaised(XimServer::gDpy, mWin);
+	mIsMapped = true;
     }
 }
 
@@ -368,17 +416,16 @@ void PeWin::clear()
 
 void PeWin::draw()
 {
-    if (!mIsMapped) {
-        do_map();
-    } else {
-        expose(mWin);
-    }
+    if (!mIsMapped)
+	do_map();
+    else
+	expose(mWin);
 }
 
 void PeWin::unmap()
 {
     if (mIsMapped) {
-        XUnmapWindow(XimServer::gDpy, mWin);
+	XUnmapWindow(XimServer::gDpy, mWin);
 	XFlush(XimServer::gDpy);
 	mIsMapped = false;
     }
@@ -387,9 +434,9 @@ void PeWin::unmap()
 //
 // PeLineWin
 //
-PeLineWin::PeLineWin(Window w) : PeWin(w)
+PeLineWin::PeLineWin(Window w, const char *im_lang, const char *encoding, const char *locale) : PeWin(w, im_lang, encoding, locale)
 {
-    set_size(400, 40);
+    set_size(400, 28); // set window height wider than its font height 16
     clear();
 }
 
@@ -404,7 +451,7 @@ void PeLineWin::draw_pe(pe_stat *p)
     m_x = 0;
     std::list<pe_ustring>::iterator i;
     for (i = p->ustrings.begin(); i != p->ustrings.end(); i++) {
-        draw_segment(&(*i));
+	draw_segment(&(*i));
     }
 }
 
@@ -412,9 +459,15 @@ void PeLineWin::draw_segment(pe_ustring *s)
 {
     uString::iterator i;
     for (i = s->s.begin(); i != s->s.end(); i++) {
-        uchar ch = *i;
-        draw_char(m_x, 20, ch, s->stat);
-        m_x += 20;
+	uchar ch = *i;
+	draw_char(m_x, 20, ch, s->stat);
+	if (s->stat & PE_UNDERLINE) {
+	    XDrawLine(XimServer::gDpy, mPixmap, mGC,
+			    m_x, 20 + UNDERLINE_HEIGHT,
+			    m_x + 16, 20 + UNDERLINE_HEIGHT);
+
+	}
+	m_x += 16; // XXX font width
     }
 }
 
@@ -422,14 +475,14 @@ void PeLineWin::calc_extent(pe_stat *p)
 {
     int c;
     c = p->get_char_count();
-    set_size(400, 40);
+    set_size(400, 28); // XXX need to extent
 }
 
 
 //
 // PeOvWin
 //
-PeOvWin::PeOvWin(Window w) : PeWin(w)
+PeOvWin::PeOvWin(Window w, const char *im_lang, const char *encoding, const char *locale) : PeWin(w, im_lang, encoding, locale)
 {
     m_mask_pix = 0;
     m_mask_pix_gc = 0;
@@ -438,21 +491,21 @@ PeOvWin::PeOvWin(Window w) : PeWin(w)
 PeOvWin::~PeOvWin()
 {
     if (m_mask_pix) {
-        XFreePixmap(XimServer::gDpy, m_mask_pix);
-        XFreeGC(XimServer::gDpy, m_mask_pix_gc);
+	XFreePixmap(XimServer::gDpy, m_mask_pix);
+	XFreeGC(XimServer::gDpy, m_mask_pix_gc);
     }
 }
 
 void PeOvWin::set_size(int w, int h)
 {
-    if (w == mWidth && h == mHeight) {
-        return ;
-    }
+    if (w == mWidth && h == mHeight)
+	return;
+
     PeWin::set_size(w, h);
     if (m_mask_pix) {
-        XFreePixmap(XimServer::gDpy, m_mask_pix);
-        m_mask_pix = XCreatePixmap(XimServer::gDpy, mWin,
-                                   mWidth, mHeight, 1);
+	XFreePixmap(XimServer::gDpy, m_mask_pix);
+	m_mask_pix = XCreatePixmap(XimServer::gDpy, mWin,
+				   mWidth, mHeight, 1);
     }
 }
 
@@ -486,11 +539,12 @@ void PeOvWin::draw_a_ce(char_ent *ce)
 {
     draw_char(ce->x, ce->y, ce->c, ce->stat);
     XFillRectangle(XimServer::gDpy, m_mask_pix, m_mask_pix_gc,
-                   ce->x, ce->y - ce->height + 2,
-                   ce->width, ce->height+1);
+		   ce->x, ce->y - ce->height + 2,
+		   ce->width, ce->height + UNDERLINE_HEIGHT - 1);
     if (ce->stat & PE_UNDERLINE) {
-        XDrawLine(XimServer::gDpy, mPixmap, mGC, ce->x, ce->y+2,
-		  ce->x+ce->width, ce->y + 2);
+	XDrawLine(XimServer::gDpy, mPixmap, mGC,
+		  ce->x, ce->y + UNDERLINE_HEIGHT,
+		  ce->x + ce->width, ce->y + UNDERLINE_HEIGHT);
     }
 }
 
@@ -501,135 +555,266 @@ Convdisp::Convdisp(InputContext *k, icxatr *a)
 {
     mKkContext = k;
     m_atr = a;
+    mIMLang = get_im_lang_from_engine(k->get_engine_name());
+    mEncoding = k->get_ic()->get_encoding();
+    mLocaleName = k->get_locale_name();
 }
 
 Convdisp::~Convdisp()
 {
 }
 
+void Convdisp::set_im_lang(const char *im_lang)
+{
+    mIMLang = im_lang;
+}
+
+void Convdisp::set_locale_name(const char *locale)
+{
+    mLocaleName = locale;
+}
+
 void Convdisp::set_pe(pe_stat *p)
 {
     m_pe = p;
 }
 
+uString Convdisp::get_pe()
+{
+    uString s;
+    std::list<pe_ustring>::iterator it;
+    for (it = m_pe->ustrings.begin(); it != m_pe->ustrings.end(); it++) {
+	append_ustring(&s, &(*it).s);
+    }
+    return s;
+}
+
+void Convdisp::set_focus()
+{
+    Canddisp *disp = canddisp_singleton();
+    disp->show();
+}
+
+void Convdisp::unset_focus()
+{
+    Canddisp *disp = canddisp_singleton();
+    disp->hide();
+}
+
+
+// Root window style
 ConvdispRw::ConvdispRw(InputContext *k, icxatr *a) : Convdisp(k, a)
 {
-    mPeWin = new PeLineWin(DefaultRootWindow(XimServer::gDpy));
+    mPeWin = NULL;
 }
 
 ConvdispRw::~ConvdispRw()
 {
-    delete mPeWin;
+    if (mPeWin)
+	delete mPeWin;
 }
 
 void ConvdispRw::update_preedit()
 {
-    if (m_pe && m_pe->get_char_count()) {
-	// preedit string exists
-	mPeWin->do_map();
-	mPeWin->draw_pe(m_pe);
-	mPeWin->draw();
-    } else {
-	mPeWin->clear();
-	mPeWin->expose(0);
-	XFlush(XimServer::gDpy);
+    if (!m_pe)
+	return; 
+
+    if (!m_pe->get_char_count()) {
+	clear_preedit();
+	return;
     }
+
+    // preedit string exists
+    if (!mPeWin)
+	mPeWin = new PeLineWin(DefaultRootWindow(XimServer::gDpy), mIMLang, mEncoding, mLocaleName);
+    
+    if (m_atr->has_atr(ICA_ClientWindow)) {
+    	int x, y;
+	Window win;
+	XWindowAttributes xattr;
+
+	XGetWindowAttributes(XimServer::gDpy, m_atr->client_window, &xattr);
+	XTranslateCoordinates(XimServer::gDpy, m_atr->client_window, DefaultRootWindow(XimServer::gDpy), 0, 0, &x, &y, &win);
+	mPeWin->set_pos(x, y + xattr.height);
+    }
+
+    mPeWin->do_map();
+    mPeWin->draw_pe(m_pe);
+    mPeWin->draw();
+
+    move_candwin();
+}
+
+void ConvdispRw::clear_preedit()
+{
+    delete mPeWin;
+    mPeWin = NULL;
 }
 
 void ConvdispRw::update_icxatr()
 {
 }
 
-//Over the spot style
+void ConvdispRw::move_candwin()
+{
+    if (m_atr->has_atr(ICA_ClientWindow)) {
+	int x, y;
+	Window win;
+	XWindowAttributes xattr;
+
+	XTranslateCoordinates(XimServer::gDpy, m_atr->client_window,
+			      DefaultRootWindow(XimServer::gDpy),
+			      0, 0, &x, &y, &win);
+
+	Canddisp *disp = canddisp_singleton();
+
+	XGetWindowAttributes(XimServer::gDpy, m_atr->client_window, &xattr);
+	disp->move(x, y + xattr.height + 28); // lower-left side under the preedit window
+    }
+}
+
+// Over the spot style
 ConvdispOv::ConvdispOv(InputContext *k, icxatr *a) : Convdisp(k, a)
 {
     m_ov_win = 0;
+    m_initial_lang = strdup(mIMLang);
+    m_initial_fontset = NULL;
+    m_lang_changed = false;
 }
 
 ConvdispOv::~ConvdispOv()
 {
-    if (m_ov_win) {
+    if (m_ov_win)
 	delete m_ov_win;
+
+    free(m_initial_lang);
+}
+
+void ConvdispOv::set_im_lang(const char *im_lang)
+{
+    mIMLang = im_lang;
+
+    if (!strcmp(m_initial_lang, im_lang)) {
+#if 0
+	if (m_initial_fontset)
+	    m_atr->font_set = m_initial_fontset;
+	else
+	    m_atr->font_set = choose_default_fontset(mLang);
+#endif
+    } else {
+#if 0
+	m_atr->font_set = choose_default_fontset(mLang);
+#endif
+	m_lang_changed = true;
     }
 }
 
 
 void ConvdispOv::update_preedit()
 {
-    if (!m_pe) {
-	return ;
-    }
+    if (!m_pe)
+	return;
+
     if (!m_pe->get_char_count()) {
-	delete m_ov_win;
-	m_ov_win = NULL;
-	return ;
+	clear_preedit();
+	return;
     }
 
     draw_preedit();
+    move_candwin();
+}
 
-    if (m_atr->has_atr(ICA_SpotLocation)){
-       
+void ConvdispOv::move_candwin()
+{
+    if (m_atr->has_atr(ICA_SpotLocation) &&
+		mKkContext == InputContext::focusedContext()) {
 	int x = -1, y = -1;
 	Window win;
 
-	if(m_atr->has_atr(ICA_FocusWindow)){
+	if (m_atr->has_atr(ICA_ClientWindow)) {
 
-	    XTranslateCoordinates(XimServer::gDpy,m_atr->focus_window,
+	    XTranslateCoordinates(XimServer::gDpy, m_atr->client_window,
 				  DefaultRootWindow(XimServer::gDpy),
 				  m_atr->spot_location.x,
 				  m_atr->spot_location.y,
 				  &x, &y, &win);
 	}
-      
-	if(m_atr->has_atr(ICA_ClientWindow)){
 
-	    XTranslateCoordinates(XimServer::gDpy,m_atr->client_window,
+	if (m_atr->has_atr(ICA_FocusWindow)) {
+
+	    XTranslateCoordinates(XimServer::gDpy, m_atr->focus_window,
 				  DefaultRootWindow(XimServer::gDpy),
 				  m_atr->spot_location.x,
 				  m_atr->spot_location.y,
 				  &x, &y, &win);
 	}
-	if(x > -1 && y > -1) {
-	    Canddisp* disp = canddisp_singleton();
-	    disp->move(x, y + 2);
+#ifdef FLASHPLAYER_WORKAROUND
+	if (revised_spot_y != -1) {
+	    XTranslateCoordinates(XimServer::gDpy, m_atr->client_window,
+				  DefaultRootWindow(XimServer::gDpy),
+				  m_atr->spot_location.x,
+				  revised_spot_y,
+				  &x, &y, &win);
 	}
+#endif
+	if (x > -1 && y > -1) {
+	    Canddisp *disp = canddisp_singleton();
+	    disp->move(x, y + UNDERLINE_HEIGHT + 1);
+	}
+#ifdef FLASHPLAYER_WORKAROUND
+    } else if (m_atr->has_atr(ICA_SpotLocation)) {
+	int x = -1, y = -1;
+	Window win;
+	if (revised_spot_y != -1) {
+	    XTranslateCoordinates(XimServer::gDpy, m_atr->client_window,
+				  DefaultRootWindow(XimServer::gDpy),
+				  m_atr->spot_location.x,
+				  revised_spot_y,
+				  &x, &y, &win);
+	    Canddisp *disp = canddisp_singleton();
+	    disp->move(x, y + UNDERLINE_HEIGHT + 1);
+	}
+#endif
     }
 }
 
+void ConvdispOv::clear_preedit()
+{
+    delete m_ov_win;
+    m_ov_win = NULL;
+}
+
 void ConvdispOv::validate_area()
 {
     Window r, win;
     int x;
     unsigned int w, h, tmp;
-    if (m_atr->has_atr(ICA_ClientWindow)) {
-	win = m_atr->client_window;
-    } else {
+    if (m_atr->has_atr(ICA_FocusWindow))
 	win = m_atr->focus_window;
-    }
+    else
+	win = m_atr->client_window;
+
     XGetGeometry(XimServer::gDpy, win,
 		 &r, &x, &x, &w, &h, &tmp, &tmp);
-    /* ̵,, (RedHat7.3Qtк֤¾Qt⤳ư)*/
+    // Absurd... (cope with Qt from RedHat7.3, and maybe some other)
     m_atr->area.width = w;
     m_atr->area.height = h;
 }
 
 void ConvdispOv::update_icxatr()
 {
-    if (!m_ov_win) {
-        return ;
-    }
+    if (!m_ov_win)
+	return;
     
     if (m_atr->is_changed(ICA_FocusWindow)) {
-	//Some clients send FocusWindow later
-        if (!check_win()) {
-            return ;
-        }
+	// Some clients send FocusWindow later
+	if (!check_win())
+	    return;
     }
     
     if (m_atr->is_changed(ICA_Area)) {
-	if (m_atr->area.width == 0) {
+	if (m_atr->area.width == 0)
 	    validate_area();
-	}
 	m_ov_win->set_size(m_atr->area.width, m_atr->area.height);
 	m_atr->unset_change_mask(ICA_Area);
     }
@@ -650,29 +835,16 @@ void ConvdispOv::update_icxatr()
     update_preedit();
 }
 
-void ConvdispOv::set_focus()
-{
-    Canddisp* disp = canddisp_singleton();
-    disp->show();
-}
-
-void ConvdispOv::unset_focus()
-{
-    Canddisp* disp = canddisp_singleton();
-    disp->hide();
-}
-
 void ConvdispOv::draw_preedit()
 {
     m_ce_len = m_pe->get_char_count();
-    if (!check_win()) {
-	return ;
-    }
+    if (!check_win())
+	return;
 
     m_ce = (char_ent *)malloc(sizeof(char_ent)*m_ce_len);
-    make_ce_array();//prepare array
-    layoutCharEnt();//layout them
-    do_draw_preedit();//do draw
+    make_ce_array();
+    layoutCharEnt();
+    do_draw_preedit();
     free(m_ce);
     m_ov_win->draw();
     XFlush(XimServer::gDpy);
@@ -680,8 +852,79 @@ void ConvdispOv::draw_preedit()
 
 void ConvdispOv::do_draw_preedit()
 {
+#ifdef FLASHPLAYER_WORKAROUND
+    revised_spot_y = -1;
+#endif
+
     if (m_atr->has_atr(ICA_Area)) {
-	m_ov_win->set_pos(m_atr->area.x, m_atr->area.y);
+#ifdef FLASHPLAYER_WORKAROUND
+	// Workaround for brain damaged flash player plugin (at least
+	// version 7.0 r25). --ekato
+	//
+	// Background: flashplayer plugin set same values for
+	// m_atr->area.y and m_atr->spot_location.y.
+	//
+	// 1. If preedit area goes beyond the parent window (browser),
+	// set the area at the upper side of client window.
+	//
+	// 2. If height for preedit is too narrow, make sure to set
+	// preedit within the client window.
+
+	if (m_atr->has_atr(ICA_SpotLocation) && (m_atr->area.y == m_atr->spot_location.y)) {
+	    XWindowAttributes clientattr, topattr;
+	    Window clientwin, topwin, win;
+	    int x, y;
+	    int font_height;
+	
+	    topattr.height = 0;
+	    if (m_atr->has_atr(ICA_FocusWindow))
+		clientwin = m_atr->focus_window;
+	    else
+		clientwin = m_atr->client_window;
+
+	    XGetWindowAttributes(XimServer::gDpy, clientwin, &clientattr);
+	    topwin = getTopWindow(XimServer::gDpy, clientwin);
+	    if (topwin)
+		XGetWindowAttributes(XimServer::gDpy, topwin, &topattr);
+	    XTranslateCoordinates(XimServer::gDpy, clientwin,
+			    DefaultRootWindow(XimServer::gDpy),
+			    m_atr->area.x,
+			    m_atr->area.y,
+			    &x, &y, &win);
+	    font_height = get_ce_font_height(m_ce, m_ce_len);
+
+	    if (topattr.height) {
+		if (y > (topattr.height + topattr.y)) {
+		    // preedit area goes beyond the top window's geometry
+		    if ((y - m_atr->area.y) < topattr.y) {
+			// Set preedit at upper side of the top
+			// window.  But it may doesn't work because it
+			// lacks height for window manager's title bar
+			// and some toolbars of browser...
+			m_ov_win->set_pos(m_atr->area.x, topattr.y - (y - m_atr->area.y));
+			revised_spot_y = font_height + topattr.y - (y - m_atr->area.y);
+		    } else {
+			// Set preedit at upper side of the flash
+			// player's window.
+			m_ov_win->set_pos(m_atr->area.x, 0);
+			revised_spot_y = font_height;
+		    }
+		} else {
+		    // preedit area is visible within the browser
+		    if ((clientattr.height - m_atr->area.y) < font_height + UNDERLINE_HEIGHT) {
+			// Make sure preedit+underline to be fit
+			// within the client window.
+			m_ov_win->set_pos(m_atr->area.x, clientattr.height - font_height - (UNDERLINE_HEIGHT + 1));
+			revised_spot_y = clientattr.height - (UNDERLINE_HEIGHT + 1);
+		    } else {
+			m_ov_win->set_pos(m_atr->area.x, m_atr->area.y);
+			revised_spot_y = font_height;
+		    }
+		}
+	    }
+	} else
+#endif // FLASHPLAYER_WORKAROUND
+	    m_ov_win->set_pos(m_atr->area.x, m_atr->area.y);
     } else {
 	m_ov_win->set_pos(0, 0);
     }
@@ -690,19 +933,14 @@ void ConvdispOv::do_draw_preedit()
 
 bool ConvdispOv::check_win()
 {
-    if (!check_atr()) {
-	//not enough information to map preedit
-	return false;
-    }
+    if (!check_atr())
+	return false; // not enough information to map preedit
   
-    if (m_ov_win && !m_atr->is_changed(ICA_FocusWindow)) {
-	// no need to update window
-	return true;
-    }
+    if (m_ov_win && !m_atr->is_changed(ICA_FocusWindow))
+	return true; // no need to update window
 
-    if (m_ov_win) {
+    if (m_ov_win)
 	delete m_ov_win;
-    }
   
     m_atr->unset_change_mask(ICA_FocusWindow);
     m_atr->unset_change_mask(ICA_Foreground);
@@ -710,13 +948,12 @@ bool ConvdispOv::check_win()
     m_atr->unset_change_mask(ICA_FontSet);
 
     Window w;
-    if (m_atr->has_atr(ICA_ClientWindow)) {
-	w = m_atr->client_window;
-    } else {
+    if (m_atr->has_atr(ICA_FocusWindow))
 	w = m_atr->focus_window;
-    }
+    else
+	w = m_atr->client_window;
 
-    m_ov_win = new PeOvWin(w);
+    m_ov_win = new PeOvWin(w, mIMLang, mEncoding, mLocaleName);
     m_ov_win->set_size(m_atr->area.width, m_atr->area.height);
     m_ov_win->set_fore(m_atr->foreground_pixel);
     m_ov_win->set_back(m_atr->background_pixel);
@@ -728,11 +965,11 @@ bool ConvdispOv::check_win()
 bool ConvdispOv::check_atr()
 {
     if (!m_atr->has_atr(ICA_FocusWindow) &&
-	!m_atr->has_atr(ICA_ClientWindow)) {
+	!m_atr->has_atr(ICA_ClientWindow))
 	return false;
-    }
+
     if (!m_atr->has_atr(ICA_SpotLocation)) {
-	//Ͼ꤬̤ʤ麸
+	// set at top-left corner unless SpotLocation is available
 	m_atr->spot_location.x = 0;
 	m_atr->spot_location.y = 0;
     }
@@ -742,20 +979,23 @@ bool ConvdispOv::check_atr()
 	m_atr->area.x = 0;
 	m_atr->area.y = 0;
     }
-    if (!m_atr->has_atr(ICA_FontSet)) {
-	m_atr->font_set = gFontset;
+    if (!m_atr->has_atr(ICA_FontSet))
+	m_atr->font_set = choose_default_fontset(mIMLang, mEncoding, mLocaleName);
+    else {
+	if (!m_initial_fontset && !m_lang_changed)
+	    m_initial_fontset = m_atr->font_set;
     }
-    if (!m_atr->has_atr(ICA_LineSpace)) {
+    if (!m_atr->has_atr(ICA_LineSpace))
 	m_atr->line_space = 14;
-    }
-    if (!m_atr->has_atr(ICA_Foreground)) {
+
+    if (!m_atr->has_atr(ICA_Foreground))
 	m_atr->foreground_pixel = BlackPixel(XimServer::gDpy,
 					     DefaultScreen(XimServer::gDpy));
-    }
-    if (!m_atr->has_atr(ICA_Background)) {
+
+    if (!m_atr->has_atr(ICA_Background))
 	m_atr->background_pixel = WhitePixel(XimServer::gDpy,
 					     DefaultScreen(XimServer::gDpy));
-    }
+
     return true;
 }
 
@@ -779,33 +1019,39 @@ void ConvdispOv::make_ce_array()
 void ConvdispOv::layoutCharEnt()
 {
     int i;
-    int x,y;
+    int x, y;
 
     x = m_atr->spot_location.x;
     y = m_atr->spot_location.y;
 
     for (i = 0; i < m_ce_len; i++) {
 	uchar ch = m_ce[i].c;
-	char buf[3];
-	int l;
-	XRectangle ink,logical;
-	if (ch < 256) {
-	    buf[0] = ch;
-	    buf[1] = 0;
-	    l = 1;
-	} else {
-	    buf[2] = 0;
-	    buf[1] = (ch & 255)|0x80;
-	    buf[0] = ((ch>>8)&255)|0x80;
-	    l = 2;
+	XRectangle ink, logical;
+
+	if (!strcmp(mEncoding, "UTF-8"))
+	    XwcTextExtents(m_atr->font_set, &ch, 1, &ink, &logical);
+	else {
+	    char utf8[6];
+	    int len = utf8_wctomb((unsigned char *)utf8, ch);
+	    utf8[len] = '\0';
+	    char *str;
+	    str = utf8_to_native_str(utf8, mEncoding);
+	    if (!str) {
+		logical.width = 0;
+		logical.height = (i > 0) ? m_ce[i - 1].height : 0;
+	    } else {
+		len = strlen(str);
+		XmbTextExtents(m_atr->font_set, str, len, &ink, &logical);
+		free(str);
+	    }
 	}
-	XmbTextExtents(m_atr->font_set, buf, l, &ink, &logical);
+
 	m_ce[i].width = logical.width;
 	m_ce[i].height = logical.height;
 	int right_limit = m_atr->area.width;
-	if (m_atr->has_atr(ICA_Area)) {
+	if (m_atr->has_atr(ICA_Area))
 	    right_limit += m_atr->area.x;
-	}
+
 	if (m_ce[i].width + x >
 	    right_limit) {
 	    // goto next line
@@ -820,23 +1066,41 @@ void ConvdispOv::layoutCharEnt()
 	for (i = 0; i < m_ce_len; i++) {
 	    m_ce[i].x -= m_atr->area.x;
 	    m_ce[i].y -= m_atr->area.y;
+#ifdef FLASHPLAYER_WORKAROUND
+	    // workaround for brain damaged flash player plugin
+	    if (m_ce[i].y == 0)
+		m_ce[i].y += m_ce[i].height;
+#endif
 	}
     }
 }
 
 int ConvdispOv::calc_ce_width(int b, int e)
 {
-    int i,w=0;
+    int i, w = 0;
     for (i = b; i < e; i++) {
-        w += m_ce[i].width;
+	w += m_ce[i].width;
     }
     return w;
 }
 
+#ifdef FLASHPLAYER_WORKAROUND
+int ConvdispOv::get_ce_font_height(char_ent *ce, int len)
+{
+    int i, h = 0;
+    for (i = 0; i < len; i++) {
+	if (ce[i].height > h)
+	    h = ce[i].height;
+    }
+    return h;
+}
+#endif
+
+// On the spot style
 ConvdispOs::ConvdispOs(InputContext *k, icxatr *a, Connection *c)
     : Convdisp(k, a)
 {
-    XimIC *ic=k->get_ic();
+    XimIC *ic = k->get_ic();
     mConn = c;
     mImid = ic->get_imid();
     mIcid = ic->get_icid();
@@ -849,18 +1113,18 @@ ConvdispOs::~ConvdispOs()
 
 void ConvdispOs::update_preedit()
 {
-    if (!m_pe) {
-        return ;
-    }
+    if (!m_pe)
+	return;
+
     TxPacket *t;
 
     int len, caret_pos;
     len = m_pe->get_char_count();
     caret_pos = m_pe->caret_pos;
 
-    if (mPrevLen == 0 && len == 0) {
-	return ;
-    }
+    if (mPrevLen == 0 && len == 0)
+	return;
+
     if (mPrevLen == 0 && len) {
 	t = createTxPacket(XIM_PREEDIT_START, 0);
 	t->pushC16(mImid);
@@ -875,11 +1139,10 @@ void ConvdispOs::update_preedit()
     t->pushC32(0); // chg_first
     t->pushC32(mPrevLen); // chg_length
 
-    if (m_pe->ustrings.size()) {
+    if (m_pe->ustrings.size())
 	t->pushC32(0);
-    } else {
+    else
 	t->pushC32(3);
-    }
   
     compose_preedit_array(t);
     compose_feedback_array(t);
@@ -899,25 +1162,50 @@ void ConvdispOs::update_preedit()
       t->pushC32(pos);
       t->pushC32(dir);
       t->pushC32(style);
-      mConn->push_passive_packet(t);
+      mConn->push_packet(t);
       */
+    move_candwin();
+}
       
-    if (m_atr->has_atr(ICA_FocusWindow)
-        && m_atr->has_atr(ICA_SpotLocation)){
+void ConvdispOs::move_candwin()
+{
+    if (m_atr->has_atr(ICA_ClientWindow)) {
 	int x, y;
 	Window win;
-	XTranslateCoordinates(XimServer::gDpy,m_atr->focus_window,
+	XWindowAttributes xattr;
+
+	if (!m_atr->has_atr(ICA_SpotLocation)) {
+	    m_atr->spot_location.x = 0;
+	    m_atr->spot_location.y = 0;
+	}
+
+	XTranslateCoordinates(XimServer::gDpy, m_atr->client_window,
 			      DefaultRootWindow(XimServer::gDpy),
 			      m_atr->spot_location.x,
 			      m_atr->spot_location.y,
 			      &x, &y, &win);
-        
-        Canddisp* disp = canddisp_singleton();
-        disp->move(x, y + 2);
+
+	Canddisp *disp = canddisp_singleton();
+
+	if (m_atr->has_atr(ICA_SpotLocation))
+	    disp->move(x, y + 36); // 36 pixel seems too long.
+				   // Is there any way to get current preedit
+				   // height?
+	else {
+	    XGetWindowAttributes(XimServer::gDpy, m_atr->client_window,
+			      &xattr);
+	    //disp->move(x + xattr.width + 2, y + 2); //upper-right side
+	    disp->move(x, y + xattr.height + 2); //lower-left side
+	}
     }    
     
 }
 
+void ConvdispOs::clear_preedit()
+{
+    mPrevLen = 0;
+}
+
 void ConvdispOs::update_icxatr()
 {
 }
@@ -927,60 +1215,72 @@ void ConvdispOs::compose_preedit_array(TxPacket *t)
     uString s;
     std::list<pe_ustring>::iterator it;
     for (it = m_pe->ustrings.begin(); it != m_pe->ustrings.end(); it++) {
-        append_ustring(&s, &(*it).s);
+	append_ustring(&s, &(*it).s);
     }
 
-    char *c = mKkContext->getServer()->uStringToCtext(&s);
-    int i,len;
-    len = strlen(c);
-    t->pushC16(len);// LENGTH
-    for (i = 0 ; i < len ; i++) {
-        t->pushC8(c[i]); // CTEXT
-    }
-    len = pad4(len+2);
+    char *c = mKkContext->getServer()->uStringToCtext(&s, mEncoding);
+    int i, len = 0;
+    if (c)
+	len = strlen(c);
+    t->pushC16(len); // LENGTH
     for (i = 0; i < len; i++) {
-        t->pushC8(0); // PADDING
+	t->pushC8(c[i]); // CTEXT
     }
-    free(c);
+    len = pad4(len + 2);
+    for (i = 0; i < len; i++) {
+	t->pushC8(0); // PADDING
+    }
+    if (c)
+	free(c);
 }
 
 void ConvdispOs::compose_feedback_array(TxPacket *t)
 {
-    int i,len,stat,xstat;
+    int i, len, stat, xstat;
     len = m_pe->get_char_count();
-    t->pushC16(len*4);
+    t->pushC16(len * 4);
     t->pushC16(0);
     std::list<pe_ustring>::iterator it;
     for (it = m_pe->ustrings.begin(); it != m_pe->ustrings.end(); it++) {
-        len = (*it).s.size();
-        stat = (*it).stat;
-        xstat = FB_None;
-        if (stat & PE_REVERSE) {
-            xstat |= FB_Reverse;
-        }
-        if (stat & PE_UNDERLINE) {
-            xstat |= FB_Underline;
-        }
-        if (stat & PE_HILIGHT) {
-            xstat |= FB_Highlight;
-        }
-        for (i = 0; i < len; i++) {
-            t->pushC32(xstat);
-        }
+	len = (*it).s.size();
+	stat = (*it).stat;
+	xstat = FB_None;
+	if (stat & PE_REVERSE)
+	    xstat |= FB_Reverse;
+
+	if (stat & PE_UNDERLINE)
+	    xstat |= FB_Underline;
+
+	if (stat & PE_HILIGHT)
+	    xstat |= FB_Highlight;
+
+	for (i = 0; i < len; i++) {
+	    t->pushC32(xstat);
+	}
     }
 }
 
-void ConvdispOs::set_focus()
+#ifdef FLASHPLAYER_WORKAROUND
+static Window getTopWindow(Display *d, Window w)
 {
-    Canddisp* disp = canddisp_singleton();
-    disp->show();
-}
+    Window root, parent, *children;
+    unsigned int nchild;
+    Status retval;
+
+    for (;;) {
+	retval = XQueryTree(d, w, &root, &parent, &children, &nchild);
+	if (retval == 0)
+	    return 0;
+	if (children)
+	    XFree(children);
+	if (parent == root)
+	    break;
+	w = parent;
+    }
+    return w;
+};
+#endif
 
-void ConvdispOs::unset_focus()
-{
-    Canddisp* disp = canddisp_singleton();
-    disp->hide();
-}
 /*
  * Local variables:
  *  c-indent-level: 4
