@@ -30,7 +30,7 @@
   SUCH DAMAGE.
 */
 
-// XIM Server currently for Japanese
+// XIM Server supporting CJK languages
 // initialize many modules
 //
 #ifdef HAVE_CONFIG_H
@@ -49,23 +49,29 @@
 #include "xim.h"
 #include "xdispatch.h"
 #include "ximserver.h"
+#include "canddisp.h"
+#include "uim/uim-util.h"
+#include "uim/uim-im-switcher.h"
 
 Display *XimServer::gDpy;
 std::map<Window, XimServer *> XimServer::gServerMap;
 
 // Configuration
 int g_option_mask;
-int scr_width,scr_height;
+int scr_width, scr_height;
 int host_byte_order;
 
 int lib_uim_fd;
 
 #define VERSION_NAME "uim-xim under the way! Version "PACKAGE_VERSION"\n"
-char *version_name=VERSION_NAME;
-char *usage=
+const char *version_name=VERSION_NAME;
+const char *usage=
 "--help , --version :Show usage or version\n"
-"--trace :trace-connection\n"
-"--trace-xim :trace-xim-message\n";
+"--list             :Show available backend conversion engines\n"
+"--engine=ENGINE    :Use ENGINE as a backend conversion engine at startup\n"
+"--trace            :trace-connection\n"
+"--trace-xim        :trace-xim-message\n";
+const char *default_engine;
 
 
 static Atom atom_locales;
@@ -78,11 +84,13 @@ Atom xim_servers;
 
 struct fd_watch_struct {
     int mask;
-    void (*fn)(int,int);
+    void (*fn)(int, int);
 };
-static std::map<int,fd_watch_struct> fd_watch_stat;
+static std::map<int, fd_watch_struct> fd_watch_stat;
 static std::map<unsigned int, WindowIf *> window_watch_stat;
 
+static char *supported_locales;
+std::list<UIMInfo> uim_info;
 
 bool
 pretrans_register()
@@ -106,24 +114,24 @@ WindowIf::resize(Window, int, int)
     // do nothing
 }
 
-static void remove_current_fd_watch(int fd)
+void remove_current_fd_watch(int fd)
 {
-    std::map<int,fd_watch_struct>::iterator i;
+    std::map<int, fd_watch_struct>::iterator i;
     i = fd_watch_stat.find(fd);
-    if (i == fd_watch_stat.end()) {
-        return ;
-    }
+    if (i == fd_watch_stat.end())
+	return;
+
     fd_watch_stat.erase(i);
 }
 
-static void add_fd_watch(int fd, int mask, void (*fn)(int, int))
+void add_fd_watch(int fd, int mask, void (*fn)(int, int))
 {
     remove_current_fd_watch(fd);
 
     fd_watch_struct s;
     s.mask = mask;
     s.fn = fn;
-    std::pair<int,fd_watch_struct> p(fd, s);
+    std::pair<int, fd_watch_struct> p(fd, s);
     fd_watch_stat.insert(p);
 }
 
@@ -133,29 +141,24 @@ void main_loop()
     while (1) {
 	FD_ZERO(&rfds);
 	FD_ZERO(&wfds);
-	std::map<int,fd_watch_struct>::iterator it;
+	std::map<int, fd_watch_struct>::iterator it;
 	int  fd_max = 0;
 	for (it = fd_watch_stat.begin(); it != fd_watch_stat.end(); it++) {
 	    int fd = it->first;
-	    if (it->second.mask & READ_OK) {
+	    if (it->second.mask & READ_OK)
 		FD_SET(fd, &rfds);
-	    }
-	    if (it->second.mask & WRITE_OK) {
+	    if (it->second.mask & WRITE_OK)
 		FD_SET(fd, &wfds);
-	    }
-	    if (fd_max < fd) {
+	    if (fd_max < fd)
 		fd_max = fd;
-	    }
 	}
 	select(fd_max + 1, &rfds, &wfds, NULL, NULL);
 	for (it = fd_watch_stat.begin(); it != fd_watch_stat.end(); it++) {
 	    int fd = it->first;
-	    if (FD_ISSET(fd, &rfds)) {
+	    if (FD_ISSET(fd, &rfds))
 		it->second.fn(fd, READ_OK);
-	    }
-	    if (FD_ISSET(fd, &wfds)) {
+	    if (FD_ISSET(fd, &wfds))
 		it->second.fn(fd, WRITE_OK);
-	    }
 	}
     }
 }
@@ -163,48 +166,46 @@ void main_loop()
 void
 add_window_watch(Window id, WindowIf *w, int mask)
 {
-    std::pair<unsigned int,WindowIf *> p(id, w);
+    std::pair<unsigned int, WindowIf *> p(id, w);
     window_watch_stat.insert(p);
 
-    //Event mash has same value defined in X,
-    //but do not depend on.
+    // Event mash has same value defined in X,
+    // but do not depend on.
     int emask = 0;
-    if (mask & EXPOSE_MASK) {
+    if (mask & EXPOSE_MASK)
 	emask |= ExposureMask;
-    }
-    if (mask & STRUCTURE_NOTIFY_MASK) {
+    if (mask & STRUCTURE_NOTIFY_MASK)
 	emask |= StructureNotifyMask;
-    }
+
     XSelectInput(XimServer::gDpy, id, emask);
 }
 
 void
 remove_window_watch(Window id)
 {
-    std::map<unsigned int,WindowIf *>::iterator i;
+    std::map<unsigned int, WindowIf *>::iterator i;
     i = window_watch_stat.find(id);
-    if (i != window_watch_stat.end()) {
+    if (i != window_watch_stat.end())
 	window_watch_stat.erase(i);
-    }
 }
 
 WindowIf *
 findWindowIf(Window w)
 {
-    std::map<unsigned int,WindowIf *>::iterator i;
+    std::map<unsigned int, WindowIf *>::iterator i;
     i = window_watch_stat.find(w);
-    if (i == window_watch_stat.end()) {
+    if (i == window_watch_stat.end())
 	return NULL;
-    }
+
     return (*i).second;
 }
 
 static int
 X_ErrorHandler(Display *d, XErrorEvent *e)
 {
-    if (g_option_mask & OPT_TRACE) {
+    if (g_option_mask & OPT_TRACE)
 	printf("X error occured.\n");
-    }
+
     return 0;
 }
 
@@ -226,10 +227,10 @@ sendSelectionNotify(XEvent *ev, char *buf, int len)
     e.xselection.time = ev->xselectionrequest.time;
     e.xselection.property = ev->xselectionrequest.property;
     XChangeProperty(XimServer::gDpy, e.xselection.requestor,
-                    e.xselection.property,
-                    e.xselection.target,
-                    8,PropModeReplace,
-                    (unsigned char *)buf, len);
+		    e.xselection.property,
+		    e.xselection.target,
+		    8, PropModeReplace,
+		    (unsigned char *)buf, len);
     XSendEvent(XimServer::gDpy, e.xselection.requestor, 0, 0, &e);
     XFlush(XimServer::gDpy);
 }
@@ -237,76 +238,58 @@ sendSelectionNotify(XEvent *ev, char *buf, int len)
 void
 notifyLocale(XEvent *ev)
 {
-    char buf[32];
-    XimServer *xs = XimServer::findServer(ev->xselectionrequest.owner);
-    strcpy(buf,"@locale=ja_JP");
-    sendSelectionNotify(ev, buf, strlen(buf)+1);
-    if (g_option_mask & OPT_TRACE) {
+    sendSelectionNotify(ev, supported_locales, strlen(supported_locales) + 1);
+    if (g_option_mask & OPT_TRACE)
 	printf("selection notify request for locale.\n");
-    }
 }
 
 void
 notifyTransport(XEvent *ev)
 {
-    sendSelectionNotify(ev, "@transport=X/", 13+1);
-    if (g_option_mask & OPT_TRACE) {
+    sendSelectionNotify(ev, "@transport=X/", 13 + 1);
+    if (g_option_mask & OPT_TRACE)
 	printf("selection notify request for transport.\n");
-    }
 }
 
 void
 ProcXEvent(XEvent *e)
 {
     Atom p;
+    WindowIf *i;
     switch (e->type) {
     case SelectionRequest:
-    {
 	p = e->xselectionrequest.property;
-	if (p == atom_locales) {
+	if (p == atom_locales)
 	    notifyLocale(e);
-	} else if(p == atom_transport) {
+	else if (p == atom_transport)
 	    notifyTransport(e);
-	} else {
+	else
 	    printf("property %s?\n",
-		   XGetAtomName(XimServer::gDpy,e->xselection.property));
-	    break;
-	}
-    }
-    break;
+		   XGetAtomName(XimServer::gDpy, e->xselection.property));
+	break;
     case Expose:
-    {
 	if (e->xexpose.count == 0) {
-	    WindowIf *i = findWindowIf(e->xexpose.window);
-	    if (i) {
+	    i = findWindowIf(e->xexpose.window);
+	    if (i)
 		i->expose(e->xexpose.window);
-	    }
 	}
-    }
-    break;
+	break;
     case ConfigureNotify:
-    {
-	WindowIf *i = findWindowIf(e->xconfigure.window);
-	if (i) {
-	    i->resize(e->xconfigure.window,
-		      e->xconfigure.x, e->xconfigure.y);
-	}
-    }
-    break;
+	i = findWindowIf(e->xconfigure.window);
+	if (i)
+	    i->resize(e->xconfigure.window, e->xconfigure.x, e->xconfigure.y);
+	break;
     case DestroyNotify:
-    {
-	WindowIf *i = findWindowIf(e->xdestroywindow.window);
-	if (i) {
+	i = findWindowIf(e->xdestroywindow.window);
+	if (i)
 	    i->destroy(e->xdestroywindow.window);
-	}
 	remove_window_watch(e->xdestroywindow.window);
-    }
-    break;
+	break;
     case ClientMessage:
 	procXClientMessage(&e->xclient);
-    break;
+	break;
     default:;
-	//printf("unknown type of X event. %d\n",e->type);
+	//printf("unknown type of X event. %d\n", e->type);
     }
 }
 
@@ -347,6 +330,95 @@ print_usage()
     exit(0);
 }
 
+static void
+get_uim_info()
+{
+    int res;
+
+    res = uim_init();
+    if (res) {
+	printf("Failed to init uim\n");
+	exit(1);
+    }
+    uim_context uc = uim_create_context(NULL, "UTF-8", NULL,
+					NULL, uim_iconv, NULL);
+
+    struct UIMInfo ui;
+
+    int nr = uim_get_nr_im(uc);
+    for (int i = 0; i < nr; i++) {
+	ui.name = strdup(uim_get_im_name(uc, i));
+	ui.lang = strdup(uim_get_im_language(uc, i));
+	ui.locale = createLocale(ui.lang);
+	ui.desc = strdup(uim_get_im_short_desc(uc, i));
+	if (!ui.locale)
+	    continue;
+	uim_info.push_back(ui);
+    }
+    uim_release_context(uc);
+}
+
+static void
+print_uim_info()
+{
+    std::list<UIMInfo>::iterator it;
+
+    printf("Supported conversion engines:\n");
+    if (uim_info.empty())
+	printf("  None.\n");
+    else
+	for (it = uim_info.begin(); it != uim_info.end(); it++)
+	    printf("  %s (%s)\n", it->name, it->lang);
+    
+}
+
+static void
+init_supported_locales()
+{
+    std::list<char *> locale_list;
+    char *name;
+    char locale[20];
+    unsigned int len;
+
+    asprintf(&supported_locales, "@locale=");
+
+    std::list<UIMInfo>::iterator i;
+    std::list<char *>::iterator j;
+    for (i = uim_info.begin(); i != uim_info.end(); i++) {
+	name = strdup(compose_localenames_from_im_lang(i->lang));
+	// Handles IM supporting several locales.
+	char *sep;
+	char *tmp = name;
+	while ((sep = strchr(tmp, ':')) != NULL) {
+	    *sep = ',';
+	    tmp = sep;
+	}
+
+	bool is_exist = false;
+	for (j = locale_list.begin(); j != locale_list.end(); j++) {
+	    if (!strcmp(*j, name)) {
+		is_exist = true;
+		break;
+	    }
+	}
+	if (is_exist == true) {
+	    free(name);
+	    continue;
+	}
+
+	locale_list.push_back(name);
+	snprintf(locale, sizeof(locale), "%s,", name); // Trailing comma seems to be OK...
+	len = strlen(supported_locales) + strlen(locale);
+	supported_locales = (char *)realloc(supported_locales,
+			     sizeof(char) * len + 1);
+	if (!supported_locales)
+	    break;
+	strcat(supported_locales, locale);
+    }
+    for (j = locale_list.begin(); j != locale_list.end(); j++) {
+	free(*j);
+    }
+}
 
 static void
 parse_args(int argc, char **argv)
@@ -361,25 +433,134 @@ parse_args(int argc, char **argv)
 		exit(0);
 	    } else if (!strcmp(opt, "help")) {
 		print_usage();
-	    } else if(!strcmp(opt, "trace")) {
+	    } else if (!strcmp(opt, "list")) {
+		get_uim_info();
+		print_uim_info();
+		exit(0);
+	    } else if (!strcmp(opt, "trace")) {
 		g_option_mask |= OPT_TRACE;
-	    } else if(!strcmp(opt, "trace-xim")) {
+	    } else if (!strcmp(opt, "trace-xim")) {
 		g_option_mask |= OPT_TRACE_XIM;
+	    } else if (!strncmp(opt, "engine=", 7)) {
+		default_engine = strdup(&argv[i][9]);
 	    }
 	}
     }
 }
 
+static void check_default_engine(const char *locale)
+{
+    bool found = false;
+    if (default_engine) {
+	std::list<UIMInfo>::iterator it;
+	for (it = uim_info.begin(); it != uim_info.end(); it++) {
+	    if (!strcmp(it->name, default_engine)) {
+		found = true;
+		break;
+	    }
+	}
+    }
+
+    if (found == false)
+	default_engine = uim_get_default_im_name(locale);
+}
+
 static void
 get_runtime_env()
 {
     int i = 1;
     char *v = (char *)&i;
-    if (*v == 1) {
+    if (*v == 1)
 	host_byte_order = LSB_FIRST;
-    } else {
+    else
 	host_byte_order = MSB_FIRST;
+}
+
+static void
+parse_helper_str_im_change(const char *level, const char *engine) {
+    InputContext *focusedContext = InputContext::focusedContext();
+
+    if (!strcmp(level, "im_change_whole_desktop")) {
+	std::map<Window, XimServer *>::iterator it;
+	for (it = XimServer::gServerMap.begin(); it != XimServer::gServerMap.end(); it++) {
+	    (*it).second->changeContext(engine);
+	}
     }
+
+    if (focusedContext) {
+	if (!strcmp(level, "im_change_this_text_area_only"))
+	    focusedContext->changeContext(engine);
+	else if (!strcmp(level, "im_change_this_application_only"))
+	    get_im_by_id(focusedContext->get_ic()->get_imid())->changeContext(engine);
+    }
+}
+
+static void
+send_im_list(void)
+{
+    char *buf = NULL, *tmp = NULL;
+    int len;
+    InputContext *focusedContext = InputContext::focusedContext();
+    const char *current_im_name =
+	    uim_get_current_im_name(focusedContext->getUC());
+    const char *encoding = focusedContext->get_ic()->get_encoding();
+    const char *client_locale = NULL;
+    
+    if (strcmp(encoding, "UTF-8"))
+	client_locale = focusedContext->get_ic()->get_lang_region();
+
+    asprintf(&buf, "im_list\ncharset=UTF-8\n");
+    if (!buf)
+	return;
+
+    std::list<UIMInfo>::iterator it;
+    for (it = uim_info.begin(); it != uim_info.end(); it++) {
+	if (client_locale) {
+	    const char *engine_locales =
+		    compose_localenames_from_im_lang(it->lang);
+	    if (!is_locale_included(engine_locales, client_locale) &&
+				strcmp(it->lang, "*"))
+		    continue;
+	}
+
+	const char *language;
+	language = uim_get_language_name_from_locale(it->lang);
+	asprintf(&tmp, "%s\t%s\t%s\t", it->name,
+				       language ? language : "",
+				       it->desc ? it->desc : "");
+	if (!tmp)
+	    return;
+	len = strlen(buf) + strlen(tmp);
+	buf = (char *)realloc(buf, sizeof(char) * len + 1);
+	if (!buf)
+	    return;
+	strcat(buf, tmp);
+	free(tmp);
+
+	if (!strcmp(it->name, current_im_name)) {
+	    asprintf(&tmp, "selected\n");
+	    if (!tmp)
+		return;
+	    len = strlen(buf) + strlen(tmp);
+	    buf = (char *)realloc(buf, sizeof(char) * len + 1);
+	    if (!buf)
+		return;
+	    strcat(buf, tmp);
+	    free(tmp);
+	} else {
+	    asprintf(&tmp, "\n");
+	    if (!tmp)
+		return;
+	    len = strlen(buf) + strlen(tmp);
+	    buf = (char *)realloc(buf, sizeof(char) * len + 1);
+	    if (!buf)
+		return;
+	    strcat(buf, tmp);
+	    free(tmp);
+	}
+    }
+    uim_helper_send_message(lib_uim_fd, buf);
+    free(buf);
 }
 
 static void
@@ -390,9 +571,8 @@ helper_str_parse(char *str)
     if (focusedContext) {
 	char *line = str;	
 	char *eol = strchr(line, '\n');
-	if (eol != NULL) {
+	if (eol != NULL)
 	    *eol = '\0';
-	}
 
 	if (strcmp("prop_list_get", line) == 0) {
 	    uim_prop_list_update(focusedContext->getUC());
@@ -400,16 +580,44 @@ helper_str_parse(char *str)
 	    uim_prop_label_update(focusedContext->getUC());
 	} else if (strcmp("prop_activate", line) == 0) {
 	    line = eol + 1;
-	    
 	    eol = strchr(line, '\n');
 	    if (eol != NULL) {
 		*eol = '\0';
 	    }
-
 	    uim_prop_activate(focusedContext->getUC(), line);
 	} else if (strncmp("focus_in", line, 8) == 0) {
-            InputContext::deletefocusedContext();
- 	}
+	    InputContext::deletefocusedContext();
+	    Canddisp *disp = canddisp_singleton();
+	    disp->hide();
+ 	} else if (strcmp("im_list_get", line) == 0) {
+	    send_im_list();
+	} else if (strncmp("im_change_", line, 10) == 0) {
+	    char *engine;
+	    engine = eol + 1;
+	    eol = strchr(engine, '\n');
+	    if (eol != NULL) {
+		*eol = '\0';
+	    }
+	    parse_helper_str_im_change(line, engine);
+	}
+    } else {
+	char *line = str;	
+	char *eol = strchr(line, '\n');
+	if (eol != NULL) {
+	    *eol = '\0';
+	}
+
+	if (strncmp("im_change_", line, 10) == 0) {
+	    char *engine;
+	    engine = eol + 1;
+	    eol = strchr(engine, '\n');
+	    if (eol != NULL) {
+		*eol = '\0';
+	    }
+	    parse_helper_str_im_change(line, engine);
+	} else if (strcmp("prop_update_custom", line) == 0) {
+	    fprintf(stderr, "prop_update_custom is not implemented yet\n");
+	}
     }
 }
 
@@ -434,56 +642,84 @@ helper_disconnect_cb(void)
 int
 main(int argc, char **argv)
 {
-    int res;
-    printf("UIM-XIM bridge. Now only supports Japanese.\n");
+    const char *locale;
+
+    printf("UIM-XIM bridge. Now testing to support multiple locales.\n");
 
     get_runtime_env();
+
     parse_args(argc, argv);
-    setlocale(LC_ALL, "ja_JP.eucJP");
-    signal(SIGPIPE , SIG_IGN);
+    signal(SIGPIPE, SIG_IGN);
 
     lib_uim_fd = uim_helper_init_client_fd(helper_disconnect_cb);
-    if (lib_uim_fd >= 0) {
+    if (lib_uim_fd >= 0)
 	add_fd_watch(lib_uim_fd, READ_OK, helper_read_cb);
-    }
 
-    res = uim_init();
-    if (res) {
-	printf("Failed to init uim\n");
-	return 1;
-    }
     XimServer::gDpy = XOpenDisplay(NULL);
     if (!XimServer::gDpy) {
-        printf("failed to open display!\n");
-        return 1;
+	printf("failed to open display!\n");
+	return 1;
     }
     if (!pretrans_register()) {
 	printf("pretrans_register failed\n");
 	return 1;
     }
-    uim_context uc = uim_create_context(NULL, "EUC-JP", NULL,
-					NULL, uim_iconv, NULL);
-    int nr = uim_get_nr_im(uc);
-    for (int i = 0; i < nr; i++) {
-	char *name = (char *)uim_get_im_name(uc, i);
-	char *lang = (char *)uim_get_im_language(uc, i);
-	Locale *lc = getLocale(lang);
-	if (!lc) {
-	    continue;
+
+    get_uim_info();
+    print_uim_info();
+
+    locale = setlocale(LC_CTYPE, "");
+    if (!locale)
+	locale = setlocale(LC_CTYPE, "C");
+
+    check_default_engine(locale);
+    init_supported_locales();
+    init_modifier_keys();
+
+    std::list<UIMInfo>::iterator it;
+    bool res = false;
+
+    // First, setup conversion engine selected by cmdline option or
+    // "default-im-name" on ~/.uim.
+    for (it = uim_info.begin(); it != uim_info.end(); it++) {
+	if (strcmp(it->name, default_engine) == 0) {
+	    XimServer *xs = new XimServer(it->locale, it->name, it->lang);
+	    res = xs->setupConnection(true);
+	    if (res)
+		printf("XMODIFIERS=@im=uim registered, selecting %s (%s) as default conversion engine\n", it->name, it->lang);
+	    else
+		delete xs;
+	    break;
 	}
-	XimServer *xs = new XimServer(lc, name, lang);
-	bool res = xs->setupConnection();
-	if (!res) {
+    }
+
+    // XXX Setup each backend as a separated XimServer, aimed for
+    // backward compatibility, and should be removed in the future.
+    for (it = uim_info.begin(); it != uim_info.end(); it++) {
+	XimServer *xs = new XimServer(it->locale, it->name, it->lang);
+	res = xs->setupConnection(false);
+	if (res)
+	    printf("XMODIFIERS=@im=uim-%s (%s) registered\n",
+			    it->name, it->lang);
+	else
 	    delete xs;
-	}
-	printf("XMODIFIERS=@im=uim-%s ,(%s) registered\n", name, lang);
     }
-    
+
+    if (!res) {
+	printf("aborting...\n");
+	return 1;
+    }
+
     connection_setup();
-    if (pretrans_setup() == -1) {
+    if (pretrans_setup() == -1)
 	return 0;
+
+    // Handle pending events to prevent hang just after startup
+    XEvent e;
+    while (XPending(XimServer::gDpy)) {
+	XNextEvent(XimServer::gDpy, &e);
+	ProcXEvent(&e);
     }
-    init_convdisp();
 
     main_loop();
     return 0;
@@ -492,7 +728,7 @@ main(int argc, char **argv)
 int
 pad4(int x)
 {
-    return (4-(x%4))%4;
+    return (4 - (x % 4)) % 4;
 }
 
 void
@@ -500,16 +736,15 @@ hex_dump(unsigned char *buf, int len)
 {
     int i, j;
     unsigned char c[16];
-    for (i = 0 ; i < len ; i += 16) {
+    for (i = 0; i < len; i += 16) {
 	printf("%8.8x ", i);
-	for (j = 0 ; j < 16 ; j++) {
+	for (j = 0; j < 16; j++) {
 	    if (i + j < len) {
-		c[j] = buf[i+j];
-		if (j == 7) {
+		c[j] = buf[i + j];
+		if (j == 7)
 		    printf("%2.2x-", c[j]);
-		} else {
+		else
 		    printf("%2.2x ", c[j]);
-		}
 	    } else {
 		c[j] = 0;
 		printf("-- ");
@@ -517,11 +752,10 @@ hex_dump(unsigned char *buf, int len)
 	}
 	printf(" ");
 	for (j = 0; j < 16; j++) {
-	    if (isprint(c[j])) {
+	    if (isprint(c[j]))
 		printf("%c", c[j]);
-	    } else {
+	    else
 		printf(".");
-	    }
 	}
 	printf("\n");
     }
