@@ -30,9 +30,8 @@
   SUCH DAMAGE.
 */
 
-// XIMץȥΥѥåȤIC(Input Context)˴ؤʪ
-// ä˥줿Ȥåʸ򥳥ߥåȤå
-// ̾⥸塼ؤȤδ֤𤹤롣
+// Handle IC (Input Context) part of XIM protocol.
+// Intermediate key handling between backend IM engine.
  
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -43,6 +42,7 @@
 #include "xim.h"
 #include "convdisp.h"
 #include "ximserver.h"
+#include "connection.h"
  
 #ifndef __GNUC__
 # ifdef HAVE_ALLOCA_H
@@ -56,60 +56,54 @@ char invalid_style_msg[]=
 XimIC *XimIC::current_ic;
 int XimIC::nrActiveIC;
 
-struct FSCache{
+struct FSCache {
     int refc;
     XFontSet fs;
     char *name;
+    char *locale;
 };
 
 static std::list<FSCache> fs_cache;
 
-static void
-ui_update_ic_stat()
-{
-}
-
-static void
-ui_update_input_mode(int mode)
-{
-}
-
-static XFontSet
-get_font_set(char *name)
+XFontSet
+get_font_set(const char *name, const char *locale)
 {
     std::list<FSCache>::iterator it;
     for (it = fs_cache.begin(); it != fs_cache.end(); it++) {
-	if (!strcmp(it->name, name)) {
+	if (!strcmp(it->name, name) && !strcmp(it->locale, locale)) {
 	    it->refc++;
 	    return it->fs;
 	}
     }
     struct FSCache fc;
-    char **missing,*def;
+    char **missing, *def;
     int nr_missing;
     fc.fs = XCreateFontSet(XimServer::gDpy, name,
 			   &missing, &nr_missing, &def);
-    if (missing) {
+    if (missing)
 	XFreeStringList(missing);
-    }
+
     fc.refc = 1;
     fc.name = strdup(name);
+    fc.locale = strdup(locale);
     fs_cache.push_back(fc);
     return fc.fs;
 }
 
 static void
-release_font_set(char *name)
+release_font_set(const char *name, const char *locale)
 {
     std::list<FSCache>::iterator it;
     for (it = fs_cache.begin(); it != fs_cache.end(); it++) {
-	if (!strcmp(it->name, name)) {
+	if (!strcmp(it->name, name) && !strcmp(it->locale, locale)) {
 	    it->refc--;
 	    if (!it->refc) {
 		XFreeFontSet(XimServer::gDpy, it->fs);
+		free(it->name);
+		free(it->locale);
 		fs_cache.erase(it);
 	    }
-	    return ;
+	    return;
 	}
     }
 }
@@ -135,7 +129,7 @@ void pe_stat::new_segment(int s)
 
 void pe_stat::push_uchar(uchar c)
 {
-    std::list<pe_ustring>::reverse_iterator i= ustrings.rbegin();
+    std::list<pe_ustring>::reverse_iterator i = ustrings.rbegin();
     (*i).s.push_back(c);
 }
 
@@ -143,7 +137,7 @@ int pe_stat::get_char_count()
 {
     std::list<pe_ustring>::iterator i;
     uString::iterator j;
-    int k=0;
+    int k = 0;
     for (i = ustrings.begin(); i != ustrings.end(); i++) {
 	for (j = (*i).s.begin(); j != (*i).s.end(); j++) {
 	    k++;
@@ -158,6 +152,7 @@ icxatr::icxatr()
     change_mask = 0;
     font_set_name = NULL;
     font_set = NULL;
+    m_locale = NULL;
     foreground_pixel = BlackPixel(XimServer::gDpy, DefaultScreen(XimServer::gDpy));
     background_pixel = WhitePixel(XimServer::gDpy, DefaultScreen(XimServer::gDpy));
 }
@@ -165,71 +160,71 @@ icxatr::icxatr()
 icxatr::~icxatr()
 {
     if (font_set_name) {
-	release_font_set(font_set_name);
+	release_font_set(font_set_name, m_locale);
 	free((void *)font_set_name);
     }
+    if (m_locale)
+	free(m_locale);
 }
 
 bool icxatr::has_atr(int id)
 {
-    return atr_mask & (1<<id);
+    return atr_mask & (1 << id);
 }
 
 void icxatr::set_atr(int id, C8 *val, int len, int o)
 {
     switch (id) {
     case ICA_InputStyle:
-	input_style = readC32(val,o);
+	input_style = readC32(val, o);
 	break;
     case ICA_ClientWindow:
-	client_window = readC32(val,o);
+	client_window = readC32(val, o);
 	break;
     case ICA_FocusWindow:
-	focus_window = readC32(val,o);
+	focus_window = readC32(val, o);
 	break;
     case ICA_Foreground:
-	foreground_pixel = readC32(val,o);
+	foreground_pixel = readC32(val, o);
 	break;
     case ICA_Background:
-	background_pixel = readC32(val,o);
+	background_pixel = readC32(val, o);
 	break;
     case ICA_SpotLocation:
-	spot_location.x = readC16(val,o);
-	spot_location.y = readC16(&val[2],o);
+	spot_location.x = readC16(val, o);
+	spot_location.y = readC16(&val[2], o);
 	break;
     case ICA_FontSet:
     {
-	int len = readC16(val,o);
-	char *new_fsn = (char *)alloca(len+1);
+	int len = readC16(val, o);
+	char *new_fsn = (char *)alloca(len + 1);
 	new_fsn[len] = 0;
 	memcpy(new_fsn, &val[2], len);
-	if (font_set_name && !strcmp(font_set_name, new_fsn)) {
+	if (font_set_name && !strcmp(font_set_name, new_fsn))
 	    break;
-	}
 
-	if (font_set_name){
+	if (font_set_name)
 	    free(font_set_name);
-	}
-	font_set_name = strdup(new_fsn);
 
-	font_set = get_font_set(font_set_name);
+	font_set_name = strdup(new_fsn);
+	font_set = get_font_set(font_set_name, m_locale);
     }
     break;
     case ICA_Area:
     {
-	area.x = readC16(&val[0],o);
-	area.y = readC16(&val[2],o);
-	area.width = readC16(&val[4],o);
-	area.height = readC16(&val[6],o);              
+	area.x = readC16(&val[0], o);
+	area.y = readC16(&val[2], o);
+	area.width = readC16(&val[4], o);
+	area.height = readC16(&val[6], o);
     }
     break;
     case ICA_LineSpace:
-	line_space=readC16(val,o);
+	line_space=readC16(val, o);
 	break;
     default:
-	//̤ΤΥȥӥ塼
-	printf("try to set unknown ic attribute %d.\n",id);
-	return ;
+	// unknown attribute
+	printf("try to set unknown ic attribute %d.\n", id);
+	return;
     }
     atr_mask |= (1 << id);
     change_mask |= (1 << id);
@@ -237,67 +232,66 @@ void icxatr::set_atr(int id, C8 *val, int len, int o)
 
 bool icxatr::is_changed(int id)
 {
-    if (change_mask & (1<<id)) {
+    if (change_mask & (1 << id))
 	return true;
-    }
+
     return false;
 }
 
 void icxatr::unset_change_mask(int id)
 {
-    change_mask &=(~(1<<id));
+    change_mask &=(~(1 << id));
 }
 
 void icxatr::print()
 {
-    if (has_atr(ICA_InputStyle)) {
-        printf("input-style %ld.\n", input_style);
-    } else {
-        printf("input-style undefined.\n");
-    }
-    if (has_atr(ICA_ClientWindow)) {
-        printf("client-window id %ld.\n", client_window);
-    } else {
-        printf("client-window id undefined.\n");
-    }
-    if (has_atr(ICA_FocusWindow)) {
-        printf("focus-window id %ld.\n", focus_window);
-    }else{
-        printf("focus-window id undefined.\n");
-    }
-    /*Preedit Attributes϶Ūͤޤ*/
-    if (has_atr(ICA_Foreground)) {
-        printf("foreground-pixel %d.\n", foreground_pixel);
-    } else {
-        printf("foreground-pixel undefined.\n");
-    }
-    if (has_atr(ICA_Background)) {
-        printf("background-pixel %d.\n", background_pixel);
-    } else {
-        printf("background-pixel undefined.\n");
-    }
-    if (has_atr(ICA_SpotLocation)) {
-        printf("spot location x=%d,y=%d.\n",
+    if (has_atr(ICA_InputStyle))
+	printf("input-style %ld.\n", input_style);
+    else
+	printf("input-style undefined.\n");
+
+    if (has_atr(ICA_ClientWindow))
+	printf("client-window id %ld.\n", client_window);
+    else
+	printf("client-window id undefined.\n");
+
+    if (has_atr(ICA_FocusWindow))
+	printf("focus-window id %ld.\n", focus_window);
+    else
+	printf("focus-window id undefined.\n");
+
+    if (has_atr(ICA_Foreground))
+	printf("foreground-pixel %d.\n", foreground_pixel);
+    else
+	printf("foreground-pixel undefined.\n");
+
+    if (has_atr(ICA_Background))
+	printf("background-pixel %d.\n", background_pixel);
+    else
+	printf("background-pixel undefined.\n");
+
+    if (has_atr(ICA_SpotLocation))
+	printf("spot location x=%d,y=%d.\n",
 	       spot_location.x, spot_location.y);
-    } else {
-        printf("spot location undefined.\n");
-    }
-    if (has_atr(ICA_FontSet)) {
-        printf("font-set-name %s\n", font_set_name);
-    } else {
-        printf("font-set-name undefined.\n");
-    }
-    if (has_atr(ICA_Area)) {
-        printf("area = x=%d y=%d width=%d height=%d.\n",
+    else
+	printf("spot location undefined.\n");
+
+    if (has_atr(ICA_FontSet))
+	printf("font-set-name %s\n", font_set_name);
+    else
+	printf("font-set-name undefined.\n");
+
+    if (has_atr(ICA_Area))
+	printf("area = x=%d y=%d width=%d height=%d.\n",
 	       area.x, area.y, area.width, area.height);
-    } else {
-        printf("area undefined.\n");
-    }
-    if (has_atr(ICA_LineSpace)) {
-        printf("line-space %d.\n", line_space);
-    } else {
-        printf("line-space undefined.\n");
-    }
+    else
+	printf("area undefined.\n");
+
+    if (has_atr(ICA_LineSpace))
+	printf("line-space %d.\n", line_space);
+    else
+	printf("line-space undefined.\n");
+
 }
 
 int icxatr::getSize(int id)
@@ -313,7 +307,14 @@ int icxatr::getSize(int id)
     return 0;
 }
 
-XimIC::XimIC(Connection *c, int imid, int icid)
+void icxatr::set_locale_name(const char *locale)
+{
+    if (m_locale)
+	free(m_locale);
+    m_locale = strdup(locale);
+}
+
+XimIC::XimIC(Connection *c, int imid, int icid, const char *engine)
 {
     mConn = c;
     mIMid = imid;
@@ -321,28 +322,29 @@ XimIC::XimIC(Connection *c, int imid, int icid)
     mIsActive = false;
 
     XimServer *svr = mConn->getXimServer();
-    m_kkContext = svr->createContext(this);
+    m_kkContext = svr->createContext(this, engine);
+
+    const char *locale = m_kkContext->get_locale_name();
+    m_xatr.set_locale_name(locale);
+
     mConvdisp = 0;
-    if (g_option_mask & OPT_TRACE) {
+    if (g_option_mask & OPT_TRACE)
 	printf("imid=%d, icid=%d ic created.\n", mIMid, mICid);
-    }
 }
 
 XimIC::~XimIC()
 {
-    if (g_option_mask & OPT_TRACE) {
-	printf("imid=%d, icid=%d ic deleted.\n",mIMid ,mICid);
-    }
-    unsetFocus();
-    if (current_ic == this) {
-	current_ic = 0;
-    }
+    if (g_option_mask & OPT_TRACE)
+	printf("imid=%d, icid=%d ic deleted.\n", mIMid , mICid);
 
-    //νϽ
+    unsetFocus();
+    if (current_ic == this)
+	current_ic = 0;
+
+    // The sequence is important.
     delete m_kkContext;
-    if (mConvdisp) {
+    if (mConvdisp)
 	delete mConvdisp;
-    }
 }
 
 bool XimIC::isActive()
@@ -362,71 +364,74 @@ int XimIC::get_imid()
 
 void XimIC::setFocus()
 {
-    if (!mIsActive) {
-	nrActiveIC ++; 
-    }
+    if (!mIsActive)
+	nrActiveIC++;
+
     current_ic = this;
     mIsActive = true;
-    ui_update_ic_stat();
-    ui_update_input_mode(m_kkContext->getMode());
     m_kkContext->focusIn();
+
     if (mConvdisp) {
-	mConvdisp->set_focus();
+	// Updating preedit here causes string mismatch if the context
+	// receives XIM_RESET_IC after XIM_SET_IC_FOCUS.  Should only
+	// update candidate window.
+	if (m_kkContext->hasActiveCandwin()) {
+	    mConvdisp->move_candwin();
+	    m_kkContext->candidate_update();
+	} else {
+	    mConvdisp->unset_focus();
+	}
     }
 }
 
 void XimIC::unsetFocus()
 {
-    if (!mIsActive) {
-	return ;
-    }
-    mIsActive = false;
-    nrActiveIC --;
-    if (current_ic == this) {
-	ui_update_ic_stat();
-    }
+    if (!mIsActive)
+	return;
 
+    mIsActive = false;
+    nrActiveIC--;
     m_kkContext->focusOut();
-    if (mConvdisp) {
-	mConvdisp->unset_focus();
-    }
+    // Since the sequence of XIM_SET_IC_FOCUS and XIM_UNSET_FOCUS
+    // events is not consistent, unsetting focus of candidate window
+    // is now handled in XimIC::setFocus() and focus_in message from
+    // helper application.
+    //
+    // if (mConvdisp && m_kkContext->hasActiveCandwin()) {
+    //     mConvdisp->unset_focus();
+    // }
 }
 
 void XimIC::OnKeyEvent(keyEventX e)
 {
     int s;
-    keyState k(&e);
+    m_keyState.check_key(&e);
 
-    s = m_kkContext->pushKey(&k);
-    if (s & COMMIT_RAW) {
+    s = m_kkContext->pushKey(&m_keyState);
+    if (s & COMMIT_RAW)
 	send_key_event(&e.ev.xkey);
-    }
 
-    if (s & UPDATE_MODE) {
-	int mode = m_kkContext->getMode();
-	ui_update_input_mode(mode);
-    }
+    if (s & UPDATE_MODE)
+	onSendPacket(); // send XIM_COMMIT
 }
 
-void XimIC::changeMode(int s)
+void XimIC::changeContext(const char *engine)
 {
-    m_kkContext->setMode(s);
-    
-    ui_update_input_mode(m_kkContext->getMode());
+   m_kkContext->changeContext(engine);
 }
 
 void XimIC::send_key_event(XKeyEvent *e)
 {
     TxPacket *t;
-    t = createTxPacket(XIM_FORWARD_EVENT,0);
+    t = createTxPacket(XIM_FORWARD_EVENT, 0);
     t->pushC16(mIMid);
     t->pushC16(mICid);
-    t->pushC16(1); //flag , synchronous
-    t->pushC16((e->serial>>16)&0xffff);
+    t->pushC16(1); // flag, synchronous
+    t->pushC16((e->serial >> 16) & 0xffff);
 
     t->pushC8(e->type);
     t->pushC8(e->keycode);
-    t->pushC16(e->serial &0xffff);
+    t->pushC16(e->serial & 0xffff);
     t->pushC32(e->time);
     t->pushC32(e->root);
     t->pushC32(e->window);
@@ -438,7 +443,7 @@ void XimIC::send_key_event(XKeyEvent *e)
     t->pushC16(e->state);
     t->pushC8(e->same_screen);
     t->pushC8(0);
-    mConn->push_passive_packet(t);
+    mConn->push_packet(t);
 }
 
 void XimIC::commit_string(const char *str)
@@ -450,12 +455,14 @@ void XimIC::commit_string(const char *str)
 
 void XimIC::extra_input(char *s)
 {
-    if (!m_kkContext->extra_input(s)) {
-        commit_string(s);
-    }
-    if (m_xatr.has_atr(ICA_FocusWindow)) {
-        force_event(m_xatr.focus_window);
-    }
+    if (!m_kkContext->extra_input(s))
+	commit_string(s);
+    if (m_xatr.has_atr(ICA_FocusWindow))
+	force_event(m_xatr.focus_window);
+}
+
+void XimIC::force_send_packet(void) {
+    (dynamic_cast<XConnection *>(mConn))->writeProc();
 }
 
 void XimIC::setICAttrs(void *val, int len)
@@ -463,7 +470,7 @@ void XimIC::setICAttrs(void *val, int len)
     unsigned char *p=(unsigned char *)val;
     int byte_order = mConn->byte_order();
     int i;
-    for (i = 0; i < len ;) {
+    for (i = 0; i < len;) {
 	int atr_id, atr_len;
 	atr_id = readC16(&p[i], byte_order);
 	i += 2;
@@ -472,10 +479,10 @@ void XimIC::setICAttrs(void *val, int len)
 	i += 2;
 
 	unsigned char *q;
-	q = (unsigned char *)alloca(atr_len+pad4(atr_len));
+	q = (unsigned char *)alloca(atr_len + pad4(atr_len));
 
 	int j;
-	for(j = 0; j < atr_len +pad4(atr_len); j++, i++) {
+	for (j = 0; j < atr_len + pad4(atr_len); j++, i++) {
 	    q[j] = p[i];
 	}
 
@@ -486,21 +493,21 @@ void XimIC::setICAttrs(void *val, int len)
 int XimIC::get_ic_atr(int id, TxPacket *t)
 {
     int l = m_xatr.getSize(id);
-    if (!t) {
+    if (!t)
 	return l;
-    }
+
     switch (id) {
     case ICA_FocusWindow:
 	t->pushC32(m_xatr.focus_window);
 	break;
     case ICA_FilterEvents:
-	t->pushC32(KeyPressMask);
+	t->pushC32(KeyPressMask|KeyReleaseMask);
 	break;
     case ICA_InputStyle:
 	t->pushC32(m_xatr.input_style);
   	break;
     default:
-	printf("try to get unknown ic attribute %d.\n",id);
+	printf("try to get unknown ic attribute %d.\n", id);
 	break;
     }
     return l;
@@ -511,67 +518,75 @@ int XimIC::lookup_style(unsigned long s)
     int i;
     struct input_style *is = mConn->getXimServer()->getInputStyles();
     for (i = 0; is[i].x_style; i++) {
-        if (is[i].x_style == (int)s) {
-            return is[i].style;
-        }
+	if (is[i].x_style == (int)s)
+	    return is[i].style;
     }
     return IS_INVALID;
 }
 
-void XimIC::set_ic_attr(int id,C8 *val, int len)
+void XimIC::set_ic_attr(int id, C8 *val, int len)
 {
-    if (id == ICA_PreeditAttribute || id == ICA_StatusAttributes) {
-        // attributeꥹ
-	setICAttrs(val, len);
-    } else {
+    if (id == ICA_PreeditAttribute || id == ICA_StatusAttributes)
+	setICAttrs(val, len); // list of attribute
+    else
 	m_xatr.set_atr(id, val, len, mConn->byte_order());
-    }
-    if (mConvdisp) {
+
+    if (mConvdisp)
 	mConvdisp->update_icxatr();
-    }
-    if (!mConvdisp && m_xatr.has_atr(ICA_InputStyle)) {
-	mConvdisp = create_convdisp(lookup_style(m_xatr.input_style),
+    else {
+	if (m_xatr.has_atr(ICA_InputStyle)) {
+	    mConvdisp = create_convdisp(lookup_style(m_xatr.input_style),
 				    m_kkContext, &m_xatr, mConn);
-	m_kkContext->setConvdisp(mConvdisp);
-    }
-    if (mConvdisp) {
-	mConvdisp->update_icxatr();
+	    m_kkContext->setConvdisp(mConvdisp);
+
+	    if (mConvdisp)
+		mConvdisp->update_icxatr();
+	}
     }
 }
 
 void XimIC::reset_ic()
 {
     TxPacket *t;
-    t = createTxPacket(XIM_RESET_IC_REPLY,0);
+    t = createTxPacket(XIM_RESET_IC_REPLY, 0);
     t->pushC16(mIMid);
     t->pushC16(mICid);
 
-    uString *s;
+    uString s;
+    const char *encoding = get_encoding();
 
-    //FIXME:m_kkContext->clear() is not implement yet
-    /*  s = m_kkContext->clear();
-    if (s) {
-        char *p;
-        int i,len;
-        p = mConn->getXimServer()->uStringToCtext(s);
-        len = strlen(p);
-        t->pushC16(len);
-        for (i = 0 ; i < len; i++) {
-            t->pushC8(p[i]);
-        }
-        len = pad4(len+2);
-        for (i = 0; i < len; i++) {
-            t->pushC8(p[i]);
-        }
-        free(p);
-        delete s;
-	}else{*/
-        t->pushC16(0);//ߥåȤʸĹ
-        // ǥߥåȤʸ롣
-        t->pushC16(0);//padding ( 2bytes)
-	//    }
+    // m_kkContext->get_preedit_string() returns uncommitted preedit
+    // strings, which will be committed in client applications.
+    s = m_kkContext->get_preedit_string();
+    if (s.size()) {
+	char *p;
+	int i, len = 0;
+	p = mConn->getXimServer()->uStringToCtext(&s, encoding);
+	if (p) {
+	    len = strlen(p);
+	    t->pushC16(len); // length of committed strings
+	    for (i = 0; i < len; i++) {
+		t->pushC8(p[i]); // put string here
+	    }
+	    len = pad4(len + 2);
+	    for (i = 0; i < len; i++) {
+		t->pushC8(0); // padding (len + 2bytes)
+	    }
+	    free(p);
+	} else {
+	    t->pushC16(0);
+	    t->pushC16(0);
+	}
+    } else {
+	t->pushC16(0);
+	t->pushC16(0);
+    }
 
     mConn->push_packet(t);
+
+    // After sending RESET_IC_REPLY, uim-xim needs to clear and reset
+    // the input context.
+    m_kkContext->clear();
 }
 
 Convdisp *XimIC::get_convdisp()
@@ -581,29 +596,34 @@ Convdisp *XimIC::get_convdisp()
 
 void XimIC::onSendPacket()
 {
-    if (!mPending.size()) {
-        return;
-    }
+    if (!mPending.size())
+	return;
+
     char *p;
-    p = mConn->getXimServer()->uStringToCtext(&mPending);
+    const char *encoding = get_encoding();
+    p = mConn->getXimServer()->uStringToCtext(&mPending, encoding);
+
+    if (!p) {
+	erase_ustring(&mPending);
+	return;
+    }
 
     TxPacket *t;
     t = createTxPacket(XIM_COMMIT, 0);
     t->pushC16(mIMid);
     t->pushC16(mICid);
 
-    t->pushC16(3); //XLookupChars|Ʊ
+    t->pushC16(3); // XLookupChars|synchronous
 
-    int i,len;
+    int i, len;
     len = strlen(p);
 
     t->pushC16(len);
-
-    for (i = 0 ; i < len ; i++) {
+    for (i = 0; i < len; i++) {
 	t->pushC8(p[i]);
     }
     len = pad4(len);
-    for (i = 0 ; i < len ; i++) {
+    for (i = 0; i < len; i++) {
 	t->pushC8(0);
     }
     free(p);
@@ -620,27 +640,39 @@ XimIC *XimIC::get_current_ic()
 
 bool XimIC::isAnyActive()
 {
-    if ( nrActiveIC ){
+    if (nrActiveIC)
 	return true;
-    }
+
     return false;
 }
 
-XimIC *create_ic(Connection *c,RxPacket *p,int imid,int icid)
+const char *XimIC::get_encoding()
+{
+    XimIM *im = get_im_by_id(mIMid);
+    return im->get_encoding();
+}
+
+const char *XimIC::get_lang_region()
+{
+    XimIM *im = get_im_by_id(mIMid);
+    return im->get_lang_region();
+}
+
+XimIC *create_ic(Connection *c, RxPacket *p, int imid, int icid, const char *engine)
 {
     XimIC *ic;
-    ic = new XimIC(c,imid,icid);
+    ic = new XimIC(c, imid, icid, engine);
     p->rewind();
-    p->getC16();//discard
+    p->getC16(); // discard
     int atr_len = p->getC16();
     unsigned char *v;
     v = (unsigned char *)alloca(atr_len);
     int i;
-    for ( i = 0 ; i < atr_len ;i++){
+    for (i = 0; i < atr_len; i++) {
 	v[i] = p->getC8();
     }
-    ic->setICAttrs((void *)v,atr_len);
-    if ( !ic->get_convdisp()){
+    ic->setICAttrs((void *)v, atr_len);
+    if (!ic->get_convdisp()) {
 	delete ic;
 	return 0;
     }
