@@ -41,68 +41,61 @@
 #include <list>
 #include <vector>
 #include <map>
-#include <uim/uim.h>
-#include <uim/uim-helper.h>
+#include "uim/uim.h"
+#include "uim/uim-helper.h"
 
 
 bool pretrans_register();
 int connection_setup();
 void main_loop();
 
-//ѴΥݡ
-// uString ʸ
-
-// preeditν
+// preedit ornament
 #define PE_NORMAL 0
 #define PE_REVERSE 1
 #define PE_UNDERLINE 2
 #define PE_HILIGHT 4
 
-typedef int uchar;
+typedef wchar_t uchar;
 typedef std::list<uchar> uString;
-struct pe_ustring{
+struct pe_ustring {
     uString s;
     int stat;
 };
 
-// ץꥨǥåȤξ
-// ˴InputContextΥ󥹥ȥ饯ǥȥ饯ǹԤʤ
+// state of preedit.
+// created in the constructor of InputContext, and deleted in the
+// destructor of it.
 class pe_stat {
 public:
     pe_stat(class InputContext *);
     void clear();
     void new_segment(int s);
-    void push_uchar(uchar );
+    void push_uchar(uchar);
     int get_char_count();
     int caret_pos;
-    std::list<pe_ustring> ustrings; //ʸڤ줿Ρ
+    std::list<pe_ustring> ustrings; // separated with segments
     class InputContext *cont;
 };
 
 void print_ustring(uString *s);
 void erase_ustring(uString *s);
-void append_ustring(uString *d,uString *s);
-void str_to_ustring(uString *d,char *s);// JIS or EUC
+void append_ustring(uString *d, uString *s);
 
-//
-void init_converter();
-
-void global_finalize(int );
-
-// 桼󥿡եȤ
+// user interfaces
 void init_convdisp();
+void init_modifier_keys();
 
 // misc
-int pad4(int );
+int pad4(int);
 
 // debug functions
-void hex_dump(unsigned char *buf,int len);
+void hex_dump(unsigned char *buf, int len);
 
 
-// ưץ󡢼˵ưΥޥɥ饤󥪥ץǷޤ롣
-// 饤ȤȤΤפ
+// for command line option
+// trace comunication between client
 #define OPT_TRACE 1
-// XIM̿פ
+// trace XIM connection
 #define OPT_TRACE_XIM 2
 
 
@@ -115,25 +108,32 @@ void hex_dump(unsigned char *buf,int len);
 // 
 extern int host_byte_order;
 extern int g_option_mask;
-extern int scr_width,scr_height;
+extern int scr_width, scr_height;
 
 #define DO_NOTHING 0
 #define COMMIT_RAW 1
 #define UPDATE_MODE 2
 
-// ȤΥ٥ȤѴԤʤäƱؿǲ
+// do convert from original event, and released within the same function
 class keyState {
 public:
-    keyState(class keyEventX *);
+    keyState();
+    ~keyState();
+    void check_key(class keyEventX *);
     int key();
     int modifier();
 
-    bool is_push();//򲡤Τ(ΥΤǤϤʤ)
+    bool is_push(); // for distinguish from release
     void print();
 private:
+    int revise_mod(int uim_mod);
     int mKey;
     int mModifier;
     bool m_bPush;
+    bool mAltOn;
+    bool mMetaOn;
+    bool mHyperOn;
+    bool mSuperOn;
 };
 
 class XimIC;
@@ -141,13 +141,11 @@ class Convdisp;
 class XimServer;
 class InputContext {
 public:
-    InputContext(XimServer *, XimIC *, char *);
+    InputContext(XimServer *, XimIC *, const char *);
     ~InputContext();
     uim_context getUC();
     int pushKey(keyState *e);
-    int getMode();
-    void setMode(int mode);
-    uString * clear();
+    void clear();
     void setConvdisp(Convdisp *);
     void focusIn();
     void focusOut();
@@ -155,20 +153,28 @@ public:
     XimServer *getServer();
     bool extra_input(char *s);
     void clear_preedit();
-    void pushback_preedit_string(int attr, char *str);
+    uString get_preedit_string();
+    void pushback_preedit_string(int attr, const char *str);
     void update_preedit();
     void candidate_activate(int nr, int display_limit);
     void candidate_select(int index);
+    void candidate_shift_page(int direction);
     void candidate_deactivate();
+    void candidate_update();
     void update_prop_list(const char *str);
     void update_prop_label(const char *str);
+    bool hasActiveCandwin();
+    const char *get_engine_name();
+    const char *get_locale_name();
+    void changeContext(const char *engine);
 public:
     static void commit_cb(void *, const char *);
     static void clear_cb(void *);
-    static void pushback_cb(void *, int attr, char *str);
+    static void pushback_cb(void *, int attr, const char *str);
     static void update_cb(void *);
     static void candidate_activate_cb(void *ptr, int nr, int index);
     static void candidate_select_cb(void *ptr, int index);
+    static void candidate_shift_page_cb(void *ptr, int direction);
     static void candidate_deactivate_cb(void *ptr);
     static void update_prop_list_cb(void *ptr, const char *str);
     static void update_prop_label_cb(void *ptr, const char *str);
@@ -182,40 +188,76 @@ protected:
     Convdisp *mConvdisp;
     uim_context mUc;
 private:
-    uim_context createUimContext(char *engine);
-    static InputContext *mFocusedContext;    
+    uim_context createUimContext(const char *engine);
+    static InputContext *mFocusedContext;
+    bool mCandwinActive;
+    int mDisplayLimit;
+    int mNumPage;
+    int current_cand_selection;
+    int current_page;
+    std::vector<const char *> active_candidates;
+    char *mEngineName;
+    char *mLocaleName;
 };
 
 class Locale {
 public:
-    virtual char *uStringToCtext(uString *us);
+    virtual char *uStringToCtext(uString *us, const char *encoding) = 0;
     virtual bool supportOverTheSpot();
+    virtual void set_localename_from_im_lang(const char *im_lang) = 0;
+private:
+    char *mLocaleName;
 };
 
-Locale *getLocale(char *lang);
+Locale *createLocale(const char *im_lang);
+const char *compose_localenames_from_im_lang(const char *im_lang);
+bool is_locale_included(const char *locales, const char *locale);
+// Sring returned by get_first_locale() is allocated with strdup().
+char *get_first_locale(const char *locales);
+const char *find_localename_from_encoding(const char *encoding);
+char *utf8_to_native_str(char *utf8, const char *encoding);
+int utf8_mbtowc(uchar *wc, const unsigned char *src, int src_len);
+int utf8_wctomb(unsigned char *dest, uchar wc); // dest must have size 6
 
 class XimServer {
 public:
-    XimServer(Locale *lc, char *name, char *lang);
-    ~XimServer(){};
-    InputContext *createContext(XimIC *);
+    XimServer(Locale *lc, const char *name, const char *lang);
+    ~XimServer() {};
+    InputContext *createContext(XimIC *, const char *engine);
+    void deleteContext(InputContext *);
 
-    bool setupConnection();
+    bool setupConnection(bool useDefaultIM);
     void setupNewConnection(XClientMessageEvent *ev);
-    char *uStringToCtext(uString *js);
+    char *uStringToCtext(uString *js, const char *encoding);
     void strToUstring(uString *d, const char *s);
     struct input_style *getInputStyles();
+    const char *getIMName();
+    const char *getIMLang();
+    void set_im(const char *name);
+    void changeContext(const char *engine);
+    Locale *getLocale();
 public:
     static XimServer *findServer(Window w);
     static Display *gDpy;
-    static std::map<Window , XimServer *>gServerMap;
+    static std::map<Window, XimServer *> gServerMap;
 private:
     Locale *mLocale;
     Window mSelectionWin;
     Atom mServerAtom;
-    char *mName, *mLang;
+    char *mIMName;
+    const char *mIMLang;
+    std::list<InputContext *> ic_list;
 };
 
+struct UIMInfo {
+    const char *lang;
+    const char *name;
+    Locale *locale;
+    const char *desc;
+};
+
+const char *get_im_lang_from_engine(const char *engine);
+
 #endif
 /*
  * Local variables:
